/*                               -*- Mode: C -*- 
 * head.c --- LiS stream head processing
 * Author          : Graham Wheeler, Francisco J. Ballesteros
 * Created On      : Tue May 31 22:25:19 1994
 * Last Modified By: John A. Boyd Jr.
 * RCS Id          : $Id: head.c,v 1.25 1996/01/29 18:12:19 dave Exp $
 * Purpose         : stream head processing stuff
 * ----------------______________________________________________
 *
 *
 *    Copyright (C) 1995  Graham Wheeler, Francisco J. Ballesteros,
 *                        Denis Froschauer
 *    Copyright (C) 1997-2000
 *			  David Grothe, Gcom, Inc <dave@gcom.com>
 *    Copyright (C) 2000  John A. Boyd Jr.  protologos, LLC
 *
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330, Cambridge,
 * MA 02139, USA.
 * 
 *
 *    You can reach us by email to any of
 *    gram@aztec.co.za, nemo@ordago.uc3m.es, 100741.1151@compuserve.com
 *
 *
 * Modified 2001-02-09 by Jeff Goldszer <Jeff_goldszer@cnt.com> 
 *                     Fixed bugs the following:
 *                     1) Process hangs on io stream calls (ex. close)
 *                        following a T_LOOK IOCTL failure.
 *                     2) Getmsg does not return length of 0 when a 
 *                        peer pipe is closed.
 *                     3) Process hangs in Poll System call because
 *                        peer pipe is closed and Process Poll thread 
 *                        is not notified.
 *
 * Modified 2001-08-09 by Jeff Goldszer <Jeff_goldszer@cnt.com>
 *   Prevent a streams oops when an application does a putmsg to a
 *   streams which has been hung up.
 *
 */

#ident "@(#) LiS head.c 2.81 8/28/02 14:15:27 "

/* BEWARE: should check:
 * tty stuff
 */


/*  -------------------------------------------------------------------  */
/*				 Dependencies                            */


/* LiS implementation modules used */

#include <sys/stream.h>
#include <sys/poll.h>
#include <sys/LiS/errmsg.h>
#if !defined(ERANGE) && defined(LINUX)
#undef _ERRNO_H
#define __need_Emath 1
#included <errnos.h>
#endif
#include <sys/lismem.h>			/* for lis_free_all_pages */
#include <sys/osif.h>


/*  -------------------------------------------------------------------  */
/*				   Symbols                               */

/*
 * Defines for measuring time intervals
 *
 * Each routine that is measuring its execution time will have a local
 * variable called "time_cell".  The measuring is enabled by a debug
 * bit.
 *
 * time_cell is declared as an unsigned long initialized to zero.
 */
#define	CLOCKON()	do { \
    if (LIS_DEBUG_MEAS_TIME) time_cell -= lis_hitime(); } while (0)

#define	CLOCKADD()	do { \
    if (LIS_DEBUG_MEAS_TIME) time_cell += lis_hitime(); } while (0)

#define	CLOCKOFF(item)	do { \
    if (LIS_DEBUG_MEAS_TIME) \
	 LisSetCounter(item, time_cell += lis_hitime()); } while (0)

/*
 *  timing functions - these are available outside the kernel also,
 *  depending on how lis_gettimeofday is defined.
 */

/*
 *  lis_hitime() -
 *
 *  This timer is restricted to a cycle of approx. 64 seconds.
 */
unsigned long lis_hitime(void)
{
    struct timeval	 tv ;

    lis_gettimeofday(&tv);
    return( (tv.tv_sec & 0x3f) * 1000000 + tv.tv_usec ) ;

} /* lis_hitime */

/*
 *  lis_usecs() -
 *
 *  this is similar to lis_hitime, but it isn't restricted to cycling
 *  at 64 seconds.  This has considerable semantic advantage; namely,
 *  it allows simpler comparison of times crossing a single cycle
 *  boundary because the full range of the underlying type is used.
 */
unsigned long lis_usecs(void)
{
    struct timeval tv;

    lis_gettimeofday(&tv);
    return( (tv.tv_sec * 1000000L) + tv.tv_usec );
}

/*
 *  lis_msecs() -
 *
 *  millisecond-resolution clock similar otherwise to lis_usecs().
 */
unsigned long lis_msecs(void)
{
    struct timeval tv;

    lis_gettimeofday(&tv);
    return( (tv.tv_sec * 1000L) + (tv.tv_usec / 1000L) );
}

/*
 *  lis_dsecs() -
 *
 *  decisecond(1/10th)-resolution clock similar otherwise to lis_usecs().
 *
 *  this is included because some protocols (which might be implemented as
 *  STREAMS modules) specify timer events at this resolution...
 */
unsigned long lis_dsecs(void)
{
    struct timeval tv;

    lis_gettimeofday(&tv);
    return( (tv.tv_sec * 10L) + (tv.tv_usec / 100000L) );
}

/*
 *  lis_secs() -
 *
 *  seconds-resolution clock similar otherwise to lis_usecs().
 */
unsigned long lis_secs(void)
{
    struct timeval tv;

    lis_gettimeofday(&tv);
    return(tv.tv_sec);
}

/*
 * Macro to do a putnext from the stream head.
 *
 * Do not call this macro while holding any locks since lis_runqueues may
 * call the queue runner, which may want locks that you are holding.  Nested
 * locks will probably save you, but best just not to do it.
 */
#define	PUTNEXT(q,m)				\
	do {					\
	    lis_putnext( (q), (m) ) ;		\
	    lis_runqueues();			\
	} while (0)

/*
 * Macro to determine whether there are any processes waiting on
 * a poll for a given stream.  When using kernel version 2.1 style
 * polling the queue head is a Linux specific structure, not our
 * portable poll list head.
 */
#if defined(PORTABLE_POLL)
#define	POLL_WAITING(hdp)	( (hdp)->sd_polllist.ph_list != NULL )
#elif defined(LINUX_POLL)
#define	POLL_WAITING(hdp)	( waitqueue_active(&(hdp)->sd_task_list) )
#else
#error "Either PORTABLE_POLL or LINUX_POLL must be defined"
#endif

#define	POLL_NOT_WAITING(hdp)	( ! POLL_WAITING(hdp) )

/*  -------------------------------------------------------------------  */
/*
 * This counter is used in keeping the code path tables and also in the
 * lock tracing.  It allows one to determine the relative order of 
 * traced events for forensic analysis.
 */
int		 lis_seq_cntr ;		/* used to establish ordering */

/*  -------------------------------------------------------------------  */
/*
 * Code path tracing.
 *
 * This is a compile-time option that is used for debugging LiS.
 */
typedef struct lis_code_path
{
    char	*fcn ;
    char	*file ;
    int		 line ;
    int		 cpu ;
    void	*ptr ;			/* arbitrary entry */
    long	 arg ;			/* arbitrary entry */
    int		 cntr ;			/* sequence counter */

} lis_code_path_t ;

#define	USE_CODE_PATH	1

#if defined(USE_CODE_PATH)

#define CP_SIZE		2048
lis_code_path_t		lis_code_path_tbl[CP_SIZE] ;
lis_code_path_t		*lis_code_path_ptr = lis_code_path_tbl ;
lis_spin_lock_t		lis_code_path_lock ;
static char		*lis_cp_fmt =		/* for printk */
			    "%u:CPU%u in %s() %s #%d -- ptr=%p long=0x%lx\n";

#if defined(LINUX)
#define CPCPU	smp_processor_id()
#else
#define CPCPU	0
#endif

#define CPFL(p,a,func,f,l)						\
    do {								\
	  int psw ;							\
	  lis_code_path_t *pp ;						\
	  lis_spin_lock_irqsave(&lis_code_path_lock,&psw) ;		\
	  pp = lis_code_path_ptr++ ;					\
	  if (lis_code_path_ptr == &lis_code_path_tbl[CP_SIZE])		\
	    lis_code_path_ptr = lis_code_path_tbl ;			\
	  lis_spin_unlock_irqrestore(&lis_code_path_lock,&psw) ;	\
	  pp->fcn  = (func) ;						\
	  pp->file = (f) ;						\
	  pp->line = (l) ;						\
	  pp->cpu  = CPCPU ;						\
	  pp->ptr  = (void *)(p) ;					\
	  pp->arg  = (long)(a) ;					\
	  pp->cntr = ++lis_seq_cntr ;					\
	  if (LIS_DEBUG_CP)						\
	      printk(lis_cp_fmt, pp->cntr, CPCPU,			\
		      		(func),(f),(l),(void *)(p),(a));	\
       } while (0)

#if !defined(__GNUC__) && !defined(HAS__FUNCTION__)
#define	__FUNCTION__	""
#endif

#define CP(p,a)		CPFL((p),(a),__FUNCTION__,__FILE__, __LINE__)

#else

#define CPFL(p,a,fc,f,l)	do {;} while (0)
#define CP(p,a)			do {;} while (0)

#endif

void	lis_cpfl(void *p, long a, char *fcn, char *f, int l)
{
    CPFL(p,a,fcn,f,l) ;
}

/*  -------------------------------------------------------------------  */
/*				  Glob. Vars                             */

lis_spin_lock_t	 lis_stdata_lock ;	/* protects global tables */
lis_spin_lock_t	 lis_qhead_lock ;	/* protects qhead, qtail & scanq */
lis_spin_lock_t	 lis_incr_lock ;	/* protects certain variable incrs */
volatile struct queue	*lis_qhead; /* first scheduled queue	*/
volatile struct queue	*lis_qtail; /* last scheduled queue		*/
volatile struct queue	*lis_scanqhead;	/* head of STREAMS scan queue	*/
volatile struct queue	*lis_scanqtail;	/* tail of STREAMS scan queue	*/
lis_atomic_t	 lis_in_syscall ;	/* processing a system call */
lis_atomic_t	 lis_queues_running ;	/* queues are being run now */
lis_atomic_t	 lis_runq_req_cnt ;	/* # req's to run queue runners */
int		 lis_num_cpus ;
lis_atomic_t	 lis_runq_cnt ;		/* # of qrun threads running */
volatile unsigned long lis_runq_cnts[LIS_NR_CPUS] ;
volatile unsigned long lis_queuerun_cnts[LIS_NR_CPUS] ;
volatile unsigned long lis_setqsched_cnts[LIS_NR_CPUS] ;
volatile unsigned long lis_setqsched_isr_cnts[LIS_NR_CPUS] ;
lis_atomic_t	 lis_runq_active_flags[LIS_NR_CPUS] ;
int		 lis_runq_pids[LIS_NR_CPUS] ;

int		 lis_nstrpush;	/* maximum # of pushed modules */
int		 lis_strhold;	/* if not zero str hold feature's activated*/
unsigned long    lis_strthresh;	/* configurable STREAMS memory limit */
unsigned int     lis_iocseq;	/* ioctl id */
stdata_t	*lis_stdata_head ;	/* list of stdata structures */
long		 lis_stdata_cnt ;	/* counts stdata structs in use */
int		 lis_open_cnt ;	/* count opens */
int		 lis_close_cnt ;/* count closes */
int              lis_doclose_cnt;
lis_semaphore_t  lis_stdata_sem;        /* stdata list semaphore */
lis_semaphore_t  lis_close_sem;         /* single threads closes */
lis_atomic_t	 lis_putnext_flag ;	/* prevents svc procs during putnext */
int		 lis_negstat ;		/* stat went negative */

extern long	 lis_max_mem ;		/* maximum to allocate */
extern long	 lis_max_msg_mem ;	/* maximum to allocate */
extern lis_spin_lock_t	  lis_msg_lock ;/* msg.c */
extern lis_spin_lock_t	  lis_bc_lock ; /* buffcall.c */
extern lis_spin_lock_t	  lis_mem_lock ;/* strmdbg.c */
extern lis_spin_lock_t	  lis_tlist_lock ; /* dki.c */

/*  -------------------------------------------------------------------  */
/*			   Streams Queue Structures                      */
/*
 * The stream head itself acts as a streams driver (sort-of).  So
 * it needs to have qinit structures just like other drivers.
 *
 * TBD:  We need a better way to set low and high water marks.
 * TBD:  Also packet sizes.
 */

struct module_info	strmhd_rdminfo =
		{0			/* mi_idnum */
		,"str_rput"		/* mi_idname */
		,LIS_MINPSZ		/* mi_minpsz */
		,LIS_MAXPSZ		/* mi_maxpsz */
		,0xFFFF			/* mi_hiwat */
		,0xF000			/* mi_lowat */
		} ;

struct module_info	strmhd_wrminfo =
		{0			/* mi_idnum */
		,"str_wput"		/* mi_idname */
		,LIS_MINPSZ		/* mi_minpsz */
		,LIS_MAXPSZ		/* mi_maxpsz */
		,0xFFFF			/* mi_hiwat */
		,0xF000			/* mi_lowat */
		} ;

struct qinit	strmhd_rdinit =
		{lis_strrput		/* qi_putp */
		,lis_strrsrv		/* qi_srvp */
		,NULL			/* qi_qopen */
		,NULL			/* qi_qclose */
		,NULL			/* qi_qadmin */
		,&strmhd_rdminfo	/* qi_minfo */
		,NULL			/* qi_mstat */
		} ;

struct qinit	strmhd_wrinit =
		{NULL			/* qi_putp */
		,&lis_strwsrv		/* qi_srvp */
		,NULL			/* qi_qopen */
		,NULL			/* qi_qclose */
		,NULL			/* qi_qadmin */
		,&strmhd_wrminfo	/* qi_minfo */
		,NULL			/* qi_mstat */
		} ;

struct streamtab strmhd_info = {
    &strmhd_rdinit,         /* read queue init */
    &strmhd_wrinit,         /* write queue init */
    NULL,
    NULL
};

/*  -------------------------------------------------------------------  */
/*			   Local functions & macros                      */

/*  -------------------------------------------------------------------  */

extern void lis_qdetach(queue_t *q, int do_close, int flag, cred_t *creds) ;
extern void lis_wakeup_close(caddr_t arg) ;
static int  lis_strdoioctl(struct file *f, stdata_t *hd,
			    strioctl_t *ioc, cred_t *creds,
			    int do_copyin) ;
int lis_doclose(struct inode *,struct file *,stdata_t *, cred_t *);
static void lis_free_stdata( struct stdata *hd ) ;
static void lis_deallocate_polllist( stdata_t *hd ) ;
static void deschedule(queue_t *q) ;
static void lis_check_m_sig(stdata_t *hd) ;

#ifdef LINUX
extern int  lis_assign_inode_to_file(struct file *f, struct inode *i) ;
extern void lis_start_qsched(void) ;
#endif						/* routine in linux-mdep.c */

extern void lis_show_inode_aliases(struct inode *);
extern int  lis_sleep_on_close_wt(void *q_str) ;	/* wait.c */
extern void lis_wakeup_close_wt(void *q_str) ;

/*  -------------------------------------------------------------------  */
/*			    lis_incr					 */
/*
 * Increment an integer under spin lock protection and return its 
 * incremented value.  Useful for allocating unique numbers for
 * ioctls, etc.
 */
int	lis_incr(int *intp)
{
    int		ret ;

    lis_spin_lock(&lis_incr_lock) ;
    ret = ++(*intp) ;
    lis_spin_unlock(&lis_incr_lock) ;
    return(ret) ;
}

/*  -------------------------------------------------------------------  */
/*                          lis_down_nintr				 */
/*
 * Do a "down" operation, but don't return until it either succeeds
 * or fails for some reason other than EINTR.
 */
int	lis_down_nintr(lis_semaphore_t *sem)
{
    int		rslt ;

    while ((rslt = lis_down(sem)) == -EINTR) ;
    return(rslt) ;
}

int	lis_down_nintr_head(stdata_t *hd, lis_semaphore_t *sem)
{
    int		rslt ;

    lis_clear_and_save_sigs(hd);		/* undo pending signals */
    while ((rslt = lis_down(sem)) == -EINTR)
	lis_clear_and_save_sigs(hd);		/* undo pending signals */
    lis_restore_sigs(hd);			/* restore old signals */
    return(rslt) ;
}


/*  -------------------------------------------------------------------  */
/*                            lis_clone_major				 */
int lis_clone_major(void)
{
    return(LIS_CLONE) ;		/* rtn major of clone device */
}

/*  -------------------------------------------------------------------  */
/*                              str_mux_cycle				 */

#if 0
	This whole mux-cycle business, as I see it, is completely
	artificial.  In order to construct a cycle you would have
	to do something like the following:

	    ioctl(fd2, I_LINK, fd3) ;		2-->3
	    ioctl(fd1, I_LINK, fd2) ;		1-->2-->3
	    ioctl(fd3, I_LINK, fd1) ;		3-->1-->2-->3 (cycle)

	However, you can''t do this because the third ioctl will
	fail.  fd3 has its STPLEX flag set so the ioctl is disallowed.

	--DMG

/* returns non-zero if the given link leads to cyclic graph
 */
static int str_mux_cycle(stdata_t *from, stdata_t *to)
{
    (void)from;
    (void)to;
    return(0);			/* no graph detection */
}
#endif

/*  -------------------------------------------------------------------  */
/*
 * new_muxid
 *
 * Get a new muxid.  Ensure that it is not in use already.
 */
static int
new_muxid(void)
{
    static int		 muxid ;
    int			 my_id ;
    stdata_t		*hd ;
    int			 n ;


    for (;;)
    {
	while ((my_id = lis_incr(&muxid)) < 0) muxid = 0 ; /* keep positive */

	if (lis_stdata_head == NULL) return(my_id) ;

	lis_spin_lock(&lis_stdata_lock) ;
	for (hd = lis_stdata_head, n = lis_stdata_cnt;
	     n > 0 || hd != lis_stdata_head;
	     hd = hd->sd_next, n--)
	{
	    if (hd->sd_mux.mx_index == my_id) break ;	/* in-use */
	}

	lis_spin_unlock(&lis_stdata_lock) ;
	if (n == 0)			/* whole list, no match */
	    break ;			/* can use this one */
    }

    return(my_id) ;

} /* new_muxid */

/* 
 * Link the muxed stream under hd.  Return the l_index value.
 * 
 * Use a global running muxid so that if the same driver has multiple
 * control streams it can have unique muxids for its lowers without
 * having to know which control stream goes with which lower.
 */
static int
str_link_mux( stdata_t *hd, stdata_t *muxed, int cmd)
{
    muxed->sd_mux.mx_cmd   = cmd ;	/* save cmd used to link */
#if 0
    muxed->sd_mux.mx_index = ++hd->sd_l_index ;
#else
    muxed->sd_mux.mx_index = new_muxid() ;
#endif
    muxed->sd_mux.mx_next  = hd->sd_mux.mx_hd ;

    hd->sd_mux.mx_hd = muxed ;

    return(muxed->sd_mux.mx_index) ;

}/*str_link_mux*/


/*  -------------------------------------------------------------------  */
/* 
 * This routine unlinks the given multiplexor from the given
 * stream head.  It returns >= 0 for success, < 0 for failure.
 *
 * The parameters 'i', 'f' and 'head' refer to the control stream
 * of the multiplexor.  This stream may be open and we are unlinking
 * a specific lower indexed by l_index, or it may be closing and we
 * are unlinking all lower streams which were i_linked via this
 * control stream.
 *
 * Any mux's below hd whose reference count is currently 1 will
 * be closed.  For any such mux's we recursively unlink any mux's
 * below those.
 *
 * We do not close hd.  We decrement its ref cnt if the command
 * is I_PUNLINK.
 */
static  int
lis_i_unlink(struct inode	*i,
	     struct file	*f,
	     stdata_t		*hd,
	     int		 l_index,
	     int		 cmd,
	     cred_t		*creds)
{
    stdata_t		 *hp ;
    stdata_t		 *prev;
    stdata_t		 *next;
    queue_t		 *qtop ;
    strioctl_t		  ioc;
    linkblk_t		  lnk;
    int			  err ;
    int			  rtn = 0 ;
    unsigned long  	  time_cell = 0 ;

    CLOCKON() ;
    for (hp = hd->sd_mux.mx_hd, prev = NULL, next = NULL; 
	 hp != NULL; 
	 hp = next)
    {
	next = hp->sd_mux.mx_next ;		/* link to next stream */

	if (l_index >= 0 && hp->sd_mux.mx_index != l_index)
	{
	    prev = hp ;
	    continue ;				/* skip this one */
	}

	if (cmd == I_UNLINK && hp->sd_mux.mx_cmd == I_PLINK)
	{
	    prev = hp ;
	    continue ;				/* skip this one */
	}

	lis_spin_lock(&hp->sd_lock) ;		/* lock strm head */
	/*
	 * Get permission from driver before unlinking
	 */
	for (qtop = hd->sd_wq; qtop->q_next != NULL; qtop = qtop->q_next) ;
	lnk.l_qtop   = qtop;
	lnk.l_qbot   = hp->sd_wq;
	lnk.l_index  = hp->sd_mux.mx_index ;

			    /* note I_UNLINK vs I_PLINK eliminated above */
	if (hp->sd_mux.mx_cmd == I_LINK)
	    ioc.ic_cmd = I_UNLINK ;		/* linked w/I_LINK */
	else
	    ioc.ic_cmd = I_PUNLINK;		/* linked w/I_PLINK */

	ioc.ic_timout= LIS_LNTIME;
	ioc.ic_len   = sizeof(linkblk_t);
	ioc.ic_dp    = (char*)&lnk;
	CLOCKADD() ;
	if (   (err = lis_strdoioctl(f,hd,&ioc,creds,0)) < 0
	    && cmd == I_UNLINK			/* ignore errs for I_PUNLINK */
	   )
	{
	    CLOCKON() ;
	    lis_spin_unlock(&hp->sd_lock) ;	/* unlock strm head */
	    rtn = err ;				/* will return error */
	    if (l_index < 0)			/* OK, skip this one */
	    {					/* lose the memory?? */
		prev = hp ;
		continue ;
	    }

	    CLOCKOFF(IOCTLTIME) ;
	    return(rtn);			/* can't unlink */
	}

	CLOCKON() ;
	if (hd->sd_mux.mx_hd == hp)		/* delinking elt at hd */
	    hd->sd_mux.mx_hd = hp->sd_mux.mx_next ;
	else					/* delinking elt down in list */
	    prev->sd_mux.mx_next = hp->sd_mux.mx_next ;

	hp->sd_mux.mx_next = NULL ;		/* clobber link */

	if ( LIS_DEBUG_IOCTL || LIS_DEBUG_LINK )
	{
	  printk(
	      "strioctl: I_UNLINK: ctl stream %s muxid=%d\n",
		    hd->sd_name,
		    lnk.l_index) ;
	  printk(
	      "                    l_qtop=\"%s\" l_qbot=\"%s\"->\"%s\"\n",
		    lis_queue_name(lnk.l_qtop),
		    hp->sd_name,
		    lis_queue_name(hp->sd_wq->q_next)) ;
	}

						/* restore queues */
	lis_setq(hp->sd_rq, &strmhd_rdinit, &strmhd_wrinit);
	hp->sd_wq->q_str	= hp ;		/* point q back to strm hd */
	hp->sd_rq->q_str	= hp ;		/* assume mux used these ptrs */

	if (   hp->sd_mux.mx_cmd == I_PLINK	/* was linked w/I_PLINK */
	    && LIS_REFCNT(hd) > 0		/* safety first */
	   )
       {
	    lis_atomic_dec(&hd->sd_refcnt);	/* decr ref count of head */
	    hd->sd_linkcnt-- ;			/* one less link for head */
	    if (hd->sd_inode)                   /* ASSERT: this is OK */
		lis_put_inode(hd->sd_inode);    /* one less inode ref */
       }

	hp->sd_linkcnt-- ;			/* one less I_LINK */
	F_CLR(hp->sd_flag,STPLEX);		/* no longer multiplexed */
	if (LIS_REFCNT(hp) == 1)		/* about to close */
	{
	    /*
	     * Close the lower stream.  The close routine may recurse
	     * on lis_i_unlink() if the lower stream is itself the
	     * control stream of a multiplexor.
	     *
	     * We do not have access to the file and inode of the
	     * lower stream.  Because the sd_refcnt is sitting at 1,
	     * the i_link is the only user of the stream.  The file
	     * and inode have been closed.
	     *
	     * lis_doclose unlocks the stream head "hp".
	     */
	    CLOCKADD() ;
	    lis_doclose(NULL, NULL, hp, creds) ;	/* close lower mux */
	    CLOCKON() ;
	}
	else
	if (LIS_REFCNT(hp) > 1)			/* be safe */
	{
	    lis_atomic_dec(&hp->sd_refcnt) ;	/* decr ref cnt */
	    lis_spin_unlock(&hp->sd_lock) ;	/* unlock strm head */
	}

	if (l_index >= 0) break ;		/* done */
    }

    CLOCKOFF(IOCTLTIME) ;
    return(rtn) ;

} /* lis_i_unlink */

/*  -------------------------------------------------------------------  */
/* lis_unlink_all
 *
 * This routine unlinks all the lowers from any control streams
 * that have the same major device number as the one for 'hd'.
 *
 * This is an AT&T feature in which the user uses I_PUNLINK to
 * construct a multiplexor and then exits, closing the control
 * stream.  Later, to dismantle the multiplexor, the user opens
 * just any old clone device to the top level of the multiplexor
 * and then issues an I_PUNLINK with the argument MUXID_ALL.
 * That is when this routine gets called.
 *
 * If it had been left up to me (DMG), I would have forced the user to
 * have a dedicated minor device for the control stream and 
 * re-open it for purposes of issuing the I_PUNLINK.  This clone
 * business opens a gaping security hole.
 *
 * Anyway, the technique is to traverse our linked list of all
 * stream heads looking for matches of major device number and
 * control stream-ness.  Perform an I_PUNLINK on all the qualified
 * stream heads, being careful about the list integrity since
 * streams heads being closed will get unlinked from the list.
 *
 * If we find a stream that matches, we effectively simulate the
 * user opening the stream (bump usage counts), doing the I_PUNLINK,
 * and then closing the stream.
 *
 * If there is an error unlinking the lowers then we really will
 * lose the memory that is allocated for the lower streams.  Chances
 * are that the module use counts will stay positive and LiS will not
 * be able to be unloaded.  Thus, it is not a good idea for a driver
 * to respond with an error to an I_PUNLINK ioctl.
 */
static  int
lis_unlink_all(struct inode	*i,
	     struct file	*f,
	     stdata_t		*hd,
	     cred_t		*creds)
{
    stdata_t		*hdp ;
    int			 n ;
    int			 maj ;
    int			 rtn = 0 ;
    int			 r ;

    maj = STR_MAJOR(hd->sd_dev) ;
again:
    if (lis_stdata_head == NULL)
	return(rtn) ;

    lis_spin_lock(&lis_stdata_lock) ;
    for (hdp = lis_stdata_head, n = 0;
	 n == 0 || hdp != lis_stdata_head;
	 hdp = hdp->sd_next, n++)
    {
	if (STR_MAJOR(hdp->sd_dev) != maj || hdp->sd_mux.mx_hd == NULL)
	    continue ;			/* not our major, or not a ctl stream */

	lis_spin_unlock(&lis_stdata_lock) ;	/* done w/stdata_head */
	lis_atomic_inc(&hdp->sd_refcnt);
	if (hdp != hd)
	    lis_spin_lock(&hdp->sd_lock) ;	/* lock strm head */
						/* lis_doclose unlocks */
	if ((r = lis_i_unlink(i, f, hdp, MUXID_ALL, I_PUNLINK, creds)) < 0)
	    rtn = r ;			/* save error for return */

	hdp->sd_mux.mx_hd = NULL ;	/* just to be sure */
	lis_doclose(NULL, NULL, hdp, creds) ;	/* close the mux */
	goto again ;			/* start over at head of list */
    }

    lis_spin_unlock(&lis_stdata_lock) ;
    return(rtn) ;

} /* lis_unlink_all */


/*  -------------------------------------------------------------------  */
/* lis_i_link
 *
 * Called from strioctl to perform the I_LINK (I_PLINK) function.
 *
 * Note about I_PLINK:  This causes the refcnt of the top stream
 * to be incremented so that a close of that stream will not cause
 * the stream to be deallocated and the multiplexor dismantled.
 * However, if the stream does get closed then the stdata structure
 * will be laying around in memory with refcnt==1 and nothing holding
 * onto it.  It will be their until the next reboot.  Hopefully,
 * this multiplexor can take care of itself until then.
 */
int	lis_i_link(struct inode	*i,
		   struct file	*f,
		   stdata_t	*hd,
		   int		 muxfd,
		   int		 cmd,
		   cred_t	*creds)
{
    stdata_t		*muxed;
    queue_t		*qtop ;
    linkblk_t		 lnk;
    strioctl_t		 ioc;
    int			 err ;
    unsigned long  	 time_cell = 0 ;

#define	RTN(v) do { CLOCKOFF(IOCTLTIME) ; return(v); } while (0)

    CLOCKON() ;
    if ( (muxed = lis_fd2str(muxfd)) == NULL )
	RTN(-EINVAL);

    lis_spin_lock(&muxed->sd_lock) ;		/* lock strm head */
    for (qtop = hd->sd_wq; qtop->q_next != NULL; qtop = qtop->q_next) ;
    lnk.l_qtop = qtop ;
    lnk.l_qbot = muxed->sd_wq;
    lnk.l_index= str_link_mux(hd, muxed, cmd);

    if ( LIS_DEBUG_IOCTL || LIS_DEBUG_LINK )
    {
	printk("strioctl: I_LINK: ctl stream %s l_index=%d\n",
		hd->sd_name,
		lnk.l_index) ;
	printk("                  l_qtop=\"%s\" l_qbot=\"%s\"->\"%s\"\n",
		lis_queue_name(lnk.l_qtop),
		muxed->sd_name,
		lis_queue_name(muxed->sd_wq->q_next)
		) ;
    }

    lis_setq(muxed->sd_rq,
	     hd->sd_strtab->st_muxrinit, hd->sd_strtab->st_muxwinit);

    ioc.ic_cmd	 = I_LINK;
    ioc.ic_timout= LIS_LNTIME;
    ioc.ic_len	 = sizeof(linkblk_t);
    ioc.ic_dp	 = (char*)&lnk;
    F_SET(muxed->sd_flag,STPLEX);
    lis_atomic_inc(&muxed->sd_refcnt);		/* so close won't deallocate */
    muxed->sd_linkcnt++ ;			/* one more I_LINK */
    lis_spin_unlock(&muxed->sd_lock) ;		/* unlock strm head */
    CLOCKADD() ;
    if ((err = lis_strdoioctl(f,hd,&ioc,creds,0)) < 0)
    {
	CLOCKON() ;
	lis_spin_lock(&muxed->sd_lock) ;	/* lock strm head */
	lis_atomic_dec(&muxed->sd_refcnt);	/* undo refcnt bumps */
	muxed->sd_linkcnt-- ;
	F_CLR(muxed->sd_flag,STPLEX);		/* not multiplexed */
	lis_setq(muxed->sd_rq,		/* put back old queues */
	     hd->sd_strtab->st_rdinit, hd->sd_strtab->st_wrinit);
	hd->sd_mux.mx_hd = muxed->sd_mux.mx_next ;/* undo str_link_mux effect */
	muxed->sd_mux.mx_next = NULL ;		/* clobber link */
	lis_spin_unlock(&muxed->sd_lock) ;	/* unlock strm head */
	RTN(err);
    }

    CLOCKON() ;
    lis_spin_lock(&muxed->sd_lock) ;	/* lock strm head again */

    /*
     * Set initial queue flags.
     *
     * We set the QWANTR flag of both sides of the queue so that
     * a putq into either queue will schedule the service procedure
     * to run.  The flag means that some service procedure "wants"
     * to "read" from the queue, that is, take messages out of it.
     */
    F_SET(muxed->sd_wq->q_flag,QWANTR);	/* yes, even tho it's the wrq */
    F_SET(muxed->sd_rq->q_flag,QWANTR);

    if (muxed->sd_mux.mx_cmd==I_PLINK)	/* linked w/I_PLINK */
    {
	lis_atomic_inc(&hd->sd_refcnt);	/* so a close won't deallocate */
	hd->sd_linkcnt++ ;		/* count that as a link */
	if (hd->sd_inode)               /* ASSERT: this is OK */
	    lis_grab_inode(hd->sd_inode);  /* keep inode too */
    }

    lis_spin_unlock(&muxed->sd_lock) ;	/* unlock strm head */
    CLOCKOFF(IOCTLTIME) ;
    return(lnk.l_index);

#undef RTN

} /* lis_i_link */


/*  -------------------------------------------------------------------  */
/* detach queues from stream and free the stream
 */
static void lis_dismantle(stdata_t *hd, cred_t *creds)
{
    queue_t	*q ;

    if (hd->sd_wq != NULL)			/* still have strm head queue */
    {
	q = hd->sd_rq ;
	if (!LIS_CHECK_Q_MAGIC(q)) return ;
	LisDownCounter(MSGQDSTRHD, lis_qsize(q)) ;
	lis_qdetach(q, 1, 0, creds);
	hd->sd_wq = NULL ;
    }

    lis_free_stdata(hd);			/* free the stdata structure */
}


/*  -------------------------------------------------------------------  */

/* set read options
 */
static int
set_readopt(int *flags, int rmode)
{
    long oflags=*flags;

    switch(rmode & RMODEMASK){
    case RNORM:
    case RMSGD:
    case RMSGN:
	*flags &= ~RMODEMASK ;			/* clear mode bits */
	*flags |= (rmode & RMODEMASK) ;		/* set mode */
	break;
    default:
	*flags=oflags;				/* restore original */
	return(-EINVAL);
    }

    switch(rmode & RPROTMASK){
    case RPROTNORM:
    case RPROTDAT:
    case RPROTDIS:
	*flags &= ~RPROTMASK ;			/* clear proto bits */
	*flags |= (rmode & RPROTMASK) ;		/* set read proto mask */
	break;
    case 0:					/* not being set */
	break;
    default:
	*flags=oflags;				/* restore original */
	return(-EINVAL);
    }

    return(0);
}

/* These will let you know if you should block on rd/wr as dictated by the
 * stream head and file flags. They should be called when the process should
 * block due to no free space or no msg avail.
 * Return non-zero if you shouldn't block and zero if you should do block 
 */
#define should_notblock_rd(hd,fp) (F_ISSET(fp->f_flags,O_NONBLOCK))

#define should_notblock_wr(hd,fp)					\
    (F_ISSET(fp->f_flags,O_NONBLOCK) && !F_ISSET(hd->sd_flag,OLDNDELAY))

/*  -------------------------------------------------------------------  */

/* Will add q to the scan list.  This happens as a result of the
 * scanq timer expiring.
 *
 * One minor problem:  If the kernel is going to run the streams
 * queues before returning from the timer interrupt then we are OK.
 * If, however, it is up to us to run ourselves then we have two
 * choices:
 *	1) wait until some user process enters the kernel on streams business
 *	2) run the queues here on top of the clock interrupt
 * I don't like either choice.  It would be best of the kernel ran
 * the streams queues from strategic places such as just before returning
 * from interrupt processing and just before returning to user level.
 *
 * NOTE: STREAMS queues are run by waking up a process so that everything
 *       is orderly. -- DMG
 */
static void
add_to_scanl( caddr_t arg )
{
  int		 psw, pswq;
  stdata_t	*hd = (stdata_t *) arg ;
  queue_t	*q ;

  /* We'll use a naive FIFO, this should be revisited
   */
  lis_spin_lock_irqsave(&lis_qhead_lock, &psw) ;
  q = hd->sd_wq ;			/* stream head write queue */
  hd->sd_scantimer = 0 ;		/* note that timer is not running */
  LIS_QISRLOCK(q, &pswq) ;		/* prevent ISR operations */
  if (F_ISSET(q->q_flag,(QSCAN|QPROCSOFF))) {
      LIS_QISRUNLOCK(q, &pswq) ;	/* already scheduled */
      lis_spin_unlock_irqrestore(&lis_qhead_lock, &psw) ;
      return;
  }
  LIS_QISRUNLOCK(q, &pswq) ;

  if (lis_scanqtail) 
    lis_scanqtail->q_scnxt = q;
  else
    lis_scanqhead = q;
  q->q_scnxt=NULL;
  lis_scanqtail=q;
  lis_atomic_inc(&lis_runq_req_cnt) ;
  LIS_QISRLOCK(q, &pswq) ;		/* prevent ISR operations */
  F_SET(q->q_flag,QSCAN);
  LIS_QISRUNLOCK(q, &pswq) ;
  lis_spin_unlock_irqrestore(&lis_qhead_lock, &psw) ;

  return ;

}/*add_to_scanl*/

static void
start_scanl_timer(stdata_t *hd)
{
    if (hd->sd_scantimer == 0)
	hd->sd_scantimer = lis_timeout_fcn(add_to_scanl,
					   (caddr_t) hd,
					   SECS_TO(LIS_RTIME),
					   "scan-list", MEM_TIMER
					   );  

} /* start_scanl_timer */


/*  -------------------------------------------------------------------  */
/* Returns a pointer to the stream head in the i-node after allocating it.
 *  The head is locked upon return (we expect that lis_stdata_sem is held
 *  by stropen during the call).
 */
static stdata_t *
alloc_stdata(void)
{
    queue_t *q = lis_allocq("stream-head");
    struct stdata *head ;

    if (!q) {
	if (LIS_DEBUG_OPEN)
	    printk(
		"alloc_stdata: failed to allocate queues for stream head\n" );
	return NULL;
    }

    head = (stdata_t*) ALLOCF(sizeof(stdata_t),"stream-head ");
    if (head != NULL)			/* successful allocation */
    {					/* initialize the structure */
	lis_bzero(head, sizeof(*head)) ;	/* clear to zero */
	head->magic=STDATA_MAGIC;
	lis_spin_lock_init(&head->sd_lock, "LiS_Head_Lock") ;
	lis_spin_lock(&head->sd_lock) ;	/* lock the structure */
	lis_sem_init(&head->sd_close_wt, 0) ; /* to wait for drainage */
	lis_sem_init(&head->sd_wioc, 1) ;
	lis_sem_init(&head->sd_wread, 0) ;	/* explicit init for emphasis */
	lis_sem_init(&head->sd_wwrite, 0) ;	/* explicit init for emphasis */
	lis_sem_init(&head->sd_wiocing, 0) ;	/* explicit init for emphasis */
	lis_sem_init(&head->sd_closing, 0) ;	/* explicit init for emphasis */

#if defined(LINUX_POLL)
# if defined(KERNEL_2_3)
	init_waitqueue_head(&head->sd_task_list) ;
# endif
#endif
	head->sd_wq  = WR(q);
	head->sd_rq  = RD(q);
	head->sd_rq->q_str = head->sd_wq->q_str = head;

	/*
	 * Link stdata structure into master list
	 */
	lis_inc_mod_cnt();
	LisUpCount(MODUSE);

	lis_spin_lock(&lis_stdata_lock) ;
	lis_stdata_cnt++ ;			/* count stdata structs */
	if (lis_stdata_head == NULL)
	{
	    head->sd_next   = head ;
	    head->sd_prev   = head ;
	    lis_stdata_head = head ;
	}
	else
	{
	    head->sd_next = lis_stdata_head ;
	    head->sd_prev = lis_stdata_head->sd_prev ;
	    lis_stdata_head->sd_prev->sd_next = head ;
	    lis_stdata_head->sd_prev = head ;
	}
	lis_spin_unlock(&lis_stdata_lock) ;
    }

    return(head) ;

}/*alloc_stdata*/

/*  -------------------------------------------------------------------  */
/* Releases the stream head in the i-node
 */
static void
lis_free_stdata( struct stdata *hd )
{
    mblk_t	*mp ;
    int		 sem_err ;

    /* must assert here that every pointer in the stdata struct which
     * may hold dynamic memory is NULL.
     */
    if (hd == NULL || hd->magic != STDATA_MAGIC) return ;

    lis_atomic_dec(&lis_in_syscall) ;	/* done processing a system call */
    lis_runqueues() ;
    sem_err = lis_down_nintr_head(hd, &lis_stdata_sem);
    lis_atomic_inc(&lis_in_syscall) ;	/* processing a system call */

    /*
     * Remove the head structure from the stdata list first thing
     * so that parallel execution of lookup_stdata cannot find this
     * head while we are freeing it.
     */
    lis_spin_lock(&lis_stdata_lock) ;
    if (lis_stdata_head != NULL)
    {
	if (lis_stdata_head == hd)		/* deleting head element */
	    lis_stdata_head = hd->sd_next ;	/* move down one */

	if (lis_stdata_head == hd)		/* only element */
	    lis_stdata_head = NULL ;		/* no more streams */
	else					/* link around "hd" */
	{
	    hd->sd_prev->sd_next = hd->sd_next ;
	    hd->sd_next->sd_prev = hd->sd_prev ;
	}
    }

    lis_stdata_cnt-- ;			/* count stdata structs */
    lis_spin_unlock(&lis_stdata_lock) ;
    lis_spin_unlock(&hd->sd_lock) ;	/* nobody should be spinning */
    /*
     *  if sd_from is set, we're holding the inode it references;
     *  release it now
     */
    if (hd->sd_from)  {
	if (LIS_DEBUG_CLOSE)
	    printk("free_stdata(0x%x) - releasing sd_from 0x%x(%d)\n",
		   (int)hd, (int)hd->sd_from, I_COUNT(hd->sd_from) );
	lis_put_inode(hd->sd_from);
	hd->sd_from = NULL;
    }

    if (hd->sd_close_timer)
	untimeout(hd->sd_close_timer) ;

    SEM_DESTROY(&hd->sd_close_wt) ;	/* de-initialize the semaphores */
    SEM_DESTROY(&hd->sd_wioc) ;
    SEM_DESTROY(&hd->sd_wread) ;
    SEM_DESTROY(&hd->sd_wwrite) ;
    SEM_DESTROY(&hd->sd_wiocing) ;
    SEM_DESTROY(&hd->sd_closing) ;

    if (hd->sd_wmsg != NULL)
    {
	freemsg(hd->sd_wmsg) ;
	hd->sd_wmsg = NULL ;
    }

    if (hd->sd_iocblk != NULL)
    {
	freemsg(hd->sd_iocblk) ;
	hd->sd_iocblk = NULL ;
    }

    if (hd->sd_siglist != NULL)
	lis_free_elist(&hd->sd_siglist);

    while ((mp = hd->sd_rput_hd) != NULL)	/* flush rput list */
    {
	hd->sd_rput_hd = mp->b_next ;		/* link around message */
	freemsg(mp) ;
    }

    lis_deallocate_polllist(hd) ;

    hd->magic ^= 1 ;			/* clobbers the magic number */
    FREE(hd) ;				/* give it back */
    lis_dec_mod_cnt();
    LisDownCount(MODUSE);

    if (sem_err == 0)
	lis_up(&lis_stdata_sem);

}/*lis_free_stdata*/


/*  -------------------------------------------------------------------  */

/* Will check that inode & device are a proper STREAMS device
 *
 * Note:  This whole thing used to be a macro but the linker did
 *        something strange to the code that caused it not to
 *        work properly.  Making it into a procedure is an attempt
 *        to do an end-run around this problem. - DMG
 */
#define CHECK_INO(i,n) do {						\
			    int		rslt ;				\
									\
			    if ((rslt = lis_check_ino(i,n)) < 0)	\
				return(rslt) ;				\
			} while (0)

int	lis_global_major ;
int lis_check_ino(struct inode *i, char *msg)
{
    int		maj ;

    if (!i)
    { 					
	lis_error(LIS_ERROR,msg,"called w/ null inode"); 	
	return(-EBADF); 			
    }						

    if (!S_ISCHR(i->i_mode)) 			
	return(-ENODEV);			

    maj = STR_MAJOR(i->i_rdev) ;
    lis_global_major = maj ;
    if (!LIS_DEVOK(maj))	
	return(-ENXIO);  /* was ENOTTY(?) */

    return(0) ;

} /* lis_check_ino */

/*  -------------------------------------------------------------------  */

/* copyin the buffer shown in usr strbuf struct sb, into *m.
 * As it's now it just copies the whole usr bufs to single blks.
 */
static  int
copyin_msgpart( struct file *f, stdata_t *hd, mblk_t **m, char *sb, int wroff )
{
    strbuf_t kbuf;
    int err;

    *m = NULL ;					/* NULL msg ptr by default */
    if ((err=lis_check_umem(f,VERIFY_READ,sb,sizeof(strbuf_t)))<0)
	return(err);

    						/* fetch strbuf */
    if ((err = lis_copyin(f,&kbuf,sb,sizeof(strbuf_t))) < 0)
	return(err) ;

    if (kbuf.len < 0 || kbuf.buf == NULL)	/* don't include this msg part */
	return(0) ;				/* not an error */

    /* This prevents an oops when the stream is hung up */
    if (F_ISSET(hd->sd_flag,STRHUP|STRCLOSE))
    {
        if (LIS_DEBUG_PUTMSG)
          printk("copyin_msgpart: stream %s: stream has received M_HANGUP\n",
                  hd->sd_name) ;
        return(-EIO);
    }

    /* The if and else if should never happen. But if the conditions
    ** should occur the user should know immediately.
    */
    if(!hd->sd_wq)
	printk("**ERROR --- Stream head has no write queue %lx\n",(long)hd);
    else if (!hd->sd_wq->q_next)
	printk( "**ERROR --- No driver associated stream head %lx q=%lx\n",
        				(long)hd, (long)hd->sd_wq);
    else if (   kbuf.len < (int)hd->sd_wq->q_next->q_minpsz
	     || (   hd->sd_wq->q_next->q_maxpsz != INFPSZ
		 && kbuf.len > (int)hd->sd_wq->q_next->q_maxpsz 
	        )
	    )
	return(-ERANGE);

    if ( (*m=(mblk_t*)allocb(kbuf.len+wroff,BPRI_LO))==NULL)
	return(-ENOMEM);

    (*m)->b_rptr = ((*m)->b_wptr += wroff);
    if (kbuf.len>0) {
	if ((err=lis_check_umem(f,VERIFY_READ,kbuf.buf,kbuf.len))<0)
	    return(err);
	if ((err = lis_copyin(f,(*m)->b_wptr,kbuf.buf,kbuf.len)) < 0)
	    return(err);
	(*m)->b_wptr+=kbuf.len;
    }
    return(0);
}/*copyin_msgpart*/

/*  -------------------------------------------------------------------  */
/* copyin ctl & dat intro M_PROTO+M_DATA msg
 */
static int
copyin_msg(struct file *f, stdata_t *hd,
		mblk_t **m, strbuf_t *uctl, strbuf_t *udat)
{
    int err;
    mblk_t *aux = NULL;

    if (m == NULL) return(-EINVAL) ;

    if ( uctl != NULL )
    {
	if ((err=copyin_msgpart(f,hd,m,(char*)uctl,0))<0)
	    return(err);

	if (*m != NULL)
	    lis_btype(*m) = M_PROTO;
    }
    else
	*m = NULL ;				/* no ctl msg part */

    if ( udat != NULL )
    {
	if ((err=copyin_msgpart(f,hd,&aux,(char*)udat,hd->sd_wroff))<0)
	{
	    if (*m != NULL)
		lis_freeb(*m);

	    return(err);
	}
    }

    if (*m == NULL && aux == NULL)		/* no message to send */
	return(0) ;

    if (*m != NULL && aux != NULL)		/* have 2 msgs to send */
	lis_linkb(*m,aux);			/* link them */
    else
    if (*m == NULL)				/* have no ctl msg */
	*m = aux ;				/* rtn just data msg */

    return(0);
	
}/*copyin_msg*/

/*  -------------------------------------------------------------------  */

/* copyout part of msg.
 * It will assume it's ctl and return MORECTL/MOREDATA if more data is left.
 * msg won't be touched if !doit 
 * this function doesn't free any block, just adjust b_[rw]ptr's
 * is_ctl is true for the ctl part / false for the data part.
 *
 * 'm' can be NULL which means do not copy anything, but update
 * the user's strbuf.
 */
static int
copyout_msgpart( struct file *f, mblk_t *m,
		    strbuf_t *sb, strbuf_t *usb, int is_ctl, int doit)
{
    int err;
    int rtn = 0 ;
    int chunk ;
    int mlen = 0 ;		/* # of bytes for ctl or dta part */
    int size_set = 0 ;
    int maxlen = sb->maxlen;	/* how much the user wants */
    int gotsome= 0 ;
    int mtype = -1 ;
    int rtnlen ;

    if (doit)			/* usb points to user space */
    {				/* ensure writeable */
	if (usb && (err=lis_check_umem(f,VERIFY_WRITE,usb,sizeof(strbuf_t)))<0)
	  return(err);
	if (maxlen >= 0 && (err=lis_check_umem(f,VERIFY_WRITE,sb->buf,maxlen))<0)
	  return(err);
    }
    else			/* usb is in kernel space */
    {				/* but sb->buf is in user space */
	if (maxlen >= 0 && (err=lis_check_umem(f,VERIFY_WRITE,sb->buf,maxlen))<0)
	  return(err);
    }

    for (rtnlen=0;
	 m != NULL;
	 m=m->b_cont)
    {
	chunk = 0 ;
	mtype = lis_btype(m) ;
	if (size_set == 0)				/* size not yet set */
	{
	    if (   (!is_ctl && mtype == M_DATA)		/* hit data msg */
		|| ( is_ctl && mtype != M_DATA))	/* hit ctl msg */
	    {
		mlen = lis_xmsgsize(m);			/* set length */
		size_set = 1 ;
	    }

	    if (size_set == 0) continue ;		/* still looking */
	}

	if (is_ctl && mtype == M_DATA)			/* hit data msg */
	    break ;					/* quit */

	if (!is_ctl && mtype != M_DATA)			/* non-data msg */
	    continue ;					/* keep looking */

	gotsome = 1 ;					/* blk of correct type*/
	if (maxlen < 0)					/* no place for it */
	    break ;					/* stop looking */

	chunk=lis_min(maxlen,lis_mblksize(m));
	if (chunk > 0 && (err=lis_copyout(f,m->b_rptr,sb->buf+rtnlen,chunk))<0)
	    return(err) ;

	if (doit)
	  lis_adjmsg(m,chunk);

	 rtnlen  += chunk ;
	 maxlen  -= chunk ;
	 if (maxlen <= 0)
	    break ;
    }					/*for blk*/

    err = 0 ;				/* return value */
    if (gotsome)			/* got some bytes from msg */
    {
	sb->len = rtnlen ;
	if (sb->len < mlen)		/* less than what was in msg */
	    rtn = (is_ctl?MORECTL:MOREDATA);
    }
    else
    if (mtype >= 0)			/* saw at least one msg block */
	sb->len = -1 ;			/* no blocks of that type */
    /* else leave sb->len unmodified from caller's value */

    if (doit)
    {
	if (usb != NULL)
	    err = lis_copyout(f,sb,usb,sizeof(strbuf_t));
    }
    else
	*usb = *sb ;			/* update kernels rtn structure */

    return(err < 0 ? err : rtn);

}/*copyout_msgpart*/

/*  -------------------------------------------------------------------  */
/* copyout msg to usr buffers given by ctl & dta.
 * m should be a proto+data msg.
 * this function doesn't free any block, just adjust b_[rw]ptr's
 * msg won't be touched if !doit 
 */
static int
copyout_msg( struct file *f, mblk_t *m, strbuf_t *kctl, strbuf_t *kdta,
	     strbuf_t *uctl, strbuf_t *udta, int doit)
{
    int errctl,errdta;
    if ((errctl=copyout_msgpart(f,m,kctl,uctl,1,doit))<0)
	return(errctl);
    else if ((errdta=copyout_msgpart(f,m,kdta,udta,0,doit))<0)
	return(errdta);
    return(errctl|errdta);
}/*copyout_msg*/



/*  -------------------------------------------------------------------  */
/*Returns non-zero if we should hold msg on write at the stream head.
 */
#define strholding(sd)	(lis_strhold && F_ISSET((sd)->sd_flag,STRHOLD))

/*  -------------------------------------------------------------------  */


/*  -------------------------------------------------------------------  */
/* qdetach - unlink queue from stream, close driver/module, and free queue
 */
void 
lis_qdetach(queue_t *q, int do_close, int flag, cred_t *creds)
{
    int		i ;
    int		psw ;
    int		psw_rq, psw_wq ;
    int		rslt ;
    const char	*name = NULL;
    queue_t     *next_q, *prev_q;
    queue_t	*rq ;
    queue_t	*wq ;
    stdata_t	*hd ;
    unsigned long  time_cell = 0 ;
    lis_semaphore_t	wakeup_sem ;

    CLOCKON() ;
    if (q == NULL)				/* supposed to be read queue */
    {
	CLOCKOFF(CLOSETIME) ;
	return;
    }

    wq = WR(q);
    rq = RD(q) ;
    if (rq == NULL || wq == NULL)
    {
	CLOCKOFF(CLOSETIME) ;
	return;
    }

    hd = (stdata_t *) q->q_str ;
    if (hd != NULL && hd->magic != STDATA_MAGIC)
	hd = NULL ;

    /*
     * Lock ordering:  when getting both the queue lock and the qhead_lock, get
     * the queue lock first.  See lis_qenable which needs the qhead_lock but is
     * always called with the queue lock being held already.
     */
    lis_lockq(wq) ;				/* lock the write queue */
    lis_lockq(rq) ;				/* lock the read queue */

    /*
     * This lock prevents the queue schedule list from being removed from.  It
     * ensures that there is no race concerning the value of the q_flag field
     * between us and queuerun.
     */
    lis_spin_lock_irqsave(&lis_qhead_lock, &psw) ;
    CP(rq,0) ;
    CP(wq,0) ;
    						/* excludes queuerun action */
    LIS_QISRLOCK(wq, &psw_wq) ;
    LIS_QISRLOCK(rq, &psw_rq) ;
    wq->q_flag |= QPROCSOFF ;			/* no more put/svc */
    rq->q_flag |= QPROCSOFF ;			/* no more put/svc */
    LIS_QISRUNLOCK(rq, &psw_rq) ;
    LIS_QISRUNLOCK(wq, &psw_wq) ;
    lis_spin_unlock_irqrestore(&lis_qhead_lock, &psw) ;

    for (i = 1, q = rq; i <= 2; i++, q = wq)	/* do twice */
    {
	if (!LIS_CHECK_Q_MAGIC(q)) return ;	/* lose the memory */
	if (LIS_DEBUG_CLOSE) name = lis_queue_name(q) ;

	if (do_close && q->q_qinfo != NULL && q->q_qinfo->qi_qclose != NULL)
	{
	    if (name)
		printk("qdetach: queue \"%s\": calling close routine\n", name);

	    CP(q,0) ;
	    CLOCKADD() ;		/* user routine might sleep */
	    /* call w/queue locked and marked as closing */
	    lis_atomic_dec(&lis_in_syscall) ;	/* "done" with a system call */
	    lis_runqueues() ;
	    (*q->q_qinfo->qi_qclose)(q, (q->q_next ? 0 : flag), creds);
	    lis_atomic_inc(&lis_in_syscall) ;	/* processing a system call */
	    CLOCKON() ;
	}
    }

    /*
     * Since we have the QPROCSOFF bits set it is now safe to let queuerun have
     * the queue lock if it has just removed this queue from the queue list.
     * It will notice the QPROCSOFF bits and not do anything.  We have to
     * release these locks before calling deschedule() so as to prevent a
     * deadly embrace with lis_qhead_lock.
     */
    lis_unlockq(rq) ;
    lis_unlockq(wq) ;

    /*
     * Now that the qprocs have been turned off and the close routine called
     * we need to ensure that the queues are not scheduled -- that is,
     * not in any scheduling list.  If we remove them from the list then
     * they won't be called.
     */
    deschedule(q) ;			/* uses lis_qhead_lock */

    /*
     * If either queue service procedure is running on another CPU at this very
     * instant then we need to wait until it finishes execution.  This is the
     * case when the QRUNNING, QENAB and QSCAN flags are reset.
     *
     * Use the qhead_lock to make sure that queue sched list removal is
     * stopped while we examine our conditions.  Use the ISR locks to
     * make sure that we and queuerun examine/manipulate the q_flags
     * field atomically.
     *
     * The q_wakeup_sem is used by queuerun to wake us up if the QWAITING
     * flag is set.
     *
     * Signal consideration:  Because this is qdetach it is somewhat likely
     * that the process that we are running on has been signalled.  We want the
     * task structure signals to reflect this after we are done.  But in the
     * meantime we need to clear the signals so that we can sleep on a
     * semaphore without being "awakened" by signals that were raised
     * previously.
     */
    lis_sem_init(&wakeup_sem, 0);
    rslt = 0 ;
    lis_spin_lock_irqsave(&lis_qhead_lock, &psw) ;
    LIS_QISRLOCK(wq, &psw_wq) ;
    LIS_QISRLOCK(rq, &psw_rq) ;
    wq->q_wakeup_sem = &wakeup_sem ;
    rq->q_wakeup_sem = &wakeup_sem ;
    while (   ((rq->q_flag | wq->q_flag) & (QENAB | QSCAN | QRUNNING))
	   && rslt == 0
	  )
    {
	CP(hd,q) ;
	F_SET(rq->q_flag, QWAITING) ;
	F_SET(wq->q_flag, QWAITING) ;
	LIS_QISRUNLOCK(rq, &psw_rq) ;
	LIS_QISRUNLOCK(wq, &psw_wq) ;
	lis_spin_unlock_irqrestore(&lis_qhead_lock, &psw) ;
	lis_spin_unlock(&hd->sd_lock) ;		/* unlock strm head */
	lis_up(&lis_close_sem);			/* allow other closes */

	lis_clear_and_save_sigs(hd);		/* undo pending signals */
	lis_atomic_dec(&lis_in_syscall) ;	/* "done" with a system call */
	lis_runqueues() ;
        rslt = lis_down(&wakeup_sem) ;
	lis_restore_sigs(hd);			/* restore old signals */

	lis_down_nintr_head(hd, &lis_close_sem) ;
	lis_spin_lock(&hd->sd_lock) ;		/* lock strm head */
	lis_atomic_inc(&lis_in_syscall) ;	/* processing a system call */
	lis_spin_lock_irqsave(&lis_qhead_lock, &psw) ;
	LIS_QISRLOCK(wq, &psw_wq) ;
	LIS_QISRLOCK(rq, &psw_rq) ;
	F_CLR(rq->q_flag, QWAITING) ;
	F_CLR(wq->q_flag, QWAITING) ;
    }
    CP(hd,rslt) ;
    wq->q_wakeup_sem = NULL ;
    rq->q_wakeup_sem = NULL ;
    LIS_QISRUNLOCK(rq, &psw_rq) ;
    LIS_QISRUNLOCK(wq, &psw_wq) ;
    lis_spin_unlock_irqrestore(&lis_qhead_lock, &psw) ;
    lis_sem_destroy(&wakeup_sem) ;

    if (rslt < 0)
	printk("lis_qdetach: lis_down(&wakeup_sem) returned %d\n", rslt) ;

    if (hd != NULL)
    {
	lis_spin_lock(&hd->sd_lock) ;		/* lock strm head */
	F_CLR(hd->sd_flag, STRFLUSHWT) ;
	lis_spin_unlock(&hd->sd_lock) ;		/* unlock strm head */
    }

    /*
     * Now everybody that needs to know should know that the queue
     * is going away, and no drivers or service procedures should be
     * messing with it.  It is safe to do the rest with the queues
     * locked, but without interrupt protection.
     */

    lis_lockq(wq) ;				/* lock the write queue */
    lis_lockq(rq) ;				/* lock the read queue */

    lis_flushq(wq, FLUSHALL);
    lis_flushq(rq, FLUSHALL);

    next_q = (wq->q_next ? wq->q_next : NULL);
    prev_q = (rq->q_next ? rq->q_next : NULL);

    /*
     *  if this queue is last in a FIFO/pipe stream, the next queue
     *  from a write queue will be a read queue.  We handle that
     *  case below.
     */
    if (next_q == rq)  next_q = prev_q;

    if (next_q) {
	if (name && LIS_DEBUG_CLOSE)
	    printk("qdetach: queue \"%s\": link around write queue\n", name);
	OTHER(next_q)->q_next = prev_q;
    }
    if (prev_q) {
	if (name && LIS_DEBUG_CLOSE)
	    printk("qdetach: queue \"%s\": link around read queue\n", name);
	WR(prev_q)->q_next = next_q;
    }

    if (name && LIS_DEBUG_CLOSE)
	printk("qdetach: queue \"%s\": de-allocate queue\n", name);

    CP(rq,0) ;
    lis_freeq(rq);			/* frees both sides of the queue */

    CLOCKOFF(CLOSETIME) ;

}/*lis_qdetach*/


/*  -------------------------------------------------------------------  */
/*
 *  lis_qopen() - calls the open routine associated with the read queue
 *  of q's pair (i.e., q may be a write queue).
 */
static int
lis_qopen( queue_t *q, dev_t *devp, int flag, cred_t *credp )
{
    queue_t	*rq = RD(q);
    int		 sflag;
    dev_t	 dev = makedevice( 0,0 );
    const char	*name = NULL;
    int		 err;

    if (!(rq->q_qinfo))  {	/* ensure open routine */
	if (LIS_DEBUG_OPEN)
	    printk( "qopen: \"%s\" does not have qinfo\n",
		    lis_queue_name(rq) );
	return (-ENOPKG);			/* pkg not installed */
    }

    if (!(rq->q_qinfo->qi_qopen))  return (0);  /* open not required */

    if (!devp)  devp = &dev;

    sflag = (WR(rq)->q_next ? MODOPEN : DRVOPEN);

    if (LIS_DEBUG_OPEN)  name = lis_queue_name(q);
    if (name && LIS_DEBUG_OPEN)
	printk("qopen: queue \"%s\": "
	       "calling %s open:%lx(dev=%x,flag=%x,sflag=%x)\n",
	       name, (sflag==MODOPEN?"module":"driver"),
	       (long) rq->q_qinfo->qi_qopen,
	       *devp, flag, sflag);
    
    if ((err = (*(rq->q_qinfo->qi_qopen)) ( rq, devp,
					    flag, sflag, credp )) != 0)  {
	if (LIS_DEBUG_OPEN)
	    printk("qopen: \"%s\" returned error %d\n",
		   lis_queue_name(q), err);
	return (err < 0 ? err : -err);
    } else
	return (0);
}
/*
 *   lis_qattach - set up a queue pair for driver/module
 */
static int
lis_qattach( stdata_t *head, struct streamtab *info, dev_t *devp,
	     int flags, cred_t *credp )
{
    queue_t    *newq;
    queue_t    *q = (head ? (head->sd_rq ? head->sd_rq : NULL) : NULL);
    queue_t    *rq, *wq ;
    int		psw, psw2;
    const char *name = (info ?
			(info->st_rdinit ? 
			 (info->st_rdinit->qi_minfo ?
			  info->st_rdinit->qi_minfo->mi_idname
			  : NULL) : NULL) : NULL);
    int         err;

    if (!q)  return -EIO;

    if (!name)  name = "unnamed-queue";

    if (!(newq = lis_allocq(name)))  {
	if (LIS_DEBUG_OPEN)
	    printk("qattach: failed to allocate queue for \"%s\"\n", name );
	return -ENOMEM;
    }

    /* Set up the queue interface stuff.
     * Do this before SPL to avoid nested SPLs.
     */
    if (info != NULL)
	lis_setq(newq, info->st_rdinit, info->st_wrinit);

    rq = q ;
    wq = WR(q) ;
    LIS_QISRLOCK(q, &psw) ;
    LIS_QISRLOCK(newq, &psw2) ;

    /* insert the queue, newq points to the read side of the queue */
    newq->q_next = rq;
    WR(newq)->q_next = (wq->q_next == rq ? newq : wq->q_next);
    if (wq->q_next)
    	OTHER(WR(newq)->q_next)->q_next = newq;
    wq->q_next = WR(newq);
    newq->q_str = WR(newq)->q_str = head;

    /*
     * Set initial queue flags.
     *
     * We set the QWANTR flag of both sides of the queue so that
     * a putq into either queue will schedule the service procedure
     * to run.  The flag means that some service procedure "wants"
     * to "read" from the queue, that is, take messages out of it.
     */
    F_SET(newq->q_flag,QWANTR);
    F_SET(LIS_WR(newq)->q_flag,QWANTR);		/* yes, even tho it's the wrq */

    LIS_QISRUNLOCK(newq, &psw2) ;
    LIS_QISRUNLOCK(q, &psw) ;

    if ((err = lis_qopen( newq, devp, flags, credp )) < 0)
	lis_qdetach( newq, 0, 0, credp );

    return(err);

}/*lis_qattach*/

/*  -------------------------------------------------------------------  */
/* Push this module in this stream and call its open proc.
 */
static int
push_mod( stdata_t *head, streamtab_t *mod,
	  dev_t *devp, int flags, cred_t *creds )
{
    int err;

    if (head->sd_maxpushcnt && head->sd_pushcnt == head->sd_maxpushcnt)
	return(-ENOSR);
    if ((err = lis_qattach( head, mod, devp, flags, creds )) < 0)
	return(err);

	head->sd_pushcnt++;

    return(0);
}/*push_mod*/

/*  -------------------------------------------------------------------  */
/* Pop the top module off this stream and call its close proc.
 */
static int
pop_mod( stdata_t *head, int flags, cred_t *creds )
{
    if (!head->sd_pushcnt)
	return(-EINVAL);
    head->sd_pushcnt--;
    if (SAMESTR(head->sd_wq))
	lis_qdetach( LIS_RD(head->sd_wq->q_next), 1, flags, creds );
    return(0);
}/*pop_mod*/

/*  -------------------------------------------------------------------  */

/* Set STREAMS options
 */
static void
set_options( stdata_t *hd, stroptions_t *so)
{
    /* just set head flags as requested
     */
    if (F_ISSET(so->so_flags,SO_MREADON))	F_SET(hd->sd_flag,SNDMREAD);
    if (F_ISSET(so->so_flags,SO_MREADOFF))	F_CLR(hd->sd_flag,SNDMREAD);
    if (F_ISSET(so->so_flags,SO_NDELON))	F_SET(hd->sd_flag,OLDNDELAY);
    if (F_ISSET(so->so_flags,SO_NDELOFF))	F_CLR(hd->sd_flag,OLDNDELAY);
    if (F_ISSET(so->so_flags,SO_ISTTY))		F_SET(hd->sd_flag,STRISTTY);
    if (F_ISSET(so->so_flags,SO_ISNTTY))	F_CLR(hd->sd_flag,STRISTTY);
    if (F_ISSET(so->so_flags,SO_TOSTOP))	F_SET(hd->sd_flag,STRTOSTOP);
    if (F_ISSET(so->so_flags,SO_TONSTOP))	F_CLR(hd->sd_flag,STRTOSTOP);
    if (F_ISSET(so->so_flags,SO_STRHOLD))	F_SET(hd->sd_flag,STRHOLD);
    if (F_ISSET(so->so_flags,SO_DELIM))		F_SET(hd->sd_flag,STRDELIM);
    if (F_ISSET(so->so_flags,SO_NODELIM))	F_CLR(hd->sd_flag,STRDELIM);


    if (F_ISSET(so->so_flags,SO_READOPT))
	set_readopt(&hd->sd_rdopt,so->so_flags);

    /* This chunk will set parameters
     */
    if (F_ISSET(so->so_flags,SO_WROFF))		
        hd->sd_wroff=so->so_wroff;
    if (F_ISSET(so->so_flags,SO_MINPSZ))	
        hd->sd_rq->q_minpsz= so->so_minpsz;
    if (F_ISSET(so->so_flags,SO_MAXPSZ))
	hd->sd_rq->q_maxpsz= so->so_maxpsz;
    if (F_ISSET(so->so_flags,SO_BAND))
    {
	if (F_ISSET(so->so_flags,SO_HIWAT))
	    lis_strqset(hd->sd_rq, (qfields_t) QHIWAT,
						    so->so_band, so->so_hiwat);
	if (F_ISSET(so->so_flags,SO_LOWAT))
	    lis_strqset(hd->sd_rq, (qfields_t) QLOWAT,
						    so->so_band, so->so_lowat);
    }
    else					/* set in q struct */
    {
	if (F_ISSET(so->so_flags,SO_HIWAT))
	    hd->sd_rq->q_hiwat= so->so_hiwat;
	if (F_ISSET(so->so_flags,SO_LOWAT))
	    hd->sd_rq->q_lowat= so->so_lowat;
    }
}/*set_options*/

/*  -------------------------------------------------------------------  */

/* Wake up polling processes
 */
static void
lis_wake_up_poll(stdata_t *hd, int ev)
{
#if defined(PORTABLE_POLL)
  polldat_t *pd;

  if ( LIS_DEBUG_POLL )
    printk("lis_wake_up_poll: stream %s: revents:%s\n",
	    hd->sd_name, lis_poll_events((short)ev)) ;

  for (pd=hd->sd_polllist.ph_list; pd != NULL; pd=pd->pd_next) 
  {
    pd->pd_events = (short) ev ;		/* events being caused */
    if (pd->pd_fn != NULL)			/* careful */
	(*pd->pd_fn)(pd->pd_arg);		/* this will wake up the
						 * polling process.
	  					 */
  }

  lis_select_wakeup(hd) ;			/* wake up selectors */

#elif defined(LINUX_POLL)

  if ( LIS_DEBUG_POLL )
    printk("lis_wake_up_poll: stream %s: revents:%s\n",
		hd->sd_name, lis_poll_events((short)ev)) ;

  wake_up_interruptible(&hd->sd_task_list);

#else
#error "Either PORTABLE_POLL or LINUX_POLL must be defined"
#endif

}/*lis_wake_up_poll*/

/*  -------------------------------------------------------------------  */

/* deallocate poll list
 * 
 * Called from stdata structure deallocation code.  Shortly after
 * this routine returns the stdata structure will be deallocated.
 * We wake up any processes that are sleeping on poll data structures
 * chained off of this stdata structure.  Immediately upon return
 * from the wakeup, we go through the list and nullify the pointer
 * to the list head in all the pdat structures.  This means that
 * the poll code, when it finally wakes up, may find this head
 * pointer to be NULL.  If the poll code has not delinked the
 * poll entry by then, it may be in trouble.
 */
static void
lis_deallocate_polllist( stdata_t *hd)
{
#if defined(PORTABLE_POLL)
  polldat_t *pd;

  if (hd->sd_polllist.ph_list != NULL)
      lis_wake_up_poll(hd, POLLHUP) ;
  else						/* no pollers */
      lis_select_wakeup(hd) ;			/* wake up selectors */

  for (pd=hd->sd_polllist.ph_list; pd != NULL; pd = pd->pd_next) 
  {
      pd->pd_headp = NULL ;
  }

#elif defined(LINUX_POLL)

# if defined(KERNEL_2_3)
  init_waitqueue_head(&hd->sd_task_list) ;
# else
  hd->sd_task_list = (struct wait_queue *) NULL ;  /* nothing to deallocate */
# endif

#else
#error "Either PORTABLE_POLL or LINUX_POLL must be defined"
#endif
}/*lis_deallocate_polllist*/

/*  -------------------------------------------------------------------  */
/* kill procs in elist w/ sig
 */
static void 
kill_procs( struct strevent *elist, int sig, short e)
{
    strevent_t *ev;
    int res;

    (void) sig ;
    for(ev = elist->se_next ; ev != elist; ev=ev->se_next )
	if ((ev->se_evs & e) != 0){
	    if (F_ISSET(ev->se_evs,S_BANDURG) && F_ISSET(ev->se_evs,S_RDBAND))
	    {
		if (LIS_DEBUG_SIG)
		    printk("kill_proc(pid=%d, sig=0x%x)\n", ev->se_pid, SIGURG);
		if ((res=lis_kill_proc(ev->se_pid,SIGURG,1))<0)
		    lis_error(LIS_ERROR,
			      "kill_procs","kill_proc: errno %d",res);

		if (ev->se_evs& e& ~S_RDBAND) /* other events -> sigpoll too */
		{
		    if (LIS_DEBUG_SIG)
			printk("kill_proc(pid=%d, sig=0x%x)\n",
				ev->se_pid, SIGPOLL);
		    if ((res=lis_kill_proc(ev->se_pid,SIGPOLL,1))<0)
			lis_error(LIS_ERROR,
				  "kill_procs","kill_proc: errno %d",res);
		}
	    }
	    else
	    {
		if (LIS_DEBUG_SIG)
		    printk("kill_proc(pid=%d, sig=0x%x)\n",
			    ev->se_pid, SIGPOLL);
		if ((res=lis_kill_proc(ev->se_pid,SIGPOLL,1))<0)
		    lis_error(LIS_ERROR,"kill_procs","kill_proc: errno %d",res);
	    }
	}
    
}/*kill_procs*/
/*  -------------------------------------------------------------------  */

/* Will send relevant signals to the usr on arrival of M_[PC]SIG containing
 * `sig' at the stream head.
 */
static void 
signalusr(int sig, stdata_t *hd)
{
    if (sig==SIGPOLL)
	kill_procs(hd->sd_siglist,SIGPOLL,SIGPOLL); /* ?? !!! */
    else if (F_ISSET(hd->sd_flag,STRISTTY))
	lis_kill_pg(hd->sd_pgrp,sig,1);
}/*signalusr*/
/*  -------------------------------------------------------------------  */

/* Send an M_READ msg downstream
 */
static int
snd_mread(queue_t *wq, unsigned int ulen)
{
    mblk_t *mread = allocb(sizeof(ulen),BPRI_LO);
    if (mread == NULL)
	return(-ENOMEM);
    else {
	lis_btype(mread)=M_READ;
	memcpy(mread->b_wptr,(char*)&ulen,sizeof(ulen));
	mread->b_wptr+=sizeof(ulen);
	lis_putnext(wq,mread);		/* don't use PUTNEXT here */
	return(0);
    }
}/*snd_mread*/

/* Send an M_FLUSH msg downstream
 */
static int
lis_snd_mflush(queue_t *wq, int flag, int band)
{
    mblk_t *mf = allocb(2*sizeof(int),BPRI_LO); /*sizeof mflush msg (?)*/
    if (mf == NULL)
	return(-ENOMEM);
    else {
	lis_btype(mf)=M_FLUSH;
	lis_putbyte(&(mf->b_wptr),(char)flag);
	if (F_ISSET(flag,FLUSHBAND))
	    lis_putbyte(&(mf->b_wptr),(char)band);
	lis_putnext(wq,mf);		/* don't use PUTNEXT here */
	return(0);
    }
}/*lis_snd_mflush*/

/*  -------------------------------------------------------------------  */
/* Send an ioctl negative ack  msg downstream reusing mp (M_IOCTL) msg.
 */
static int
snd_iocnak(queue_t *wq, mblk_t *mp,int code)
{
    iocblk_t *iocb = (iocblk_t*)(mp->b_rptr);
    mblk_t   *dtab;
    lis_btype(mp)=M_IOCNAK;
    iocb->ioc_error=code;
    dtab=lis_unlinkb(mp);
    lis_freemsg(dtab);
    lis_putnext(wq,mp);			/* don't use PUTNEXT here */
    return(0);
}/*snd_iocnak*/


/*  -------------------------------------------------------------------  */

/* Is this ioc blk out of sequence?
 */
static int
bad_ioc_seq( mblk_t *mp, stdata_t *hd)
{
    iocblk_t *iocb = (iocblk_t*)mp->b_rptr;

    return(iocb->ioc_id != (uint) hd->sd_iocseq);

}/*bad_ioc_seq*/

/*  -------------------------------------------------------------------  */

/* Call every open procedure in this queue
 */
static int
open_mods( stdata_t *head, dev_t *devp, int flags, cred_t *creds )
{
    queue_t *wq = (head ? head->sd_wq : NULL);
    dev_t    odev = *devp;
    int 	 err;

    if (wq == NULL)  return(-ENOPKG);		/* pkg not installed */

    while (wq)  {
	if ((err = lis_qopen( RD(wq), devp, flags, creds )) < 0)  return (err);

	if (!SAMESTR(wq))  return (0);

	if (odev != *devp)  return (0);

	wq = wq->q_next;
    }

    return(0);

}/*open_mods*/

/*  -------------------------------------------------------------------  */
/* lis_process_rput - called from read service to actually process msg
 *
 * This is running as a service procedure and, therefore, does not have
 * to worry about interrupts.
 *
 * The message was conveyed via a special list from strrput.  Now we 
 * put the message into the official read queue and wake up processes
 * that might be waiting for messages.
 *
 * LOCK ORDERING:  We are entered with both the head and the queue
 * unlocked.  We need to lock both by locking the head first, and then
 * the queue -- the same order as the stream head code in strread or
 * strgetpmsg.  We return with both locks available.
 */
void lis_process_rput(stdata_t *shead, queue_t *q, mblk_t *mp)
{
    int		typ ;

    CP(shead,mp) ;
    typ = lis_btype(mp);
    lis_spin_lock(&shead->sd_lock) ;		/* lock strm head */
    CP(shead,mp) ;
    lis_lockq(q) ;				/* lock the queue */
    CP(q,mp) ;
    if (F_ISSET(shead->sd_flag, STRCLOSE))	/* stream is closing */
    {						/* no file above stream */
	CP(q,mp) ;
	switch (typ)
	{
	case M_PCPROTO:
	case M_DATA:
	case M_PROTO:
	case M_SIG:
	case M_PCSIG:
	case M_COPYIN: 
	case M_COPYOUT:
	case M_ERROR:
	case M_HANGUP:
	case M_FLUSH:
	case M_SETOPTS:
	case M_IOCTL:				/* ioctl from below */
	    if (LIS_DEBUG_STRRPUT)
	      printk(
	       "lis_process_rput: "
	       "discarding %s from \"%s\" on closed stream %s\n",
			lis_msg_type_name(mp), lis_queue_name(backq(q)),
			shead->sd_name) ;
	    lis_freemsg(mp) ;			/* no hope of retrieving msg */
	    if (typ == M_FLUSH && F_ISSET(shead->sd_flag, STRFLUSHWT))
	    {					/* close-time flush */
		CP(q,0) ;
		F_CLR(shead->sd_flag, STRFLUSHWT) ;
		lis_wakeup_close_wt(shead) ;
	    }
	    goto return_point ;
	case M_PASSFP:
	    lis_free_passfp(mp);
	    goto return_point;
	case M_IOCNAK: 
	case M_IOCACK: 
	    if (q->q_flag & (QCLOSING | QPROCSOFF))
	    {
		CP(q,mp) ;
		mp->b_next = NULL ;
		freemsg(mp) ;
		goto return_point;
	    }
	    /* q not closing yet, go ahead and process IOCACKs */
	    break ;
	}
    }

    if (LIS_DEBUG_STRRPUT)
	printk("lis_process_rput: "
		"received %s (%d) from \"%s\" on stream %s size %d\n",
		lis_msg_type_name(mp), typ, lis_queue_name(backq(q)),
		shead->sd_name, lis_msgsize(mp)) ;

    CP(q,mp) ;
    switch(typ)
    {
    case M_PCPROTO:
	if (F_ISSET(shead->sd_flag,STRPRI)){
	    lis_freemsg(mp);
	    goto return_point;
	}
	else 
	    F_SET(shead->sd_flag,STRPRI);
	/* fall thru... */
    case M_DATA:
    case M_PROTO:
    case M_PASSFP:
	CP(mp,0) ;
	LisUpCount(MSGQDSTRHD) ;
        putq(q,mp);			/* put in official queue */
	if (   shead->sd_sigflags == 0			/* no signals */
	    && POLL_NOT_WAITING(shead)			/* no polling */
	    && !F_ISSET(shead->sd_flag,STRSELPND)	/* no selecting */
	   )
	{
	    lis_wake_up_wread(shead);	/* wake up if someone sleeping */
	    break ;
	}

	if (typ == M_PCPROTO || mp == q->q_first)
	{				/* hipri msg or 1st msg in queue */
	    short evs=0;
	    short polls=0 ;

	    if (lis_hipri(typ))
	    {
		F_SET(polls,POLLPRI) ;
		F_SET(evs,S_HIPRI);
	    }
	    else
	    {
		F_SET(polls,POLLIN) ;
		F_SET(evs,S_INPUT);

		if (lis_bband(mp)==0)
		{
		    F_SET(polls,POLLRDNORM) ;
		    F_SET(evs,S_RDNORM);
		}

		if (lis_bband(mp)>0)
		{
		    F_SET(polls,POLLRDBAND) ;
		    F_SET(evs,S_RDBAND);
		}
	    }

	    if (evs &= shead->sd_sigflags)
		kill_procs(shead->sd_siglist,SIGPOLL,evs);

	    if (   polls
		&& (   POLL_WAITING(shead)
		    || F_ISSET(shead->sd_flag,STRSELPND)
		   )
	       )
		lis_wake_up_poll(shead,polls);
	}
	lis_wake_up_wread(shead);	/* wake up if someone sleeping */
	break;
    case M_ERROR:
	{
	    unsigned char	errnum ;

	    CP(mp,0) ;
	    switch (mp->b_wptr - mp->b_rptr) 	/* how many bytes?  */
	    {
	    default:				/* not valid format */
		mp->b_rptr = mp->b_datap->db_base;
		mp->b_wptr = mp->b_rptr ;
		*mp->b_wptr++ = EINVAL ;	/* supply default errnum */
		/*
		 * Fall into the one-byte case
		 */
	    case 1:				/* set errnum for both sides */
		errnum = *mp->b_rptr++;
		if (errnum != NOERROR)
		{
		    shead->sd_rerror = errnum ;
		    shead->sd_werror = errnum ;

		    if (errnum == 0)
		    {
			F_CLR(shead->sd_flag,STRDERR);
			F_CLR(shead->sd_flag,STWRERR);
		    }
		    else
		    {
			F_SET(shead->sd_flag,STRDERR);
			F_SET(shead->sd_flag,STWRERR);
		    }
		}
		break ;

	    case 2:				/* set two sides individually */
		errnum = *mp->b_rptr++;
		if (errnum != NOERROR)
		{
		    shead->sd_rerror = errnum ;

		    if (errnum == 0)
			F_CLR(shead->sd_flag,STRDERR);
		    else
			F_SET(shead->sd_flag,STRDERR);
		}

		errnum = *mp->b_rptr++;
		if (errnum != NOERROR)
		{
		    shead->sd_werror = errnum ;

		    if (errnum == 0)
			F_CLR(shead->sd_flag,STWRERR);
		    else
			F_SET(shead->sd_flag,STWRERR);
		}

		break ;
	    }

	    lis_freemsg(mp);
	    if (F_ISSET(shead->sd_sigflags,S_ERROR))
		kill_procs(shead->sd_siglist,SIGPOLL,S_ERROR);

	    if (   POLL_WAITING(shead)
		|| F_ISSET(shead->sd_flag,STRSELPND)
	       )
		lis_wake_up_poll(shead,POLLERR);

	    lis_wake_up_all_wwrite(shead);
	    lis_wake_up_all_wread(shead);
	    if (   (shead->sd_werror || shead->sd_rerror)
		&& F_ISSET(shead->sd_flag,IOCWAIT)
	       )
		lis_wake_up_wiocing(shead);
#if 0				/* let close do the mflush */
	    if (shead->sd_werror || shead->sd_rerror)
		lis_snd_mflush(LIS_WR(q),FLUSHW,0);
#endif
	    break;
	}
    case M_HANGUP:
	CP(mp,0) ;
	lis_freemsg(mp);
	F_SET(shead->sd_flag,STRHUP);
	if (F_ISSET(shead->sd_sigflags,S_HANGUP))
	    kill_procs(shead->sd_siglist,SIGPOLL,S_HANGUP);
	if (   POLL_WAITING(shead)
	    || F_ISSET(shead->sd_flag,STRSELPND)
	   )
	    lis_wake_up_poll(shead,POLLHUP);

	if (F_ISSET(shead->sd_flag,STRATTACH) && !(shead->sd_rfdcnt)) {
	    F_CLR(shead->sd_flag,STRATTACH);
	    lis_fdetach_stream(shead);
	}

	lis_wake_up_all_wwrite(shead);
	lis_wake_up_all_wread(shead);
	break;

    case M_FLUSH:
	if ( LIS_DEBUG_FLUSH )
	    printk("lis_process_rput: "
		    "M_FLUSH from \"%s\" on stream %s flags=0x%x\n",
		    lis_queue_name(backq(q)),
		    shead->sd_name, *mp->b_rptr & 0xFF) ;
	if (*mp->b_rptr & FLUSHR)
	{					/* flush read queue */
	    int		msgs_before = lis_qsize(q) ;

	    CP(mp,0) ;
	    if (F_ISSET(*mp->b_rptr,FLUSHBAND))
		lis_flushband(LIS_RD(q),mp->b_rptr[1],FLUSHDATA);
	    else
		lis_flushq(LIS_RD(q),FLUSHDATA);
	    if (   shead->sd_rq->q_first == NULL
		|| !lis_hipri(shead->sd_rq->q_first->b_datap->db_type)
	       )
		F_CLR(shead->sd_flag,STRPRI);	/* no PCPROTO at head */
	    *mp->b_rptr &= ~FLUSHR ;		/* turn off flush read bit */
	    LisDownCounter(MSGQDSTRHD, msgs_before - lis_qsize(q)) ;
	}

	if (*mp->b_rptr & FLUSHW)
	{					/* flush write queue */
	    CP(mp,0) ;
	    if (F_ISSET(*mp->b_rptr,FLUSHBAND))
		lis_flushband(LIS_WR(q),mp->b_rptr[1],FLUSHDATA);
	    else
		lis_flushq(LIS_WR(q),FLUSHDATA);

	    if (!(mp->b_flag & MSGNOLOOP))
	    {
		mp->b_flag |= MSGNOLOOP;
		lis_spin_unlock(&shead->sd_lock) ;	/* unlock strm head */
		lis_unlockq(q) ;
		lis_putnext(LIS_WR(q),mp);	/* send back downstream */
		return ;			/* w/head unlocked */
						/* do not use PUTNEXT */
	    }
	    else
	        lis_freemsg(mp);	/* free message */
	}
	else					/* not flushing write */
	    lis_freemsg(mp);

	break;

    case M_SIG:
	CP(mp,0) ;
	putq(q,mp);			/* strread will signalusr() */
	LisUpCount(MSGQDSTRHD) ;
	lis_check_m_sig(shead) ;	/* see if it is the only msg in the q */
	break;
    case M_PCSIG:
	CP(mp,0) ;
	signalusr(*mp->b_rptr,shead);
	if (   POLL_WAITING(shead)
	    || F_ISSET(shead->sd_flag,STRSELPND)
	   )
	    lis_wake_up_poll(shead,POLLMSG);
	lis_freemsg(mp);
	break;
    case M_IOCNAK: 
    case M_IOCACK: 
	CP(mp,0) ;
	if (!F_ISSET(shead->sd_flag,IOCWAIT)|| 
	    F_ISSET(shead->sd_flag,(STWRERR|STRDERR)) ||
	    bad_ioc_seq(mp,shead)
	   )
	    lis_freemsg(mp);
	else {			/* everything's ok: now we process it */
	    CP(mp,0) ;
	    shead->sd_iocblk=mp;
	    lis_wake_up_wiocing(shead); /* this is not an unlock! */
	}
	break ;
    case M_COPYIN: 
	CP(mp,0) ;
	if (!F_ISSET(shead->sd_flag,IOCWAIT))
	    lis_freemsg(mp);
	else if (F_ISSET(shead->sd_flag,STWRERR) ||
		 F_ISSET(shead->sd_flag,STRDERR) || bad_ioc_seq(mp,shead)){
	    copyresp_t *res=(copyresp_t*)mp->b_rptr;
	    mp->b_wptr=(char*)(res+1);
	    mp->b_datap->db_type = M_IOCDATA ;
	    res->cp_rval=(caddr_t)1;
	    lis_spin_unlock(&shead->sd_lock) ;	/* unlock strm head */
	    lis_unlockq(q) ;
	    lis_putnext(LIS_WR(q),mp);		/* do not use PUTNEXT */
	    return ;				/* w/head unlocked */
	}
	else {			/* everything's ok: now we process it */
	    shead->sd_iocblk=mp;
	    lis_wake_up_wiocing(shead); /* this is not an unlock! */
	}
	break;
    case M_COPYOUT:
	CP(mp,0) ;
	if (!F_ISSET(shead->sd_flag,IOCWAIT))
	    lis_freemsg(mp);
	else if (F_ISSET(shead->sd_flag,STWRERR) ||
		 F_ISSET(shead->sd_flag,STRDERR) ||  bad_ioc_seq(mp,shead)){
	    copyresp_t *res=(copyresp_t*)mp->b_rptr;
	    mp->b_wptr=(char*)(res+1);
	    mp->b_datap->db_type = M_IOCDATA ;
	    res->cp_rval=(caddr_t)1;
	    lis_spin_unlock(&shead->sd_lock) ;	/* unlock strm head */
	    lis_unlockq(q) ;
	    lis_putnext(LIS_WR(q),mp);		/* do not use PUTNEXT */
	    return ;				/* w/head unlocked */
	}
	else {			/* everything's ok: now we process it */
	    shead->sd_iocblk=mp;
	    lis_wake_up_wiocing(shead); /* this is not an unlock! */
	}
	break;
    case M_SETOPTS:
	CP(mp,0) ;
	set_options(shead,(stroptions_t*)mp->b_rptr);
	lis_freemsg(mp);
	break ;
    case M_IOCTL:			/* upside-down ioctl */
	CP(mp,0) ;
	lis_spin_unlock(&shead->sd_lock) ;	/* unlock strm head */
	snd_iocnak(LIS_WR(q),mp,1);
	lis_unlockq(q) ;
	return ;				/* w/head unlocked */
    default:
	CP(mp,0) ;
	lis_freemsg(mp);
	break ;
    }

return_point:
    CP(mp,0) ;
    lis_unlockq(q) ;
    lis_spin_unlock(&shead->sd_lock) ;	/* unlock strm head */
}


/*  -------------------------------------------------------------------  */
/* lis_strrsrv - stream head read service procedure.
 *
 * Process messages queued by the receive put procedure.  Running this
 * as a service procedure allows us not to have to worry so much about
 * interrupts.  The lis_strrput procedure *could* be entered from interrupt
 * level.  This service procedure is guaranteed to be running at task level.
 *
 * The only interrupt level worry that we have is in taking messages out
 * of the rput list.
 *
 * LOCK ORDERING:  We are entered from the streams scheduler with the queue
 * locked.  The routine lis_process_rput needs both the head lock and the queue
 * lock.  It needs to get the head lock first, which means that we need to
 * release the qeuue lock.  Before we do that we need to mark the queue as
 * unscheduleable  by setting the QENAB bit.  This prevents another copy of
 * this service procedure from being scheduled while we process messages.  We
 * reset the flag when we are done under protection of SPL.
 */
int	lis_strrsrv(queue_t *q)
{
    mblk_t   	*mp ;
    stdata_t	*hd ;
    int		 psw ;

    if (   !LIS_CHECK_Q_MAGIC(q)
        || (hd = q->q_str) == NULL
	|| hd->magic != STDATA_MAGIC
	|| hd->sd_rq != q
       )
    {
	printk("lis_strrsrv: msg to improperly formed queue\n") ;
	return(0) ;
    }

    CP(q,0) ;
    LIS_QISRLOCK(q, &psw) ;
    q->q_flag |= QENAB;			/* prevent qenable */
    LIS_QISRUNLOCK(q, &psw) ;
    lis_unlockq(q) ;			/* unlock the queue */
					/* nothing locked now */
    LIS_QISRLOCK(q, &psw) ;
    while ((mp = hd->sd_rput_hd) != NULL)
    {
	hd->sd_rput_hd = mp->b_next ;	/* link around message */

	LIS_QISRUNLOCK(q, &psw) ;
	mp->b_next = NULL ;
	lis_process_rput(hd, q, mp) ;
	LIS_QISRLOCK(q, &psw) ;
    }

    hd->sd_rput_tl = NULL ;		/* list now empty */
    q->q_flag &= ~QENAB;		/* allow qenable */
    LIS_QISRUNLOCK(q, &psw) ;
    CP(q,0) ;
    lis_lockq(q) ;			/* re-lock queue, but not head */
    CP(q,0) ;
    return(0) ;
}


/*  -------------------------------------------------------------------  */
/* lis_strrput - stream head read put procedure
 *
 * Don't do anything but just queue the message for the service procedure.
 *
 * This routine could be called from interrupt context, though that is not good
 * programming practice.  Therefore, it needs to use SPL protection when
 * inserting the message into the list.  Because of the SPL we don't need to
 * lock the stream head.
 */
int
lis_strrput(queue_t *q, mblk_t *mp)
{
    int		 psw ;
    stdata_t	*hd ;

    if (   !LIS_CHECK_Q_MAGIC(q)
        || (hd = q->q_str) == NULL
	|| hd->magic != STDATA_MAGIC
	|| hd->sd_rq != q
       )
    {
	printk("lis_strrput: msg to improperly formed queue\n") ;
	freemsg(mp) ;
	return(0) ;
    }

    CP(q,mp) ;
    if (q->q_flag & QPROCSOFF)
    {
	CP(q,mp) ;
	if ( LIS_DEBUG_CLOSE )
	    printk("lis_strput: discarding message from closing queue\n");

	freemsg(mp) ;
	return(0) ;
    }

    mp->b_next = NULL ;
    LIS_QISRLOCK(q, &psw) ;
    /*
     * The flush wait logic in close has disable qenables so that the
     * stream head read service procedure cannot run.  We need to catch
     * this case here so that close does not have to suffer a 50ms timeout.
     */
    if (F_ISSET(hd->sd_flag, STRFLUSHWT) && mp->b_datap->db_type == M_FLUSH)
    {						/* close-time flush */
	CP(q,mp) ;
	F_CLR(hd->sd_flag, STRFLUSHWT) ;
	lis_wakeup_close_wt(hd) ;
    }

    if (hd->sd_rput_hd == NULL)
    {						/* empty list */
	hd->sd_rput_tl = mp ;
	hd->sd_rput_hd = mp ;
	LIS_QISRUNLOCK(q, &psw) ;
	CP(q,mp) ;
	qenable(q) ;				/* enable svc proc */
    }
    else
    {						/* elements in list */
	hd->sd_rput_tl->b_next = mp ;
	hd->sd_rput_tl = mp ;
	LIS_QISRUNLOCK(q, &psw) ;
	CP(q,mp) ;
    }

    return(0) ;

}/*lis_strrput*/


/*  -------------------------------------------------------------------  */
/*  lis_strwsrv
 *
 * This is the stream head write service procedure.  It is used
 * for detecting that the stream below the stream head no longer
 * has messages queued for output so the stream can be closed.
 */

int	lis_strwsrv(queue_t *q)
{
    stdata_t	*hd ;
    mblk_t	*mp ;
    int		 couldput ;

    if (!LIS_CHECK_Q_MAGIC(q)) return(0) ;

    hd = (stdata_t *) q->q_str ;
    if (hd == NULL)				/* should not happen */
    {
	printk("lis_strwsrv: queue %s: q_str is NULL\n", lis_queue_name(q));
	return(0) ;
    }
    
    if (hd->magic != STDATA_MAGIC || hd->sd_wq != q) return(0) ;

    CP(hd,0) ;
    while (   (couldput = lis_bcanputnext(q,0))	/* can send downstream */
	   && (mp = getq(q)) != NULL		/* have queued messages */
	  )
    {
	CP(q,mp) ;
	lis_putnext(q, mp) ;			/* send downstream */
						/* don't use PUTNEXT */
    }

    if (   F_ISSET(hd->sd_flag, STRCLOSE)	/* stream is closing */
	&& lis_qcountstrm(q) == 0 		/* queues are empty */
       )
   {
	if ( LIS_DEBUG_CLOSE )
	    printk("lis_strwsrv: stream %s empty, wake up close routine\n",
		    hd->sd_name) ;

	CP(hd,0) ;
	lis_wakeup_close((caddr_t) hd) ;	/* wake up close */
   }

   if (hd->sd_sigflags != 0)
   {
	if (F_ISSET(hd->sd_sigflags,S_OUTPUT)
	    && lis_bcanputnext(q,0)
	   )
	{
	    CP(hd,0) ;
	    kill_procs(hd->sd_siglist,SIGPOLL,S_OUTPUT);
	}

	/*
	 * AT&T says in STREAMS Programmer's Guide:
	 *
	 * "S_WRBAND
	 *
	 * "A priority band greater than 0 of a queue downstream exists
	 * and is writeable.  This notifies the user that there is room
	 * on the queue for sending (or writing) priority data downstream."
	 */
	if (F_ISSET(hd->sd_sigflags,S_WRBAND)
	    && lis_bcanputnext_anyband(q)
	   )
	{
	    CP(hd,0) ;
	    kill_procs(hd->sd_siglist,SIGPOLL,S_WRBAND);
	}
   }

   if (couldput && lis_atomic_read(&hd->sd_wrcnt) != 0)
					        /* write waiting flow ctrl */
   {
	CP(hd,0) ;
	lis_wake_up_all_wwrite(hd);		/* wake up all write/putmsg */
   }

    if (  (   POLL_WAITING(hd)
	   || F_ISSET(hd->sd_flag,STRSELPND)
          )
	&& lis_bcanputnext(q,0)			/* can still send downstream */
       )
    {
	lis_spin_lock(&hd->sd_lock) ;		/* lock strm head */
	CP(hd,0) ;
	lis_wake_up_poll(hd,POLLOUT);
	lis_spin_unlock(&hd->sd_lock) ;		/* unlock strm head */
    }

    return(0) ;

} /* lis_strwsrv */


/*  -------------------------------------------------------------------  */
/* This is the tmout handling function.
 * it will be called on response to a tmout and is responsible to
 * take the right actions.
 * It should be called w/ an stream head as argument.
 */
static void
lis_do_tmout( unsigned long arg )
{
    stdata_t *hd = (stdata_t*)arg;

    if (hd->magic != STDATA_MAGIC){
	lis_error(LIS_ERROR,"lis_do_tmout","timeout with bad stream 0x%x",arg);
	return;
    }
    if (F_ISSET(hd->sd_flag,STIOCTMR)){     /* ioctl timer: awake sleeper*/
	CP(hd,0) ;
	F_CLR(hd->sd_flag,STIOCTMR);
	lis_wake_up_wiocing(hd);
    }
}/*lis_do_tmout*/

static void
lis_do_rd_tmout( unsigned long arg )
{
    stdata_t *hd = (stdata_t*)arg;

    if (hd->magic != STDATA_MAGIC){
	lis_error(LIS_ERROR,"lis_do_rd_tmout",
			    "timeout with bad stream 0x%x",arg);
	return;
    }

    lis_wake_up_wread(hd);

}/*lis_do_tmout*/


/*  -------------------------------------------------------------------  */
/* lis_wait_for_wiocing - wait until ioc response arrived or timeout occur
 */
static int
lis_wait_for_wiocing(stdata_t *hd, int tmout)
{
    struct timer_list tl;
    int		      rslt ;

    if (tmout == 0)
	tmout = LIS_DFLT_TIM ;

    while( hd->sd_iocblk == NULL )
    {
	if ( F_ISSET(hd->sd_flag,(STRDERR|STWRERR)) )
	    return(-hd->sd_rerror) ;
	F_SET(hd->sd_flag,STIOCTMR);
	if (tmout != INFTIM)
	    lis_tmout(&tl,lis_do_tmout,(long)hd, SECS_TO(tmout));
	CP(hd,tmout) ;
	rslt = lis_sleep_on_wiocing(hd);	/* unlocks and relocks head */
	if (rslt < 0)
	{
	    CP(hd,rslt) ;
	    if (tmout != INFTIM)
		lis_untmout(&tl);
	    F_CLR(hd->sd_flag,STIOCTMR);
	    return(rslt) ;
	}

	if (!F_ISSET(hd->sd_flag,STIOCTMR)) /* timer expired */
	{
	    CP(hd,-ETIME) ;
	    return(-ETIME);
	}
	else
	{
	    CP(hd,0) ;
	    if (tmout != INFTIM)
		lis_untmout(&tl);
	    F_CLR(hd->sd_flag,STIOCTMR);
	}
    }
    return(0);			/* got msg */
}/*lis_wait_for_wiocing*/

/*  -------------------------------------------------------------------  */

/* copyout bmlks for reads
 */
static int 
copyout_blks(struct file *f, char *ubuff, long count, mblk_t *mp)
{
    mblk_t *mb;
    long ocount=count;
    int len;
    int err = 0 ;

    for (mb=mp ; mb && count > 0; mp=mb)
    {
	mb=mp->b_cont;				/* next buffer in chain */
	mp->b_cont = NULL ;			/* unchain this buffer */
	if (err == 0)
	{
	    len=lis_min(count,lis_mblksize(mp));	/* size of current bfr */
	    err = lis_copyout(f,mp->b_rptr,ubuff,len); /* cpy out current bfr */
	    count -=len;
	    ubuff +=len;
	}
	lis_freeb(mp);				/* free current buffer */
    }

    if (err < 0)
	return(err) ;				/* something went wrong */

    if (count != 0)			/* hopefully, copied whole buffer */
	lis_error(LIS_WARN,"copyout_blks",
		  "count (%d) doesn't match data (%d)", ocount, ocount-count);
    return(0) ;
}

/*  -------------------------------------------------------------------  */
/* strdoioctl - ioctl handler used by strioctl
 *
 * There is a special circumstance in which this routine is called
 * with a NULL pointer for 'f'.  This is from the i_unlink code in
 * which the file is not known.  We do not worry about null pointer
 * reference here since our only use of 'f' occurs with 'do_copyin'
 * is passed as a "true", which it is not from i_unlink code.
 */
static int
lis_strdoioctl(struct file *f, stdata_t *hd,
		strioctl_t *ioc, cred_t *creds, int do_copyin)
{
    mblk_t *mioc,*mdta;
    iocblk_t *iocb;
    int err = 0 ;
    int msgtype ;
    unsigned long  time_cell = 0 ;

    CLOCKON() ;
    if (ioc->ic_len != TRANSPARENT && (ioc->ic_len < 0 || ioc->ic_timout < -1))
    {
	ioc->ic_len = 0 ;		/* no data */
	CLOCKOFF(IOCTLTIME) ;
	return(-EINVAL);
    }
    if ((mioc=allocb(sizeof(copyreq_t),BPRI_LO))==NULL)
    {
	ioc->ic_len = 0 ;		/* no data */
	CLOCKOFF(IOCTLTIME) ;
	return(-ENOMEM);
    }
    if (ioc->ic_len){
	if (do_copyin && 
	    (err=lis_check_umem(f,VERIFY_READ,ioc->ic_dp,ioc->ic_len))<0)
	{
	    ioc->ic_len = 0 ;		/* no data */
	    CLOCKOFF(IOCTLTIME) ;
	    return(err);
	}
	if ((mdta=allocb(
	            ioc->ic_len == TRANSPARENT ? sizeof(ioc->ic_dp)
					       : ioc->ic_len,BPRI_LO))==NULL)
	{
	    ioc->ic_len = 0 ;		/* no data */
	    lis_freemsg(mioc);
	    CLOCKOFF(IOCTLTIME) ;
	    return(-ENOMEM);
	}
	if (do_copyin)
	{
	    if ((err = lis_copyin(f,mdta->b_wptr,ioc->ic_dp,ioc->ic_len)) < 0)
	    {
		ioc->ic_len = 0 ;		/* no data */
		lis_freemsg(mioc);
		CLOCKOFF(IOCTLTIME) ;
		return(err) ;
	    }
	    mdta->b_wptr += ioc->ic_len;
	}
	else
	if (ioc->ic_len != TRANSPARENT)
	{
	    memcpy(mdta->b_wptr,ioc->ic_dp,ioc->ic_len);
	    mdta->b_wptr += ioc->ic_len;
	}
	else
	{
	    void	**ptr_ptr = (void **) mdta->b_wptr ;
	    *ptr_ptr = ioc->ic_dp ;		/* pass in ptr */
	    mdta->b_wptr += sizeof(*ptr_ptr) ;
	}
	lis_linkb(mioc,mdta);
    }
    iocb=(iocblk_t*)mioc->b_wptr;
    mioc->b_wptr += sizeof(iocblk_t);
    mioc->b_datap->db_type = M_IOCTL ;
    iocb->ioc_cmd=ioc->ic_cmd;
    iocb->ioc_cr = creds ;
    iocb->ioc_count=ioc->ic_len;
    iocb->ioc_error=0;
    iocb->ioc_rval=0;
    iocb->ioc_filler[0] = 0 ;
    iocb->ioc_filler[1] = 0 ;
    iocb->ioc_filler[2] = 0 ;
    iocb->ioc_filler[3] = 0 ;
    /*
     * We only send one ioctl at a time downstream on a per stream basis.
     * To enforce this we hold this lock until the ioctl transaction with
     * the driver is complete, or until some error condition occurs.
     */
    if ((err = lis_lock_wioc(hd)) < 0)
    {					/* signalled probably */
	ioc->ic_len = 0 ;		/* no data */
	lis_freemsg(mioc);
	CLOCKOFF(IOCTLTIME) ;
	return(err) ;
    }

    iocb->ioc_id = hd->sd_iocseq =  lis_incr(&lis_iocseq);

  again:				/* wioc lock is always held when here */
    if (err < 0)
    {
	CP(hd,err) ;
	ioc->ic_len = 0 ;		/* no data */
	if (mioc != NULL) lis_freemsg(mioc);
	lis_unlock_wioc(hd);
	CLOCKOFF(IOCTLTIME) ;
	return(err) ;
    }

    if ( LIS_DEBUG_IOCTL )
	printk("lis_strdoioctl: send %s to \"%s\" on stream %s, cmd(0x%x)\n",
			lis_msg_type_name(mioc),
			lis_queue_name(hd->sd_wq->q_next),
			hd->sd_name, iocb->ioc_cmd) ;
    F_SET(hd->sd_flag,IOCWAIT);		/* for strrput */
    CLOCKADD() ;			/* exclude driver time */
    lis_spin_unlock(&hd->sd_lock) ;	/* unlock strm head */
    CP(hd,0) ;
    lis_putnext(hd->sd_wq,mioc);	/* do not use PUTNEXT */
    CLOCKON() ;
    CP(hd,0) ;
    mioc=NULL;			
    CLOCKADD() ;
    lis_spin_lock(&hd->sd_lock) ;	/* lock strm head again */
    CP(hd,0) ;
    err = lis_wait_for_wiocing(hd,ioc->ic_timout) ;
    if (err < 0 || (mioc = hd->sd_iocblk) == NULL)	/* rtnd ioctl msg */
    {
	CP(hd,err) ;
	F_CLR(hd->sd_flag,IOCWAIT);	/* now completed */
	ioc->ic_len = 0 ;		/* no data */
	if (hd->sd_werror)		/* pick up error from M_ERROR if pres */
	    err = hd->sd_werror ;
	else
	if (hd->sd_rerror)
	    err = hd->sd_rerror ;
	lis_unlock_wioc(hd);
	CLOCKOFF(IOCTLTIME) ;
	return(err);
    }
    /* Here we have! */
    CLOCKON() ;
    F_CLR(hd->sd_flag,IOCWAIT);		/* now completed */
    hd->sd_iocblk = NULL ;
    if ( LIS_DEBUG_IOCTL )
	printk("lis_strdoioctl: %s from \"%s\" on stream %s\n",
		    lis_msg_type_name(mioc),
		    lis_queue_name(hd->sd_wq->q_next),
		    hd->sd_name) ;

    msgtype = lis_btype(mioc) ;
    CP(hd,msgtype) ;
    switch(msgtype)
    {
    case M_IOCACK:
	iocb=(iocblk_t*)mioc->b_rptr;
	if (iocb->ioc_error){
	    ioc->ic_len = 0 ;		/* no data */
	    lis_freemsg(mioc);
	    lis_unlock_wioc(hd);
	    CLOCKOFF(IOCTLTIME) ;
	    return(-(iocb->ioc_error));
	}

	{
	    int rval=iocb->ioc_rval;
	    int len = iocb->ioc_count ;
	    mblk_t *dat=lis_unlinkb(mioc);

	    if (do_copyin)
	    {
		if (len > 0)
		{				/* data to return */
		    if ((err=lis_check_umem(f,VERIFY_WRITE,ioc->ic_dp,len)) < 0)
			rval = err ;		/* oops... */
		    else
		    {
			copyout_blks(f,ioc->ic_dp, len, dat);
			dat = NULL;		/* see copyout_blks() */
			ioc->ic_len = len ;	/* length */
		    }
		}
		else				/* no data to return */
		if (dat)			/* however, have empty bfr */
		    ioc->ic_len = 0 ;		/* so set len to zero */
	    }

	    if (dat)
		lis_freemsg(dat);

	    lis_freemsg(mioc);
	    lis_unlock_wioc(hd);
	    CLOCKOFF(IOCTLTIME) ;
	    return(rval);
	}
    case M_IOCNAK:
	iocb=(iocblk_t*)mioc->b_rptr;
	{
	    int res=-(iocb->ioc_error);
	    ioc->ic_len = 0 ;		/* no data */
	    lis_freemsg(mioc);
	    lis_unlock_wioc(hd);
	    CLOCKOFF(IOCTLTIME) ;
	    return(res ? res : -EINVAL);
	}
	break ;
    case M_COPYIN:
	{
	    copyreq_t *cp=(copyreq_t*)mioc->b_rptr;
	    copyresp_t*cr=(copyresp_t*)mioc->b_rptr;
	    u_int      len=cp->cq_size;
	    mblk_t    *dat=NULL;
	    char      *ubuf=cp->cq_addr;

	    err = 0 ;

	    dat = mioc->b_cont ;
	    if (dat != NULL)
	    {
		mioc->b_cont = NULL ;
		freemsg(dat) ;
		dat = NULL ;
	    }
	    lis_btype(mioc)=M_IOCDATA;
	    if (   (err=lis_check_umem(f,VERIFY_READ,ubuf,len)) < 0
		|| (dat=allocb(len,BPRI_LO))==NULL
		|| (err = lis_copyin(f,dat->b_wptr,ubuf,len)) < 0
	       )
	    {
		if (err < 0)
		    cr->cp_rval=(caddr_t)(-err);
		else
		    cr->cp_rval=(caddr_t)ENOMEM;
		if (dat != NULL)
		    lis_freemsg(dat);
	    }
	    else
	    {
		dat->b_wptr += len ;
		lis_linkb(mioc,dat);
		cr->cp_rval=0;
	    }

	    mioc->b_wptr = mioc->b_rptr + sizeof(*cr) ;
	    goto again;			/* keep holding wioc lock */
	}
    case M_COPYOUT:
	{
	    copyreq_t *cp=(copyreq_t*)mioc->b_rptr;
	    copyresp_t*cr=(copyresp_t*)mioc->b_rptr;
	    char      *ubuf=cp->cq_addr;
	    mblk_t    *dat;
	    u_int      len=cp->cq_size;

	    lis_btype(mioc)=M_IOCDATA;
	    if ((err=lis_check_umem(f,VERIFY_WRITE,ubuf,len)) >= 0){
		dat=lis_unlinkb(mioc);
		if ((err = copyout_blks(f,ubuf,len,dat)) < 0)
		    cr->cp_rval=(caddr_t)(-err);
		else
		    cr->cp_rval=0;
	    } 
	    mioc->b_wptr = mioc->b_rptr + sizeof(*cr) ;
	    goto again;			/* keep holding wioc lock */
	}
    default:
	ioc->ic_len = 0 ;		/* no data */
	lis_freemsg(mioc);
	lis_unlock_wioc(hd);
	CLOCKOFF(IOCTLTIME) ;
	return(-EINVAL) ;
    }

}/*lis_strdoioctl*/

/*  -------------------------------------------------------------------  */
/*
 * lis_lookup_stdata
 *
 * Look through all the stdata structures for one that has the
 * given device id.  Return a pointer to it.
 *
 * Return with the spin lock held for the stdata structure.
 */
static stdata_t *
lis_lookup_stdata( dev_t *dev, struct inode *from, stdata_t *excluded )
{
    stdata_t *hd;
    int       n;
    dev_t     odev = *dev;

    if (lis_stdata_head == NULL)
	return(NULL);			/* no stdatas in use */

    lis_spin_lock(&lis_stdata_lock) ;
    for (hd = lis_stdata_head, n = 0;
	 n == 0 || hd != lis_stdata_head;
	 hd = hd->sd_next, n++)
    {
	if (hd->magic != STDATA_MAGIC)
	{
	    printk("lis_lookup_stdata: stream head at %lx: bad magic %x\n",
			    (long)hd, hd->magic) ;
	    break ;				/* don't believe sd_next */
	}

	if (hd != excluded)
	{
	    /*
	     * As long as the stream head is allocated the following
	     * tests are valid -- we don't need the head lock.  Owning
	     * the stdata_lock prevents free_stdata from actually freeing
	     * the stream head, but if it is that far along we will notice
	     * the STRCLOSE flag and omit looking at it.
	     *
	     * We need to release the stdata_lock before acquiring the
	     * head lock.
	     */
	    if (F_ISSET(hd->sd_flag, STRCLOSE))	/* stream is closing */
		continue ;			/* don't look at this entry */
	    /*
	     *  this allows us to reopen a cloned FIFO inode
	     */
	    if (hd->sd_from && hd->sd_inode && (hd->sd_from == from)) 
	    {
		if (LIS_DEBUG_OPEN && LIS_DEBUG_ADDRS)
		    printk(
			"lookup_stdata( 0x%x, 0x%x, 0x%x ) to 0x%x dev 0x%x\n",
			(int)odev, (int)from, (int)excluded,
			(int)hd->sd_inode, (int)hd->sd_dev );
		*dev = hd->sd_dev;
		lis_spin_unlock(&lis_stdata_lock) ;	/* done with lock */
		lis_spin_lock(&hd->sd_lock) ;	/* lock strm head */
		return hd;			/* holding lock */
	    }
	    if (   hd->sd_dev == odev			/* same maj/minor */
		&& !F_ISSET(hd->sd_flag, STRCLOSE)	/* stream not closing */
	       )
	    {
		if (LIS_DEBUG_OPEN && LIS_DEBUG_ADDRS)
		    printk( "lookup_stdata( 0x%x, 0x%x, 0x%x ) => 0x%x\n",
			    (int)odev, (int)from, (int)excluded, (int)hd );
		lis_spin_unlock(&lis_stdata_lock) ;	/* done with lock */
		lis_spin_lock(&hd->sd_lock) ;	/* lock strm head */
		return(hd);			/* holding lock */
	    }
	}
    }

    lis_spin_unlock(&lis_stdata_lock) ;
    if (LIS_DEBUG_OPEN && LIS_DEBUG_ADDRS)
	printk( "lookup_stdata( 0x%x, 0x%x, 0x%x ) => NULL\n",
		(int)odev, (int)from, (int)excluded );

    return(NULL);			/* not found */
}

/*  -------------------------------------------------------------------  */
/*			Exported functions & macros                      */
/*  -------------------------------------------------------------------  */





/*  -------------------------------------------------------------------  */
/* lis_stropen - open a stream (see Magic Garden 7.9.3 and 7.9.4)
 *
 * The stream head represents a single stream.  It has the status of
 *  an inode in the operating system.  As such it can be referenced by
 *  multiple files, but there must be a single inode per stream.  This
 *  routine enforces that relationship.  Moreover, the inode corresponding
 *  to a stream will be independent of any (other) filesystem.  This
 *  makes it safe (once again) to have stdata structures and inodes keep
 *  references to each other.
 */

#ifdef KERNEL_2_1				/* linux >= 2.1.0 */
#define	PRINT_ADDRS				\
    printk("            f=%lx i=%lx d=%lx\n", \
	   (long) f, (long) i, (long) f->f_dentry ) 
#else						/* linux < 2.1.0 or portable */
#define	PRINT_ADDRS				\
    printk("            f=%lx i=%lx f->f_inode=%lx\n", \
	   (long) f, (long) i, (long) f->f_inode)
#endif

int 
lis_stropen( struct inode *i, struct file *f )
{
    struct stdata *head = NULL;
    const char	  *name = "(unknown)";
    const char	  *majmin = "(unknown)";
    cred_t	   creds;
    dev_t	   odev ;
    dev_t	   ndev ;
    int		   err = 0;
    int		   hd_locked = 0;
    int		   stdata_locked = 0 ;	/* have stdata_sem */
    int		   maj, mnr;
    unsigned long  time_cell = 0;
    unsigned long this_open ;
#if defined(LINUX) && defined(KERNEL_2_1)
    struct dentry *oldd ;
#endif

    CLOCKON() ;

    this_open = lis_incr(&lis_open_cnt);

#ifdef LINUX
    lis_start_qsched() ;		/* ensure q running process going */
#endif

    CHECK_INO(i,"stropen");		/* may return */

    /*
     *  we use lis_stdata_sem to serialize stropen(), up to the point that
     *  we know we have a head for the stream being opened.  Primarily,
     *  lis_stdata_sem protects the allocation of new streams.
     */
    if ((err = lis_down(&lis_stdata_sem)) < 0)  {
	CLOCKOFF(OPENTIME);
	return(err);
    }
    stdata_locked = 1 ;

    if (LIS_DEBUG_OPEN)
    {
	printk("stropen#%ld: "
	       "i_rdev=(%d,%d) i_count=%ld f_count=%d "
	       "f_flags=%x"
#ifdef KERNEL_2_1				/* linux >= 2.1.0 */
	       " d_count=%d"
#endif
	       " (%ld open stream(s))\n",
	       this_open,
	       STR_MAJOR(i->i_rdev), STR_MINOR(i->i_rdev),
	       I_COUNT(i), F_COUNT(f),
	       f->f_flags
#if defined(KERNEL_2_1)				/* linux >= 2.1.0 */
# if defined(KERNEL_2_3)
	       , atomic_read(&f->f_dentry->d_count)
# else
	       , f->f_dentry->d_count
# endif
#endif
	       , lis_stdata_cnt
	       ) ;

	if (LIS_DEBUG_ADDRS) PRINT_ADDRS ;

	lis_show_inode_aliases(i);
    }

    creds.cr_uid  = (uid_t) EUID(f);
    creds.cr_gid  = (gid_t) EGID(f);
    creds.cr_ruid = (uid_t) UID(f);
    creds.cr_rgid = (gid_t) GID(f);

    lis_atomic_inc(&lis_in_syscall) ;	/* processing a system call */
    odev = i->i_rdev;
    maj = STR_MAJOR(odev);
    mnr = STR_MINOR(odev);
#if defined(LINUX) && defined(KERNEL_2_1)
    oldd = lis_dget(f->f_dentry) ;	/* see comment @ end of routine */
#endif
    if (!(head = FILE_STR(f)))
    {					/* opening new stream */
	/*
	 * Always get a new stream head structure for a clone open.
	 */
	if (   !LIS_DEV_IS_CLONE(maj)
	    && (head = lis_lookup_stdata(&odev, FILE_INODE(f), NULL)) != NULL
	   )
	{
	    hd_locked = 1 ;
	    /*
	     *  ASSERT: head->sd_inode is not NULL
	     */
	    if (head->sd_inode) {
		if (i != head->sd_inode)
		    i = lis_old_inode( f, head->sd_inode );
		SET_FILE_STR(f, head);
	    } else {
		err = -EINVAL;
		goto error_rtn;
	    }
	}
    }
    if (head)			/* via file or via lookup */
    {
reuse_head:			/* changed maj/mnr from clone open */
	CP(head,0) ;
	if (head->magic != STDATA_MAGIC)
	{
	    err = -EINVAL ;
	    printk("stropen#%ld: "
	    	   "head@%lx has magic number %lx, should be %lx\n",
		   this_open, (long)head, head->magic, STDATA_MAGIC) ;
	    head = NULL ;		/* pretend no strm head */
	    goto error_rtn ;
	}

	/* Stream was created by previous open of device: sleep if still
           opening, then call every module open() and awake sleepers */

	if (!hd_locked)
	    lis_spin_lock(&head->sd_lock) ;   /* lock strm head */
	lis_atomic_inc(&head->sd_refcnt);     /* another use counted */
	head->sd_file = f;     		      /* in case a mod needs to know */
	lis_spin_unlock(&head->sd_lock) ;     /* unlock while call drvr */
	hd_locked = 0 ;

	if ((err = open_mods( head, &odev, f->f_flags, &creds )) < 0)
	{
	    printk("stropen#%ld: error %d from open_mods()\n",
		   this_open, err );
	    goto error_rtn ;
	}

	lis_spin_lock(&head->sd_lock) ;	/* lock strm head */
	hd_locked = 1 ;
	head->sd_file = NULL;  /* done with this now */

	/*
	 *  'connld' (or the like) may have completely replaced
	 *  the stream - check for that
	 */
	if (i != FILE_INODE(f))
	{
	    i = FILE_INODE(f);
	    head = FILE_STR(f);
	}

	goto successful_rtn ;
    }

    /* First open of the stream: allocate stream head (and queue pair) */

    if (!(head = alloc_stdata())){		/* allocate and lock */
	if (LIS_DEBUG_OPEN)
	    printk("stropen#%ld: failed to allocate stream head\n",this_open);
	err = -ENOSR ;
	goto error_rtn ;
    }

    /*
     *  alloc_stdata() returns with the stdata structure locked
     */
    CP(head,0) ;
    head->sd_rdopt = RPROTNORM | RNORM ;
    head->sd_wropt = 0 ;
    head->sd_maxpsz   = LIS_MAXPSZ;
    head->sd_minpsz   = LIS_MINPSZ;
    head->sd_closetime	= LIS_CLTIME;

    SET_FILE_STR(f, head);			/* point file to strm hd */
    hd_locked = 1 ;
    lis_atomic_set(&head->sd_refcnt, 1) ;

    i = lis_set_up_inode(f, i) ;		/* allocate an LiS inode */
    if (i == NULL)
    {
	err = -ENOMEM ;				/* couldn't allocate struct */
	goto error_rtn ;
    }

    SET_INODE_STR(i, head);			/* point new inode to strm hd */
    head->sd_strtab= LIS_DEVST(maj).f_str;	/* use strmtab of opened strm */

    head->sd_dev = odev ;
    ndev = odev ;
    head->sd_open_flags = f->f_flags;

    if (LIS_DEV_IS_FIFO(maj) ||
	(LIS_DEV_IS_CLONE(maj) && LIS_DEV_IS_FIFO(mnr)))
    {
	CP(head,0) ;
	if (LIS_DEV_IS_CLONE(maj))
	    lis_setq( head->sd_rq,
		      clone_info.st_rdinit, clone_info.st_wrinit );
	else
	    lis_setq( head->sd_rq,
		      fifo_info.st_rdinit, fifo_info.st_wrinit );

	name = lis_strm_name(head) ;
	majmin = lis_maj_min_name(head) ;
	strncpy(head->sd_name, name, sizeof(head->sd_name)-1) ;
	strncat(head->sd_name, majmin, sizeof(head->sd_name)-1) ;
	lis_mark_mem(head, head->sd_name, MEM_STRMHD) ;
	lis_new_file_name(f, head->sd_name) ;

	head->sd_peer = head;
	F_SET( head->sd_flag, STFIFO );
	F_SET( head->sd_wropt, SNDZERO|SNDPIPE );

	if (LIS_DEBUG_OPEN)
	    printk( "stropen#%ld: FIFO from 0x%x \"%s\"\n",
		    this_open, (int)FILE_INODE(f), FILE_NAME(f) );

	/*
	 * Let go of the stdata lock while we call the user's open routine.
	 * Then reacquire it to finish the open.
	 */
	if (stdata_locked)
	{
	    lis_up(&lis_stdata_sem);
	    stdata_locked = 0 ;
	}

	if (hd_locked)				/* head is locked */
	    lis_spin_unlock(&head->sd_lock) ;	/* unlock strm head */
	lis_atomic_dec(&lis_in_syscall) ;	/* "done" with system call */
	lis_runqueues() ;
	err = lis_qopen( head->sd_rq, &ndev, f->f_flags, &creds ); 
	CP(head,0) ;
	lis_spin_lock(&head->sd_lock) ;
	CP(head,0) ;
	hd_locked = 1 ;
	if (err == 0)
	{
	    lis_setq(head->sd_rq, fifo_info.st_rdinit, fifo_info.st_wrinit);
	    stdata_locked = (err = lis_down(&lis_stdata_sem)) == 0 ;
	}
	lis_atomic_inc(&lis_in_syscall) ;	/* processing a system call */
    }
    else
    {
	CP(head,0) ;
	lis_setq( head->sd_rq, &strmhd_rdinit, &strmhd_wrinit );

	name = lis_strm_name(head) ;
	majmin = lis_maj_min_name(head) ;
	strncpy(head->sd_name, name, sizeof(head->sd_name)-1) ;
	strncat(head->sd_name, majmin, sizeof(head->sd_name)-1) ;
	lis_mark_mem(head, head->sd_name, MEM_STRMHD) ;
	lis_new_file_name(f, head->sd_name) ;

	/*
	 * We are about to call the driver's open routine.  If that routine
	 * sleeps we could get another open on this same device.  The
	 * second open should lead to this same stream head and a second
	 * call on the driver's open routine.  But it can't do that unless
	 * the sd_inode field is set in stdata.
	 */
	if (!head->sd_inode)
	{
	    CP(head,0) ;
	    head->sd_inode = i ;
	}

	/*
	 * Let go of the stdata lock while we call the user's open routine.
	 * Then reacquire it to finish the open.
	 */
	if (stdata_locked)
	{
	    lis_up(&lis_stdata_sem);
	    stdata_locked = 0 ;
	}

	if (hd_locked)				/* head is locked */
	{
	    lis_spin_unlock(&head->sd_lock) ;	/* unlock strm head */
	    hd_locked = 0 ;
	}

	lis_atomic_dec(&lis_in_syscall) ;	/* "done" with a system call */
	lis_runqueues() ;
	err = lis_qattach( head, head->sd_strtab, &ndev, f->f_flags, &creds );
	if (err == 0)
	    stdata_locked = (err = lis_down(&lis_stdata_sem)) == 0 ;
	lis_atomic_inc(&lis_in_syscall) ;	/* processing a system call */
    }

    if (err < 0)				/* failed qopen */
	goto error_rtn ;			/* go cleanup */

    if (ndev != odev)				/* clone open */
    {						/* changed the device */
	stdata_t	*other_hd ;
	const char	*clone_name = name ;
	const char	*clone_majmin = majmin ;

	if (hd_locked)
	{
	    lis_spin_unlock(&head->sd_lock) ;	/* unlock strm head */
	    hd_locked = 0 ;
	}

	other_hd = lis_lookup_stdata( &ndev, FILE_INODE(f), head );
						/* chk for other strm */
	if (other_hd != NULL)			/* have stdata struct already */
	{					/* use it instead of head */
	    /*
	     * lookup_stdata returns holding the spin lock for the other_hd
	     * stdata structure.
	     */
	    CP(other_hd,0) ;
	    if ( LIS_DEBUG_OPEN )
		printk("stropen#%ld: clone open stream to '%s' "
			"reusing existing stream\n",
			this_open, other_hd->sd_name) ;
	    
	    if (hd_locked)			/* head is locked */
		lis_spin_unlock(&head->sd_lock) ; /* unlock strm head */

	    if (stdata_locked)			/* let go for close */
	    {
		lis_up(&lis_stdata_sem);
		stdata_locked = 0 ;
	    }
	    lis_atomic_dec(&lis_in_syscall) ;	/* "done" with a system call */

	    lis_runqueues() ;			/* no locks owned */

	    if (lis_doclose(i, f, head, &creds))/* close "new" stream */
	    {					/* not last close */
		if (hd_locked)			/* head is locked */
		    lis_spin_unlock(&head->sd_lock) ; /* unlock strm head */
	    }

	    head = other_hd ;			/* use "old" stream */
	    hd_locked = 1 ;			/* from lookup_stdata */
	    if (err == 0)
		stdata_locked = (err = lis_down(&lis_stdata_sem)) == 0 ;
	    lis_atomic_inc(&lis_in_syscall) ;	/* processing a system call */
	    odev = ndev ;			/* switch to new device */
	    if (err) goto error_rtn ;
	    goto reuse_head ;
	}

	CP(head,0) ;
	if (!hd_locked)
	{
	    lis_spin_lock(&head->sd_lock) ;	/* ensure head locked */
	    hd_locked = 1 ;
	}

	head->sd_dev = ndev ;
	maj  = STR_MAJOR(ndev);
	mnr  = STR_MINOR(ndev);
	head->sd_strtab = LIS_DEVST(STR_MAJOR(ndev)).f_str;
	head->sd_name[0] = 0 ;			/* recompute stream name */
	name = lis_strm_name(head) ;		/* update names */
	majmin = lis_maj_min_name(head) ;
	strncpy(head->sd_name, name, sizeof(head->sd_name)-1) ;
	strncat(head->sd_name, majmin, sizeof(head->sd_name)-1) ;
	lis_mark_mem(head, head->sd_name, MEM_STRMHD) ;
	lis_new_file_name(f, head->sd_name) ;
						/* allocate new inode */
	if (!lis_new_inode( f, ndev )) {	/* yet another inode */
	    err = -ENOMEM ;
	    goto error_rtn ;
	}
	head->sd_inode = i = FILE_INODE(f);
	SET_INODE_STR(i, head);

	if (LIS_ISDEV(maj,LIS_FIFO)) {
	    if (LIS_DEBUG_OPEN && LIS_DEBUG_ADDRS)
		printk( "stropen#%ld: FIFO to 0x%x \"%s\"\n",
			this_open, (int)head->sd_inode, FILE_NAME(f) );
	}

	if ( LIS_DEBUG_OPEN )
	{
	    printk("stropen#%ld: open clone '%s%s' to '%s'\n",
		    this_open, clone_name, clone_majmin,
		    head->sd_name
		    ) ;
	    if (LIS_DEBUG_ADDRS)
#ifdef KERNEL_2_1				/* linux >= 2.1.0 */
	      printk(
		"stropen#%ld: clone: f=%lx i=%lx i->i_rdev=%x sd_dev=%x\n",
		this_open, (long) f, (long) i, i->i_rdev, head->sd_dev) ;
#else						/* linux < 2.1.0 or portable */
	      printk(
		"stropen#%ld: clone: f=%lx f->f_inode=%lx i=%lx i->i_rdev=%x "
		"sd_dev=%x\n",
		this_open, (long) f, (long) f->f_inode, (long) i, i->i_rdev,
		       head->sd_dev) ;
#endif
	}
    }
    else
    {			/* ndev == odev non-clone */
	/*
	 *  we always want a new inode for a stream; if head->sd_inode
	 *  isn't set yet, we may have a non-STREAMS inode
	 *
	 *  the reason we always want a new inode for a stream is so
	 *  that non-STREAMS file systems won't muck about with them;
	 *  i.e., we want streams inodes to be independent of other
	 *  filesystems
	 *
	 *  coincidentally (for now), this also provides a means to
	 *  identify a stream; the i_dev major in an inode for a
	 *  stream will be lis_major, and the i_dev minor will be the
	 *  major of the stream's driver.
	 */
	CP(head,0) ;
	if (!head->sd_inode)
	{
	    CP(head,0) ;
	    head->sd_inode = i ;
	}
    }

    { /* Do autopush */
	const char *mods[MAXAPUSH];
	int j, cnt;

	cnt = lis_apushm(i->i_rdev, mods);
	for (j = 0; j < cnt; ++j) {
	    modID_t mod = lis_findmod(mods[j]);
	    streamtab_t *st;

	    /* printk("autopushing %s\n", mods[j]); */
	    if (mod == LIS_NULL_MID) {
		/*
		 *  Module not present. Probably one of:
		 *  a) Module open for a previous autopush slept, giving
		 *     kerneld(8) or root time to unload our loadable module.
		 *  b) A configured loadable module could not be loaded.
		 */
		err = -ENOSR; /* Assume (a), retry may succeed */
		goto error_rtn;
	    }

	    st = lis_fmod_sw[mod].f_str;
	    ASSERT(st != NULL);
	    
	    if (hd_locked)				/* head is locked */
		lis_spin_unlock(&head->sd_lock) ;	/* unlock strm head */
	    CP(head,st) ;
	    err = push_mod( head, st, &ndev, f->f_flags, &creds ) ;
	    if (err < 0)
	        goto error_rtn; /* push failure */
	    lis_spin_lock(&head->sd_lock) ;
	    hd_locked = 1 ;
	}
    }

successful_rtn:					/* returning success */

    lis_task_to_creds(&head->sd_creds) ;	/* save kernel creds */
    if (stdata_locked)
	lis_up(&lis_stdata_sem);

    err = 0;

#ifdef S_ZERO_WR				/* Linux >= 2.0.36 */
    i->i_flags |= S_ZERO_WR ;
#endif

    if ( LIS_DEBUG_OPEN )
    {
	lis_show_inode_aliases(i);
	printk("stropen#%ld: "
	       "open #%d on stream to '%s' (%ld open stream(s))\n",
		this_open, LIS_REFCNT(head), head->sd_name, lis_stdata_cnt) ;
	printk("            head %lx wq:\"%s\" next wq:\"%s\"\n",
	       (long)head, lis_queue_name(head->sd_wq),
	       (SAMESTR(head->sd_wq)?
		lis_queue_name(head->sd_wq->q_next):"<none>"));
	if (LIS_DEBUG_ADDRS) PRINT_ADDRS ;
    }

    if (hd_locked)
	lis_spin_unlock(&head->sd_lock) ;	/* unlock strm head */

    /*
     *  FIFO files:
     *
     *  we "grab" the current inode and keep it as sd_from, so
     *  we can be sure it's valid if we need to check it
     *  (we don't keep 'sd_from' if cloning from a clone major)
     */
#if 0
    if (LIS_DEV_IS_FIFO(maj))
    {
	head->sd_from = lis_grab_inode(FILE_INODE(f));
	if (LIS_DEBUG_OPEN)
	    printk( "stropen#%ld: holding 0x%x(%d) as sd_from\n",
		    this_open,
		    (int)head->sd_from, I_COUNT(head->sd_from));
    }
#endif

#if defined(LINUX)
    /*
     *  synchronize FIFO read/write openers after (almost) everything else 
     *
     *  this is really only needed for FIFOs being opened O_RDONLY
     *  or O_WRONLY; the only likely failure case is O_WRONLY|O_NONBLOCK
     *  when no reader is open, which will return -ENXIO.
     *
     *  note that this is also done for pipe ends but really isn't needed,
     *  since pipe ends are always opened O_RDWR; unfortunately, we
     *  can't tell at this point whether a FIFO being opened will become
     *  a pipe end or not (FIXME).
     */
    if (F_ISSET(head->sd_flag,STFIFO))
	if ((err = lis_fifo_open_sync(i,f)) < 0)
	    goto error_rtn;
# if defined(KERNEL_2_1)
    /*
     * For the 2.2 and 2.4 kernel we have done an extra dget on the original
     * dentry of the file.  This was balanced by a dput in the routine
     * lis_new_inode.  Since we have replace that dentry with a new one
     * we now need to do the dput that reduces the count to zero on the
     * original dentry.
     *
     * Also see comment at error_rtn.
     */
    if (f->f_dentry != oldd)		/* yes, a new dentry */
	lis_dput(oldd);
# endif
#endif

    CP(head,err) ;
    CLOCKOFF(OPENTIME) ;
    lis_atomic_dec(&lis_in_syscall) ;	/* done processing a system call */
    lis_runqueues();
    return(err);

error_rtn:				/* come here if fail */

    if (stdata_locked)
	lis_up(&lis_stdata_sem);

    if ( head && LIS_DEBUG_OPEN )  {
	name = lis_strm_name(head);
	majmin = lis_maj_min_name(head);
    }

    if ( LIS_DEBUG_OPEN )
    {
	printk(
	    "stropen#%ld: "
	    "open stream to \"%s\" returning error %d (%ld open stream(s))\n",
	    this_open, name, err, lis_stdata_cnt) ;
	if (LIS_DEBUG_ADDRS) PRINT_ADDRS ;
    }

    if (head)			        /* a head was used */
    {
	lis_clean_up_file(f, head, err) ;
	lis_atomic_dec(&head->sd_refcnt) ;
	if (LIS_REFCNT(head) <= 0)	/* last use of head */
        {
	    if (head->sd_rq)  lis_freeq(head->sd_rq);
	    head->sd_wq = NULL;
	    head->sd_rq = NULL;

#ifdef S_ZERO_WR				/* Linux 2.0.36 */
	    i->i_flags &= ~S_ZERO_WR ;
#endif
	    lis_free_stdata(head) ;	/* free our structure */

	    SET_FILE_STR(f, NULL);		/* unhook from file */
	    f->f_op = NULL;
        }
	else
	if (hd_locked)			/* head is locked */
	    lis_spin_unlock(&head->sd_lock) ; /* unlock strm head */
    }

#if defined(LINUX) && defined(KERNEL_2_1)
    /*
     * The 2.2 and 2.4 kernel has saved the original dentry from the file and
     * will "dput" it when the open fails.  We have done an extra
     * "dget" on it balanced by a "dput" of our own in the lis_new_inode
     * routine.  What we now need to do is to dput the new dentry that
     * we allocated since the kernel won't do it for us.
     *
     * Also see comment at successful return.
     */
    if (f->f_dentry != oldd)		/* yes, a new dentry */
	lis_dput(f->f_dentry);
#endif

    CP(head,err) ;
    CLOCKOFF(OPENTIME) ;
    lis_atomic_dec(&lis_in_syscall) ;	/* done processing a system call */
    lis_runqueues();
    return(err) ;

#undef	PRINT_ADDRS

}/*lis_stropen*/

/*  -------------------------------------------------------------------  */
/* strwrite - write to a stream (see Magic Garden 7.9.5)
 */
#if defined(KERNEL_2_1)
ssize_t
lis_strwrite(struct file *fp, const char *ubuff, size_t ulen, loff_t *op)
#else
int
lis_strwrite(struct inode *i, struct file *fp, const char *ubuff,  int ulen)
#endif
{
    stdata_t *hd;
    mblk_t   *held;
    int		 chunk;
    int		 written=0; 
    int		 flags ;
    int		 err,newmsg;
    int		 hd_locked = 0 ;
    int		 inode_locked = 0 ;
    unsigned long  time_cell = 0 ;
#if defined(KERNEL_2_1)
    struct inode  *i = FILE_INODE(fp);
#endif

#define RTN(e)  do { err=(e); goto return_point; } while (0)

    /* Just in case we're holding messages in the write side we use
       head->sd_wmsg msg to hold data to be sent. When it's time to send it we
       just lis_putnext() such message.  If we're not holding then we just
       lis_putnext() the message w/ the usr data.       */

    CHECK_INO(i,"lis_strwrite"); /* may return */

    hd = FILE_STR(fp) ;
    if (hd == NULL)
	return(-ENODEV);

    CLOCKON() ;
    lis_spin_lock(&hd->sd_lock) ;		/* lock strm head */
    hd_locked = 1 ;
    lis_atomic_inc(&lis_in_syscall) ;		/* processing a system call */

    if (ulen < hd->sd_minpsz)
    {
	if (LIS_DEBUG_WRITE)
	  printk(
	    "strwrite: stream %s: write length %d less than min size of %ld\n",
			hd->sd_name,
			ulen, hd->sd_minpsz) ;

	RTN(-EINVAL);
    }

    if (!ulen && !F_ISSET(hd->sd_wropt,SNDZERO)) /* we don't send anything */
    {
	if (LIS_DEBUG_WRITE)
	  printk(
	    "strwrite: stream %s: write length zero and SNDZERO not set\n",
			hd->sd_name) ;

	RTN(0);
    }

#if defined(LINUX) && defined(KERNEL_2_1)
    if (F_ISSET(hd->sd_flag,STFIFO)) {
	if ((err = lis_fifo_write_sync( i, 0 )) < 0) {
	    RTN(err);
	}
    }
#endif

#if defined(LINUX) && defined(KERNEL_2_3)
    LOCK_INO(i);            /* sys_write no longer does this */
    inode_locked = 1 ;
#endif

    if ((err=lis_check_umem(fp,VERIFY_READ,ubuff,ulen))<0)
    {
	if (LIS_DEBUG_WRITE)
	  printk(
	    "strwrite: stream %s: user address 0x%lx cannot be read from\n",
			hd->sd_name, (long)ubuff) ;

	RTN(err);
    }

    if (F_ISSET(hd->sd_flag,STPLEX))
    {
	if (LIS_DEBUG_WRITE)
	  printk("strwrite: stream %s: stream has been I_LINKed\n",
		    hd->sd_name) ;
	RTN(-EINVAL) ;
    }

    do {
	if (F_ISSET(hd->sd_flag,STWRERR))
	{
	    if (LIS_DEBUG_WRITE)
	      printk(
		"strwrite: stream %s: stream has received M_ERROR %d\n",
			hd->sd_name, hd->sd_werror) ;
	    RTN(-hd->sd_werror);
	}

	if (F_ISSET(hd->sd_flag,STRHUP|STRCLOSE))
	{
	    if (LIS_DEBUG_WRITE)
	      printk("strwrite: stream %s: stream has received M_HANGUP\n",
		      hd->sd_name) ;
	    RTN(-ENODEV);
	}

	/* write packets until all data has been written. */
	if (   lis_atomic_read(&hd->sd_wrcnt) != 0
	    || !lis_bcanputnext(hd->sd_wq,0))
	{
	    if (should_notblock_wr(hd,fp))
	    {
		if (LIS_DEBUG_WRITE)
		    printk(
		     "strwrite: stream %s: flw cntrl blocked, return EAGAIN\n",
			    hd->sd_name) ;
		RTN(-EAGAIN);
	    }

	    if (LIS_DEBUG_WRITE)
		printk("strwrite: stream %s: flw cntrl blocked, sleep\n",
			    hd->sd_name) ;
	    ULOCK_INO(i) ;		/* sys_write/user_write locked inode */
	    CLOCKADD() ;
	    err = lis_wait_on_wwrite(hd);	/* unlocks & relocks head */
	    CLOCKON() ;
	    if (err < 0)
		RTN(err) ;
	    LOCK_INO(i) ;		/* re-acquire inode lock */
	    continue ;		/* go back and retry errs and canput */
	}
	/* get mblk if there's not one */
	held=hd->sd_wmsg;
	chunk=lis_min((int)hd->sd_maxpsz,ulen-written);
	newmsg = (held==NULL);
	if ( held == NULL && 
	     (held=hd->sd_wmsg=allocb(chunk+hd->sd_wroff,BPRI_LO))==NULL)
	    RTN(written?written:-ENOMEM);

	/* now copy from usr to mblk (type M_DATA by default) */
	if (newmsg)
	{
	    held->b_rptr =
	    held->b_wptr = held->b_datap->db_base + hd->sd_wroff;
	    if (F_ISSET(hd->sd_flag,STRDELIM))		/* delimited msg */
		held->b_flag |= MSGDELIM ;
	}

	if (held->b_datap->db_lim - held->b_wptr >= chunk)    /* enough room */
	{
	    if ((err = lis_copyin(fp,held->b_wptr,ubuff,chunk)) < 0)
		RTN(err) ;

	    ubuff        += chunk;
	    written      += chunk;
	    held->b_wptr += chunk;

	    if (strholding(hd) && 
		held->b_datap->db_lim - held->b_wptr > written)
	    {					/* hold for more data */
		if (LIS_DEBUG_WRITE)
		    printk("strwrite: stream %s: "
			   "hold message %d bytes for scan timer\n",
			    hd->sd_name, chunk) ;

		if (newmsg) 
		    start_scanl_timer(hd); 

		RTN(written);
	    }
	}        

        /*
         * Either there was not enough room in the existing message
         * to hold the next write, or we wrote to a new message
         * and the write-holding mechanism is not active.
         *
         * In either case, now is the time to send the message downstream.
         */
        lis_spin_lock_irqsave(&lis_qhead_lock, &flags) ;
        if (hd->sd_scantimer != 0)		/* scantimer running */
        {
	    untimeout(hd->sd_scantimer) ;	/* cancel the timer */
	    hd->sd_scantimer = 0 ;		/* note timer is not running */
        }
	lis_spin_unlock_irqrestore(&lis_qhead_lock, &flags) ;

	if (LIS_DEBUG_WRITE)
	    printk(
		"strwrite: stream %s: send %d bytes downstream to %s\n",
		hd->sd_name, chunk, lis_queue_name(hd->sd_wq->q_next)) ;

	CLOCKADD() ;				/* exclude driver time */
	lis_spin_unlock(&hd->sd_lock) ;		/* unlock strm head */
	hd_locked = 0 ;
        PUTNEXT(hd->sd_wq,hd->sd_wmsg);		/* put & maybe run queues */
	CLOCKON() ;
	lis_spin_lock(&hd->sd_lock) ;		/* lock strm head */
	hd_locked = 1 ;
        hd->sd_wmsg = NULL ;			/* don't reuse the message */

    } while (written<ulen);

    RTN(written);

return_point:
    if (inode_locked)
	ULOCK_INO(i);

    if (hd_locked)
	lis_spin_unlock(&hd->sd_lock) ;		/* unlock strm head */

    CLOCKOFF(WRITETIME) ;
    lis_atomic_dec(&lis_in_syscall) ;	/* done processing a system call */
    lis_runqueues();
    return(err) ;

#undef RTN

}/*lis_strwrite*/

/*  -------------------------------------------------------------------  */
/* lis_check_m_sig
 *
 * Check the stream head read queue to see if there is an M_SIG at the
 * head of the queue.  If there is, remove it and signal the user.
 *
 * Caller is assumed to own the sd_lock when this routine is called.
 */
static void lis_check_m_sig(stdata_t *hd)
{
    mblk_t	*hdmp ;

    hdmp  = hd->sd_rq->q_first ;
    if (hdmp == NULL || lis_btype(hdmp) != M_SIG)
	return ;

    hdmp  = getq(hd->sd_rq) ;
    LisDownCount(MSGQDSTRHD) ;			/* one fewer msg queued */
    signalusr(*hdmp->b_rptr,hd);
    freemsg(hdmp) ;
    if (   POLL_WAITING(hd)
	|| F_ISSET(hd->sd_flag,STRSELPND)
       )
	lis_wake_up_poll(hd,POLLMSG);
}

/*  -------------------------------------------------------------------  */
/* lis_requeue - requeue a message in stream head queue
 *
 * We need to pay attention to intr disables.  If we are re-inserting
 * a high priority message and an interrupt has meanwhile inserted
 * a high priority message at the head of the queue, then we should
 * delete the new message and insert our old one.  As if the new
 * message arrived at the stream head with ours still in place.
 */
static void lis_requeue(stdata_t *hd, mblk_t *mp)
{
    mblk_t	*hdmp ;
    queue_t	*hd_rq;

    hd_rq = hd->sd_rq;
    /*
     * If we are about to insert a PCPROTO into the queue we need
     * to check to see if the current head message is also a
     * PCPROTO, and delete it if it is.
     *
     * Setting the STRPRI bit keeps strrput from adding any more
     * PCPROTOs to the queue.
     */
    if (lis_btype(mp) == M_PCPROTO)
    {
	F_SET(hd->sd_flag,STRPRI);
	hdmp  = hd_rq->q_first ;
	if (hdmp != NULL && lis_btype(hdmp) == M_PCPROTO)
	{				/* delete msg at head of q */
	    lis_freemsg(lis_getq(hd_rq)) ;
	    LisDownCount(MSGQDSTRHD) ;	/* one fewer msg queued */
	}
    }

    lis_putbq(hd_rq,mp);
    LisUpCount(MSGQDSTRHD) ;		/* one more msg queued */

} /* lis_requeue */

/*  -------------------------------------------------------------------  */
/* lis_strread - read from a stream head
 */
#if defined(KERNEL_2_1)
ssize_t 
lis_strread(struct file *fp, char *ubuff, size_t ulen, loff_t *op)
#else
int 
lis_strread(struct inode *i, struct file *fp, char *ubuff,  int ulen)
#endif
{
    stdata_t *hd;
    queue_t  *hd_rq = NULL ;
    mblk_t   *mp;
    int		 count=0;
    int		 qlocked = 0 ;
    unsigned int chunk;
    int		 nbytes ;
    int		 msg_lgth ;
    int		 msg_marked ;
    int		 msgs_read = 0 ;
    int		 mread_sent ;
    int		 msg_is_pc_proto = 0 ;
    int		 hd_locked = 0 ;
    int err;
    unsigned long  time_cell = 0 ;
#ifdef KERNEL_2_1
    struct inode *i = FILE_INODE(fp);
#endif

#define	RTN(val)  do { err = val; goto return_point; } while (0)

    CHECK_INO(i,"lis_strrread");


    hd = FILE_STR(fp) ;
    if (hd == NULL)
	return(-ENODEV);

    CLOCKON() ;
    LOCK_INO(i);
    lis_spin_lock(&hd->sd_lock) ;		/* lock strm head */
    hd_locked = 1 ;
    lis_atomic_inc(&lis_in_syscall) ;		/* processing a system call */

    if ((err=lis_check_umem(fp,VERIFY_WRITE,ubuff,ulen))<0)
    {
	if (LIS_DEBUG_READ)
	  printk(
	    "strread: stream %s: user buffer at 0x%lx cannot be written into\n",
		hd->sd_name, (long)ubuff) ;

	RTN(err);
    }

    if (F_ISSET(hd->sd_flag,STPLEX))
    {
	if (LIS_DEBUG_READ)
	  printk("strread: stream %s: stream has been I_LINKed\n", hd->sd_name);

	RTN(-EINVAL) ;
    }

    hd_rq=hd->sd_rq;
    if (!LIS_CHECK_Q_MAGIC(hd_rq)) RTN(-EINVAL) ;


    lis_lockq(hd_rq) ;	/* exclude strrsrv */
    qlocked = 1 ;		/* return flag so as to unlock */
    do {
	mread_sent = 0 ;
	while (hd_rq->q_first == NULL && msgs_read == 0)
	{					/* nothing to read */
	    if (F_ISSET(hd->sd_flag,STRDERR))
	    {
		if (LIS_DEBUG_READ)
		  printk(
		    "strread: stream %s: stream has received M_ERROR %d\n",
			hd->sd_name, hd->sd_rerror) ;
		RTN(-hd->sd_rerror);
	    }
	    
	    if (F_ISSET(hd->sd_flag,STRHUP|STRCLOSE))
	    {
		if (LIS_DEBUG_READ)
		  printk("strread: stream %s: stream has received M_HANGUP\n",
			 hd->sd_name) ;

		RTN(0);
	    }

	    if (!mread_sent && F_ISSET(hd->sd_flag,SNDMREAD))
	    {				/* only send once per user call */
		lis_unlockq(hd_rq) ;
		qlocked = 0 ;
		lis_spin_unlock(&hd->sd_lock) ;		/* unlock strm head */
		hd_locked = 0 ;
		if (LIS_DEBUG_READ)
		    printk(
			"strread: stream %s: send M_READ downstream to %s\n",
			hd->sd_name, lis_queue_name(hd->sd_wq->q_next)) ;

		err = snd_mread(hd->sd_wq,ulen);
		if (err < 0)		/* probably ENOMEM */
		    RTN(err) ;

		mread_sent = 1 ;
		lis_spin_lock(&hd->sd_lock) ;		/* lock strm head */
		hd_locked = 1 ;
		lis_lockq(hd_rq) ;
		qlocked = 1 ;
		continue ;		/* go around again */
	    }

	    if (should_notblock_rd(hd,fp))
	    {
		if (LIS_DEBUG_READ)
		    printk(
			"strread: stream %s: O_NONBLOCK set, return EAGAIN\n",
			hd->sd_name) ;

		RTN(-EAGAIN);
	    }

	    if (LIS_DEBUG_READ)
		printk("strread: stream %s: O_NONBLOCK not set, sleep\n",
			hd->sd_name) ;
	    ULOCK_INO(i);
	    CLOCKADD() ;
	    err = lis_sleep_on_wread(hd);	/* ulocks & relocks q */
	    CLOCKON() ;
	    LOCK_INO(i);
	    if (err < 0)
		RTN(err) ;
	}

	mp = lis_getq(hd_rq);			/* while q still locked */
	lis_check_m_sig(hd) ;			/* watch for M_SIG in q */
	lis_unlockq(hd_rq) ;
	qlocked = 0 ;
	if (mp == NULL)				/* out of data to read */
	{
	    RTN(count);				/* rtn amount read */
	}

	LisDownCount(MSGQDSTRHD) ;
	msg_marked = mp->b_flag & MSGMARK;
	switch(lis_btype(mp))
	{
	case M_DATA:
	    msg_lgth = 0 ;
	    err = 0 ;
	    msgs_read++ ;
	    do
	    {
		/*
		 * If rptr > wptr the message block is ill-formed
		 * and nbytes will be negative.
		 */
		if (lis_btype(mp) != M_DATA)
		    nbytes = -1 ;		/* pretend invalid */
		else
		    nbytes = mp->b_wptr - mp->b_rptr;

		if (nbytes > 0)			/* safety check */
		{
		    chunk=lis_min(ulen-count,nbytes);
		    if ((err = lis_copyout(fp,mp->b_rptr,ubuff,chunk)) < 0)
			break ;			/* stop reading */

		    msg_lgth   += nbytes ;	/* keep track of msg lgth */
		    mp->b_rptr += chunk;	/* message read ptr */
		    ubuff      += chunk ;	/* user buffer ptr */
		    count      += chunk;	/* bytes copied to user */
		    nbytes     -= chunk ;	/* bytes left in msg */
		}

		if (nbytes <= 0)		/* time for nxt blk */
		{				/* includes bad blocks */
		    mblk_t *dust=mp;		/* and zero-lgth msgs */
		    mp=lis_unlinkb(mp);
		    lis_freeb(dust);
		}

	    } while(count < ulen && mp);

	    if (mp != NULL)			/* not at msg boundary */
	    {
		if ( err == 0 && (hd->sd_rdopt & RMODEMASK) == RMSGD )
		    lis_freemsg(mp);		/* discard rest of msg */
		else				/* RMSGN or RNORM */
		    lis_requeue(hd, mp) ;	/* save rest of msg */
	    }
	    if (LIS_DEBUG_READ)
	      printk(
	      "strread: stream %s: RDOPT=0x%x user buffer %d bytes, "
	      "read %d error %d\n",
		   hd->sd_name, hd->sd_rdopt, ulen, count, err) ;
	    if (   ulen == count		/* filled user buffer */
		|| (hd->sd_rdopt & RMODEMASK) != RNORM
		|| msg_lgth == 0		/* 0-lgth msg */
		|| msg_marked			/* at marked msg */
	       )				/* return to user */
	    {
	        RTN(count);
	    }

	    if (err < 0)			/* something went wrong */
		RTN(err) ;

	    break;
	case M_PCPROTO:
		F_CLR(hd->sd_flag,STRPRI);
		msg_is_pc_proto = 1 ;
		/* fall into next case */
	case M_PROTO: 
	    msgs_read++ ;
	    if (LIS_DEBUG_READ)
		printk(
		    "strread: stream %s: RDOPT=0x%x",
		    hd->sd_name, hd->sd_rdopt) ;
	    if (F_ISSET(hd->sd_rdopt,RPROTNORM))
	    {					/* return with EBADMSG */
		if (F_ISSET(hd->sd_rdopt,RPROTDIS))/* both bits set */
		{
		    if (LIS_DEBUG_READ)
			printk(" discard M_PROTO msg, return EBADMSG\n") ;
		    lis_freemsg(mp);		/* discard the messasge */
		}
		else
		{
		    if (LIS_DEBUG_READ)
			printk(" re-queue M_PROTO msg, return EBADMSG\n") ;
		    lis_requeue(hd, mp) ;	/* put back in queue */
		}
		RTN(-EBADMSG);			/* return */
	    }

	    if (F_ISSET(hd->sd_rdopt,RPROTDAT))/* treat M_PROTO as M_DATA */
	    {
		if (LIS_DEBUG_READ)
		    printk(" change M_PROTO msg to M_DATA\n") ;
		lis_btype(mp) = M_DATA;		/* change to M_DATA */
		lis_requeue(hd, mp) ;		/* put back in queue */
		break ;				/* go around again */
	    }

	    if (F_ISSET(hd->sd_rdopt,RPROTDIS))/* discard M_PROTO */
	    {					/* deliver M_DATA */
		mblk_t	*proto = mp;		/* the M_PROTO */
		if (LIS_DEBUG_READ)
		    printk(" discard M_PROTO block, retain M_DATA\n") ;
		mp = lis_unlinkb(mp);		/* mp is M_DATA or NULL */
		lis_freeb(proto);		/* discard proto */
		if (mp != NULL)			/* still have a msg */
		    lis_requeue(hd, mp) ;	/* put back in queue */
	    }					/* go around again */
	    else
	    {
		if (LIS_DEBUG_READ)
		    printk(" invalid rdopt, discard M_PROTO msg\n") ;

		lis_freemsg(mp);		/* discard the messasge */
		RTN(-EBADMSG);			/* return */
	    }

	    break;

	case M_READ:		/* add to-be-discarded msgs here */
	    if (LIS_DEBUG_READ)
		printk(
		    "strread: stream %s: read %s: discard\n",
		     hd->sd_name, lis_msg_type_name(mp)) ;
	    lis_freemsg(mp);
	    break;

	case M_PASSFP:
	    lis_free_passfp(mp);
	    RTN(-EBADMSG);  /* DON'T discard */
	    
	case M_SIG:			/* should have been removed already */
	    lis_requeue(hd, mp) ;
	    mp = NULL ;
	    lis_check_m_sig(hd) ;		/* watch for M_SIG in q */
	    RTN(-EAGAIN) ;

	default:
	    if (LIS_DEBUG_READ)
		printk(
		    "strread: stream %s: read %s: return EBADMSG\n",
		     hd->sd_name, lis_msg_type_name(mp)) ;
	    lis_freemsg(mp);
	    RTN(-EBADMSG);
	}/*sw db_type*/
    } while (1);

return_point:
    if (hd_rq != NULL && qlocked)		/* do we have the rq locked? */
    {
	lis_check_m_sig(hd) ;			/* watch for M_SIG in q */
	lis_unlockq(hd_rq) ;
    }

    if (hd_locked)
	lis_spin_unlock(&hd->sd_lock) ;		/* unlock strm head */
    ULOCK_INO(i);
    CLOCKOFF(READTIME) ;
    lis_atomic_dec(&lis_in_syscall) ;	/* done processing a system call */
    lis_runqueues();				/* run the streams queues */
    return(err) ;				/* return to user */

#undef	RTN

}/*lis_strrread*/

/*  -------------------------------------------------------------------  */
/* strputmsg - stream head handler for putmsg
 */
int
lis_strputpmsg(struct inode *i, struct file *fp, 
	       void *ctlp, void *datp, int band, int flags)
{
    int		    err = 0 ;
    int		    couldput ;
    int		    hd_locked = 0 ;
    mblk_t	   *msg;
    stdata_t	   *hd;
    const char     *name = "" ;
    strbuf_t	   *ctl = (strbuf_t *) ctlp ;
    strbuf_t	   *dat = (strbuf_t *) datp ;
    queue_t	   *q ;
    unsigned long  time_cell = 0 ;

#define	RTN(val)  do { err = val; goto return_point; } while (0)

    CHECK_INO(i,"lis_strputmsg"); /* may return */

    hd = FILE_STR(fp) ;
    if (hd == NULL)
	return(-ENODEV);

    CLOCKON() ;
    lis_spin_lock(&hd->sd_lock) ;		/* lock strm head */
    hd_locked = 1 ;
    lis_atomic_inc(&lis_in_syscall) ;	/* processing a system call */
    if (LIS_DEBUG_PUTMSG) name = lis_strm_name(hd) ;

    if (F_ISSET(hd->sd_flag,STPLEX))
    {
	if (LIS_DEBUG_PUTMSG)
	  printk( "strputmsg: stream %s: stream has been I_LINKed\n",
		    hd->sd_name) ;

	RTN(-EINVAL) ;
    }

#if defined(LINUX) && defined(KERNEL_2_1)
    if (F_ISSET(hd->sd_flag,STFIFO)) {
	ULOCK_INO(i);
	if ((err = lis_fifo_write_sync( i, 0 )) < 0) {
	    RTN(err);
	}
	LOCK_INO(i);
    }
#endif

    if (band >= 0)				/* putpmsg */
    {
	if ( (flags & (MSG_HIPRI|MSG_BAND)) == (MSG_HIPRI|MSG_BAND) )
	    RTN(-EINVAL) ;			/* flags are exclusive */
	if ( flags & ~(MSG_HIPRI|MSG_BAND) )
	    RTN(-EINVAL) ;			/* other flags set */
	if (flags == 0)
	    RTN(-EINVAL) ;			/* flags must be non-zero */
	if (F_ISSET(flags,MSG_HIPRI))		/* usr wants hi priority msg */
	    band = 0 ;				/* must be band zero */
    }
    else					/* putmsg */
    {
	if (flags & ~RS_HIPRI)
	    RTN(-EINVAL) ;			/* other flags set */
    }

    if ( (err=copyin_msg(fp,hd,&msg,ctl,dat))<0)	/* allocates msg */
    {
	RTN(err);
    }

    if (msg == NULL)			/* no message to send */
    {
	if (band >= 0)			/* putpmsg */
	{
	    if (flags & ~MSG_BAND)	/* asking for hi-pri */
		RTN(-EINVAL) ;		/* no can do */
	}
	else				/* putmsg */
	{
	    if (flags & RS_HIPRI)	/* asking for hi-pri */
		RTN(-EINVAL) ;		/* no can do */
	}

	RTN(0) ;			/* no options, no error */
    }

    if (band >= 0)			/* putpmsg */
    {
	if (F_ISSET(flags,MSG_HIPRI))		/* usr wants hi priority msg */
	{
	    if (lis_btype(msg) == M_PROTO)	/* is there a ctl part */
		lis_btype(msg)=M_PCPROTO;	/* yes */
	    else
	    {
		lis_freemsg(msg);
		RTN(-EINVAL) ;		/* no, sorry */
	    }
	}
	else					/* regular priority msg */
	if (F_ISSET(flags,MSG_BAND))		/* non-zero band */
	{
	    if (band >=0 && band <=LIS_MAX_BAND)	/* range check */
		lis_bband(msg) = (unsigned char) band;
	    else
	    {
		lis_freemsg(msg);
		RTN(-EINVAL);
	    }
	}
    }
    else					/* putmsg */
    {
	if (F_ISSET(flags,RS_HIPRI))		/* usr wants hi priority msg */
	{
	    if (lis_btype(msg) == M_PROTO)	/* is there a ctl part */
		lis_btype(msg)=M_PCPROTO;	/* yes */
	    else
	    {
		lis_freemsg(msg);
		RTN(-EINVAL) ;			/* no, sorry */
	    }
	}
    }

    /*
     * wait for flow control 
     *
     * If we are sending a high priority message (M_PCPROTO) ignore flow
     * control and send the messasge downstream unconditionally.
     *
     * Always check for error or hangup.
     */
    do
    {
	if (F_ISSET(hd->sd_flag,STWRERR))
	{
	    if (LIS_DEBUG_PUTMSG)
	      printk("strputmsg: stream %s: stream has received M_ERROR %d\n",
			hd->sd_name, hd->sd_werror) ;

	    freemsg(msg) ;
	    RTN(-hd->sd_werror);
	}

	if (F_ISSET(hd->sd_flag,STRHUP|STRCLOSE))
	{
	    if (LIS_DEBUG_PUTMSG)
	      printk("strputmsg: stream %s: stream has received M_HANGUP\n",
			hd->sd_name) ;
	    freemsg(msg) ;
	    RTN(-ENODEV);
	}

	if (lis_btype(msg) == M_PCPROTO)	/* high priority msg */
	{
	    couldput = 1 ;			/* pretend flw ctl enabled */
	    break ;
	}

	couldput = (lis_atomic_read(&hd->sd_wrcnt) == 0) &&
		   lis_bcanputnext(hd->sd_wq,
				   (unsigned char) (band < 0 ? 0 : band));
						    /* check the band */
	if (   !couldput			/* flow control blocked */
	    && should_notblock_wr(hd,fp)	/* NDELAY or NONBLOCK */
	   )
	{
	    if (LIS_DEBUG_PUTMSG)
	      printk("strputmsg: stream %s: flw cntrl blocked, return EAGAIN\n",
		      hd->sd_name) ;
	    freemsg(msg) ;

	    RTN(-EAGAIN);
	}

	if (!couldput)				/* flw cntrl blocked */
	{
	    if (LIS_DEBUG_PUTMSG)
		printk("strputmsg: stream %s: flw cntrl blocked, sleep\n",
			hd->sd_name) ;

	    CLOCKADD() ;
	    err = lis_wait_on_wwrite(hd);	/* unlocks & relocks head */
	    CLOCKON() ;
	    if (err < 0)
	    {
		lis_freemsg(msg) ;
		RTN(err) ;
	    }
	}

    } while (!couldput) ;
	    
    if (LIS_DEBUG_PUTMSG)
    {
	printk(
	   "strputmsg: stream %s: send %d/%d ctl/data bytes downstream to %s\n",
	    hd->sd_name,
	    msg->b_datap->db_type == M_DATA ? 0 : lis_xmsgsize(msg),
	    lis_msgdsize(msg),
	    lis_queue_name(hd->sd_wq->q_next)) ;

	if (LIS_DEBUG_ADDRS)
	    printk("strputmsg: q = 0x%lx, msg = 0x%lx\n",
		    (long)hd->sd_wq, (long)msg);
    }

    q = hd->sd_wq ;
    CLOCKOFF(PUTMSGTIME) ;
    if (hd_locked)
	lis_spin_unlock(&hd->sd_lock) ;		/* unlock strm head */
    lis_putnext(q,msg);				/* send downstream */
    						/* don't use PUTNEXT */
    err = 0 ;
    goto return_unlocked ;

return_point:
    CLOCKOFF(PUTMSGTIME) ;
    if (hd_locked)
	lis_spin_unlock(&hd->sd_lock) ;		/* unlock strm head */
return_unlocked:
    lis_atomic_dec(&lis_in_syscall) ;	/* done processing a system call */
    lis_runqueues();
    return(err) ;

#undef RTN

}/*lis_strputpmsg*/

/*  -------------------------------------------------------------------  */
/* strgetmsg - stream head handler for getmsg
 *
 * If 'doit' is set then this is a getmsg or getpmsg system call.
 * In this case 'ctl', 'dat' and 'flagsp' are all user space addresses.
 * 'bandp' can be NULL if the call is from getmsg().  It will be
 * a non-null user-space address if the call is from getpmsg().
 *
 * If 'doit' is not set then this is an I_PEEK call.  In this case
 * 'ctl', 'dat' and 'flagsp' and 'bandp' point to kernel memory.
 *
 * Note that getmsg() and I_PEEK use one encoding of the flags
 * (the RS_...) and that getpmsg() uses another (the MSG_...).
 */
int
lis_strgetpmsg(struct inode *i, struct file *fp, 
	       void *ctlp, void *datp, int *bandp, int *flagsp,
	       int doit)
{
    strbuf_t	*ctl = (strbuf_t *) ctlp ;
    strbuf_t	*dat = (strbuf_t *) datp ;
    const char	*ctl_type = "" ;
    const char	*dat_type = "" ;
    int		 err= 0 ;
    int		 rtn = 0 ;
    int		 qlocked = 0 ;
    int		 hd_locked = 0 ;
    stdata_t	*hd;
    strbuf_t	 kctl,kdat;
    mblk_t	*mp;
    queue_t	*rdq ;
    int		 band = 0 ;
    int		 flags = 0 ;
    int		 mtype ;
    unsigned long  time_cell = 0 ;

#define RTN(v)	do { err=(v); goto err_return_point; } while (0)

    CHECK_INO(i,"lis_strgetmsg"); /* may return */

    hd = FILE_STR(fp) ;
    if (hd == NULL)
	return(-ENODEV);

    CLOCKON() ;
    lis_spin_lock(&hd->sd_lock) ;		/* lock strm head */
    hd_locked = 1 ;
    lis_atomic_inc(&lis_in_syscall) ;	/* processing a system call */
    if (F_ISSET(hd->sd_flag,STRDERR))
	RTN(-hd->sd_rerror);
    else if (doit && ctl != NULL &&
            (err=lis_check_umem(fp,VERIFY_READ,ctl,sizeof(strbuf_t)))<0)
       RTN(err);
    else if (doit && dat != NULL &&
            (err=lis_check_umem(fp,VERIFY_READ,dat,sizeof(strbuf_t)))<0)
       RTN(err);
    else if (flagsp == NULL)
	RTN(-EFAULT);
    else if (doit && (err=lis_check_umem(fp,VERIFY_READ,flagsp,sizeof(int)))<0)
	RTN(err);
    else if (doit && bandp != NULL
	     && (err=lis_check_umem(fp,VERIFY_READ,bandp,sizeof(int)))<0)
	RTN(err);

    if (F_ISSET(hd->sd_flag,STPLEX))
    {
	if (LIS_DEBUG_GETMSG)
	  printk("strgetmsg: stream %s: stream has been I_LINKed\n",
		    hd->sd_name) ;

	RTN(-EINVAL) ;
    }

    rdq = hd->sd_rq ;
    if (!LIS_CHECK_Q_MAGIC(rdq)) RTN(-EINVAL) ;

    if (doit)
    {					/* fetch structs from usr space */
	if (ctl != NULL)
	{
	    if ((err = lis_copyin(fp,&kctl,ctl,sizeof(strbuf_t))) < 0)
		RTN(err) ;
	}
	else
	{
	    kctl.maxlen = -1 ;
	    kctl.len    = -1 ;
	    kctl.buf    = NULL ;
	}

	if (dat != NULL)
	{
	    if ((err = lis_copyin(fp,&kdat,dat,sizeof(strbuf_t))) < 0)
		RTN(err) ;
	}
	else
	{
	    kdat.maxlen = -1 ;
	    kdat.len    = -1 ;
	    kdat.buf    = NULL ;
	}

	if ((err = lis_copyin(fp,&flags,flagsp,sizeof(flags))) < 0)
	    RTN(err) ;
	if (doit && bandp != NULL)
	{
	    if ((err = lis_copyin(fp,&band,bandp,sizeof(band))) < 0)
		RTN(err) ;
	}
	else
	    band  = 0 ;

	if (kctl.maxlen >= 0 || kdat.maxlen >= 0)
	{
	    lis_lockq(rdq) ;
	    qlocked = 1 ;
	    while (   rdq->q_first == NULL
	           && (hd->sd_flag & (STRHUP | STRDERR | STRCLOSE)) == 0
	          )
	    {
		if (should_notblock_rd(hd,fp))
		{
		    lis_unlockq(rdq) ;
		    RTN(-EAGAIN);
		}

		ULOCK_INO(i);
		CLOCKADD() ;
		err = lis_sleep_on_wread(hd);	       /* unlocks & relocks q */
		CLOCKON() ;
		LOCK_INO(i);
		if (err < 0) 
		{
		    lis_unlockq(rdq) ;
		    RTN(err) ;
		}
	    }

	    lis_unlockq(rdq) ;
	    qlocked = 0 ;
	}
    }
    else				/* fetch structs from kernel space */
    {
	kctl  = *ctl ;
	kdat  = *dat ;
	flags = *flagsp ;
	band  = *bandp ;
    }

    /*
     * Lock the rd queue.  Every exit from here on should go out
     * via the "return_point" label for cleanup purposes.  The queue
     * lock keeps the read service procedure from running and messing
     * with the queue.
     */
    lis_lockq(rdq) ;
    qlocked = 1 ;

    if ((mp = rdq->q_first) == NULL)		/* mp is 1st msg */
    {						/* empty read queue */
	kctl.len = -1;				/* nothing present on queue */
	kdat.len = -1;
	if (F_ISSET(hd->sd_flag,STRHUP|STRCLOSE))
	{
	    kctl.len=0;			      /* so sez STREAMS pgrmr's guide */
	    kdat.len=0;
	    rtn = 0 ;
	}
	else
	if (F_ISSET(hd->sd_flag,STRDERR))
	    err = -hd->sd_rerror;
	/*
	 * Return the strbuf structures whether or not there are errors.
	 * copyout_msg will not modify our len values due to the NULL
	 * message parameter.
	 */
	rtn=copyout_msg(fp,NULL,&kctl,&kdat,
		    (strbuf_t*)ctl,(strbuf_t*)dat,doit) ;

	goto return_point ;			/* go return 'rtn' */
    }

    if (LIS_DEBUG_GETMSG && LIS_DEBUG_DMP_MBLK && LIS_DEBUG_ADDRS)
    {
	printk("lis_strgetpmsg: q=%lx, mp=%lx, mp->b_rptr=%lx\n",
				(long)rdq, (long)mp, (long)mp->b_rptr);
    }

    /*
     * At this point we have a message.  Ensure that it is
     * of the type that we want to process.
     */
    mtype = lis_btype(mp) ;
    if (LIS_DEBUG_GETMSG)
    {
	if (mtype != M_DATA)
	{
	    ctl_type = lis_msg_type_name(mp) ;
	    dat_type = lis_msg_type_name(mp->b_cont) ;
	}
	else
	{
	    ctl_type = "NULL-MSG" ;
	    dat_type = lis_msg_type_name(mp) ;
	}
    }

    switch (mtype)
    {
    case M_DATA:
    case M_PROTO:
    case M_PCPROTO:
    case M_SIG:
	break ;				/* OK, keep going */

    case M_PASSFP:
	mp = lis_getq(rdq);			/* remove from queue */
	LisDownCount(MSGQDSTRHD) ;
	lis_free_passfp(mp);
	mp = NULL ;
	lis_check_m_sig(hd) ;			/* watch for M_SIG in q */
	err = -EBADMSG;
	goto return_point;
    default:
	err = -EBADMSG;
	goto return_point;
    }

    if (doit)					/* it's get[p]msg() */
    {
	if (lis_hipri(mtype))			/* M_PCPROTO */
	{
	    F_CLR(hd->sd_flag,STRPRI);		/* about to remove the msg */
	    flags = MSG_HIPRI ;
	    band  = 0 ;
	}
	else					/* normal priority msg */
	{
	    switch (flags)
	    {
	    case MSG_HIPRI:			/* user wanted hi-pri only */
						/* same code as RS_HIPRI */
		err = -EBADMSG;			/* hi-pri not at head of q */
		goto return_point ;		/* go return 'rtn' */

	    case MSG_BAND:			/* wants from particular band */
		if (bandp == NULL)		/* getmsg, not getpmsg */
		{
		    err = -EINVAL;		/* invalid flags */
		    goto return_point ;		/* go return 'rtn' */
		}
		if (mp->b_band < (unsigned char) band) /* msg in wrong band */
		{
		    err = -EBADMSG;
		    goto return_point ;		/* go return 'rtn' */
		}
		break ;				/* OK, keep going */

	    case MSG_ANY:			/* wants msg from any band */
		if (bandp == NULL)		/* getmsg, not getpmsg */
		{
		    err = -EINVAL;		/* invalid flags */
		    goto return_point ;		/* go return 'rtn' */
		}
		flags = MSG_BAND ;		/* reduce to MSG_BAND case */
		break ;

	    default:				/* other flags or zero */
		if (   flags != 0 		/* invalid flags */
		    || (bandp != NULL && flags == 0)	/* getpmsg w/flags=0 */
		   )
		{
		    err = -EINVAL;			/* invalid flags */
		    goto return_point ;		/* go return 'rtn' */
		}
		break ;
	    }

	    band  = mp->b_band ;		/* band to return */
	}

	/*
	 * Now we commit to removing the message from the read queue
	 * Up until here error conditions could have left the message
	 * in the queue.
	 */
	mp = lis_getq(rdq);			/* remove from queue */
	if ( mp == NULL )
	{
	    printk("strgetpmsg: empty queue: addr of queue 0x%lx\n",
		    (long) rdq);
	    err = -EINVAL;
	    goto return_point;
	}

	lis_check_m_sig(hd) ;		/* watch for M_SIG in q */
	lis_unlockq(rdq) ;		/* now we can unlock the queue */
	qlocked = 0 ;

	LisDownCount(MSGQDSTRHD) ;
	if (LIS_DEBUG_GETMSG && LIS_DEBUG_DMP_DBLK)
	{
	    lis_print_msg(mp, "lis_strgetpmsg", PRINT_DATA_RDWR) ;
	}

	if (mtype == M_SIG)			/* process M_SIG */
	{
	    lis_requeue(hd, mp) ;
	    mp = NULL ;
	    lis_check_m_sig(hd) ;		/* watch for M_SIG in q */
	    err = -EAGAIN ;
	    goto return_point ;
	}
    }
    else					/* it's I_PEEK */
    {						/* have to keep queue locked */
	if (mtype == M_SIG)			/* not suitable msg */
	{					/* msg still in queue */
	    lis_check_m_sig(hd) ;		/* watch for M_SIG in q */
	    err = -EAGAIN ;
	    goto return_point ;
	}

	if (lis_hipri(mtype))			/* M_PCPROTO */
	    flags = MSG_HIPRI ;
	else					/* M_DATA or M_PROTO */
	if (flags == RS_HIPRI)			/* wants hi-priority only */
	{
	    err = -EBADMSG;
	    goto return_point ;			/* go return 'rtn' */
	}
	else
	    flags = MSG_BAND ;			/* regular priority */
    }

    /* 
     * copyout_msg will copy data to usr returning MORECTL,MOREDATA status
     */
    if ((rtn=copyout_msg(fp,mp,&kctl,&kdat,(strbuf_t*)ctl,(strbuf_t*)dat,doit))<0)
    {
	if (doit)
	    lis_freemsg(mp);
    } 
    else 
    if (doit)				/* getmsg/getpmsg */
    {
	if (rtn)			/* there's still ctl or data in msg */
	  lis_requeue(hd, mp) ;		/* re-queue */
	else
	  lis_freemsg(mp);		/* done with message now */
    }
    else				/* I_PEEK */
	rtn = kctl.len >= 0 || kdat.len >= 0 ;	/* got something */


return_point:

    if (rdq != NULL && qlocked)
    {
	lis_check_m_sig(hd) ;		/* watch for M_SIG in q */
	lis_unlockq(rdq) ;
    }

    if (err == 0 && rtn >= 0)		/* return flags to user */
    {
	if (!doit || bandp == NULL)	/* I_PEEK or getmsg */
	{				/* different encoding of flags */
	    switch (flags)
	    {
	    case MSG_HIPRI:
		flags = RS_HIPRI ;
		break ;
	    case MSG_BAND:
	    case MSG_ANY:
	    case 0:
	    default:
		flags = 0 ;
		break ;
	    }
	}

	if (!doit)
	    *flagsp = flags ;		/* return flags to kernal space */
	else				/* return flags to user space */
	if ((err=lis_check_umem(fp,VERIFY_WRITE,flagsp,sizeof(int))) == 0)
	    err = lis_copyout(fp,&flags,flagsp,sizeof(flags));
    }

    if (err == 0 && rtn >= 0 && doit && bandp != NULL)		/* getpmsg() */
    {
	if ((err=lis_check_umem(fp,VERIFY_WRITE,bandp,sizeof(int))) == 0)
	    err = lis_copyout(fp,&band,bandp,sizeof(band));
    }
    else				/* getmsg or I_PEEK */
    if (!doit && bandp != NULL)		/* kernel pointer */
	*bandp = band ;

    if (LIS_DEBUG_GETMSG)
    {
      printk("strgetmsg:    maxlen     len    return=0x%x flags=0x%x band=%d",
					  err ? err : rtn,
							  flags,     band) ;
      printk(" stream %s\n", hd->sd_name) ;
      printk("  control:    %6d  %6d    %s\n",
		  kctl.maxlen, kctl.len, ctl_type) ;
      printk("     data:    %6d  %6d    %s\n",
		  kdat.maxlen, kdat.len, dat_type) ;
    }

err_return_point:				/* return "err" */
    if (hd_locked)
	lis_spin_unlock(&hd->sd_lock) ;		/* unlock strm head */

    CLOCKOFF(GETMSGTIME) ;
    lis_atomic_dec(&lis_in_syscall) ;	/* done processing a system call */
    lis_runqueues();
    return(err != 0 ? err : rtn);

#undef RTN

}/*lis_strgetpmsg*/


/*  -------------------------------------------------------------------  */
/* strioctl - handle IOCTLs at the stream head
 *
 * NOTE: This functions needs pretty much cleanup. Perhaps ioctling stuff
 * should be moved to strioc.c (or alike). 
 */
int
lis_strioctl( struct inode *i, struct file *f, unsigned int cmd,
	      unsigned long arg )
{
    stdata_t *hd;
    int err=0,flush_pri=0;
    cred_t creds;
    streamtab_t *st;
    unsigned long  time_cell = 0 ;
    int		   hd_locked = 0 ;
    pid_t pid=(pid_t)PID(f);

#define RTN(v)	do { err=(v); goto return_point; } while (0)

    CHECK_INO(i,"lis_strioctl"); /* may return */

    if (LIS_DEBUG_IOCTL && LIS_DEBUG_ADDRS)
#ifdef KERNEL_2_1				/* linux >= 2.1.0 */
	printk("lis_strioctl: f=%lx i=%lx i->i_rdev=%x cmd=%x\n",
	       (long) f, (long) i, i->i_rdev,cmd) ;
#else						/* linux < 2.1.0 or portable */
	printk("lis_strioctl: f=%lx f->f_inode=%lx i=%lx i->i_rdev=%x cmd=%x\n",
	       (long) f, (long) f->f_inode, (long) i, i->i_rdev,cmd) ;
#endif

    hd = FILE_STR(f) ;
    if (hd == NULL)
	return(-ENODEV);

    CLOCKON() ;
    lis_spin_lock(&hd->sd_lock) ;		/* lock strm head */
    hd_locked = 1 ;
    lis_atomic_inc(&lis_in_syscall) ;		/* processing a system call */
    creds.cr_uid  = (uid_t) EUID(f);
    creds.cr_gid  = (gid_t) EGID(f);
    creds.cr_ruid = (uid_t) UID(f);
    creds.cr_rgid = (gid_t) GID(f);

    if (F_ISSET(hd->sd_flag,STPLEX))
    {
	if (LIS_DEBUG_IOCTL)
	  printk("strioctl: stream %s: stream has been I_LINKed\n",
		    hd->sd_name) ;

	RTN(-EINVAL) ;
    }


    switch( cmd ){		/* the often an ioc is called the 
				 * closer it should be to the begin of the sw*/
    case I_PUSH:		/* tested */
	if (F_ISSET(hd->sd_flag,STRHUP|STRCLOSE))
	    RTN(-ENXIO);

	if ((err=lis_check_umem(f,VERIFY_READ,(char*)arg,FMNAMESZ+1))<0)
	    RTN(err);
	else {
	    char *mname;
	    if((err=lis_copyin_str(f,(char*)arg,&mname,FMNAMESZ+1))<0)
		RTN(err);

	    if ( LIS_DEBUG_IOCTL )
		printk("strioctl: I_PUSH \"%s\" below %s: ",
			mname, hd->sd_name) ;

	    if ((st = lis_modstr(lis_findmod(mname))) == NULL)
	    {
		err=-EINVAL;
		if ( LIS_DEBUG_IOCTL )
		    printk("module not found\n") ;

		FREE((caddr_t)mname);
	    }
	    else
	    {
		dev_t ddev = i->i_rdev;

		FREE((caddr_t)mname);
		lis_spin_unlock(&hd->sd_lock) ;		/* unlock strm head */
		err = push_mod( hd, st, &ddev, f->f_flags, &creds );
		lis_spin_lock(&hd->sd_lock) ;		/* lock strm head */
		if ( LIS_DEBUG_IOCTL )
		{
		    if (err)	printk("error %d\n", err) ;
		    else	printk("OK\n") ;
		}

		if (F_ISSET(hd->sd_flag,STRISTTY)){
		    /* 
		       we should make this the controling terminal...
		     */
		    hd->sd_pgrp = PGRP(f);
		}
	    }
	}
	break;
    case I_POP:		/* tested */
      {
	const char	*qname = "";

	if (F_ISSET(hd->sd_flag,STRHUP|STRCLOSE))
	    RTN(-ENXIO);

	if ( LIS_DEBUG_IOCTL )
	    qname = lis_queue_name(hd->sd_wq->q_next) ;

	lis_spin_unlock(&hd->sd_lock) ;		/* unlock strm head */
	err=pop_mod(hd,f->f_flags, &creds); 
	lis_spin_lock(&hd->sd_lock) ;		/* lock strm head */

	if ( LIS_DEBUG_IOCTL )
	{
	    printk("strioctl: I_POP \"%s\" from stream %s: ",
		    qname, hd->sd_name) ;

	    if (err)	printk("error %d\n", err) ;
	    else	printk("OK\n") ;
	}
      }
      break;
    case I_LOOK:		/* tested */
      {
	int	namelen ;
	char   *name ;
	if ((err=lis_check_umem(f,VERIFY_WRITE,(char*)arg,FMNAMESZ+1))<0)
	{
	    if ( LIS_DEBUG_IOCTL )
		printk("strioctl: I_LOOKED lis_check_umem failed with err=%d\n",
			err);
	    RTN(err);
	}
	if (!hd->sd_pushcnt)
	{
	    if ( LIS_DEBUG_IOCTL )
		printk("strioctl: I_LOOKED failed - no modules on stream %d\n",
		-EINVAL);
	    RTN(-EINVAL);
	}
	if ( LIS_DEBUG_IOCTL )
	  printk("strioctl: I_LOOK: top driver/module of stream %s is \"%s\"\n",
		    hd->sd_name, lis_queue_name(hd->sd_wq->q_next)) ;

	name = (char *) lis_queue_name(hd->sd_wq->q_next) ;
	namelen = strlen(name) ;
	if (namelen > FMNAMESZ)
	    namelen = FMNAMESZ ;

	err = lis_copyout(f,name, (char*)arg, namelen+1);
	break;
      }
    case I_FLUSHBAND:
	{
	    struct bandinfo bi;
	    if (   (err=lis_check_umem(f,VERIFY_READ,
				    (char*)arg,sizeof(bandinfo_t)))<0
		|| (err=lis_copyin(f,&bi,(char*)arg,sizeof(bandinfo_t))) < 0
	       )
		RTN(err);
	    arg = (bi.bi_flag & FLUSHRW) | FLUSHBAND;
	    flush_pri=bi.bi_pri;
	}
        /*fall thru...*/
    case I_FLUSH:
	if (F_ISSET(hd->sd_flag,STRHUP|STRCLOSE))
	    RTN(-ENXIO);
	if ( LIS_DEBUG_IOCTL )
	    printk("strioctl: I_FLUSH arg=0x%lx\n", arg) ;

	switch(arg & FLUSHRW){
	case FLUSHR:
#ifdef FIFO_IMPL
	    if (LIS_DEV_IS_FIFO(STR_MAJOR(i->i_rdev))){
	      /*flush rw/other-wr for fifo;*/
	    }
	    else
#endif
	    {
	      int res = lis_snd_mflush(hd->sd_wq,arg,flush_pri);
	      RTN(res);
	    }
	    break;
	case FLUSHW:
#ifdef FIFO_IMPL
	    if (LIS_DEV_IS_FIFO(STR_MAJOR(i->i_rdev))){
		/*flush rw/other-wr for fifo;*/
	    }
	    else
#endif
	    {
	      int res=lis_snd_mflush(hd->sd_wq,arg,flush_pri);
	      RTN(res);
	    }
	    break ;
	case FLUSHRW:
#ifdef FIFO_IMPL
	    if (LIS_DEV_IS_FIFO(STR_MAJOR(i->i_rdev))){
		/*flush rd/other-rd;*/
	    }
	    else
#endif
	    {
		int res=lis_snd_mflush(hd->sd_wq,arg,flush_pri);
		RTN(res);
	    }
	default:
	    RTN(-EINVAL);
	}
	break;
    case I_SETSIG:
	if (!arg){
	    if (lis_del_from_elist(&(hd->sd_siglist),pid,S_ALL))
		err = -EINVAL;
	    else 
		if (!hd->sd_siglist)
		    hd->sd_sigflags=0;
	}
	else 
	    if (lis_add_to_elist(&(hd->sd_siglist),pid, (short) arg))
		err = -EAGAIN;
	    else
		F_SET(hd->sd_sigflags,arg);

	if ( LIS_DEBUG_IOCTL || LIS_DEBUG_SIG )
	{
	    printk("strioctl: I_SETSIG: stream %s signal %ld: ",
		    hd->sd_name, arg) ;
	    if (err)	printk("error %d\n", err) ;
	    else	printk("OK\n") ;
	}

	break;
    case I_GETSIG:
	{
	    int evs;

	    if ( LIS_DEBUG_IOCTL )
		printk("strioctl: I_GETSIG: stream %s: ", hd->sd_name) ;

	    err=lis_check_umem(f,VERIFY_WRITE,(char*)arg,sizeof(int)) ;
	    if (   err == 0
		&& (evs=(int)lis_get_elist_ent(hd->sd_siglist,pid)) == 0
	       )
		err = -EINVAL;

	    if ( LIS_DEBUG_IOCTL )
	    {
		if (err)	printk("error %d\n", err) ;
		else		printk("signal %d\n", evs) ;
	    }

	    if (err == 0)
		err = lis_copyout(f,&evs,(char*)arg,sizeof(int));
	}
	break;
    case I_FIND:		/* tested */
	{
	    char *mname = NULL;
	    queue_t *q;

	    if ((err=lis_check_umem(f,VERIFY_READ,(char*)arg,FMNAMESZ+1))<0)
		RTN(err);
	    if (hd->sd_pushcnt){
		if ((err=lis_copyin_str(f,(char*)arg,&mname,FMNAMESZ+1))<0)
		    RTN(err);
		if ( LIS_DEBUG_IOCTL )
		    printk("strioctl: I_FIND: stream %s module \"%s\"\n",
				hd->sd_name, mname);
		for (q=hd->sd_wq; 
		     q && strncmp(lis_queue_name(q),mname,FMNAMESZ);
		     q=q->q_next )
		{
		    if (!LIS_CHECK_Q_MAGIC(q))
		    {
			FREE(mname) ;
			RTN(-EINVAL);
		    }
		}
		if (q){		
		    FREE(mname) ;
		    RTN(1);	/* found! */
		}
	    }
	    if (mname == NULL)
	        RTN(0);		/* not present */

	    if (!lis_findmod(mname))
	    {
		FREE(mname) ;
		RTN(-EINVAL);	/* not a valid module */
	    }

	    FREE(mname) ;
	    RTN(0);		/* not present */
	}
    case I_PEEK:
	{
	    strpeek_t pb;
	    strbuf_t *ctl=&(pb.ctlbuf), *dat=&(pb.databuf);
	    int band =0;
	    int res,flags;
	    if (   (err=lis_check_umem(f,VERIFY_READ,
				    (char*)arg,sizeof(strpeek_t)))<0
		|| (err=lis_check_umem(f,VERIFY_WRITE,
				    (char*)arg,sizeof(strpeek_t)))<0
		|| (err=lis_copyin(f,&pb,(char*)arg,sizeof(strpeek_t)))<0
	       )
		RTN(err);
            flags=(int)pb.flags;
	    lis_spin_unlock(&hd->sd_lock) ;		/* unlock strm head */
	    res=lis_strgetpmsg(i,f,ctl,dat,&band,&flags,0); /*0-> just peek!*/
	    lis_spin_lock(&hd->sd_lock) ;		/* lock strm head */
            pb.flags=(long)flags;
	    err = 0 ;
	    if (res >= 0)
		err = lis_copyout(f,&pb,(char*)arg,sizeof(strpeek_t));
	    if (err < 0)		/* copyout failed */
		res = err ;		/* need to return error */
	    if ( LIS_DEBUG_IOCTL )
		printk("strioctl: I_PEEK: stream %s flags=0x%x rslt=%d\n",
				hd->sd_name, flags, res) ;
	    RTN(res);
	}
	break;
    case I_SRDOPT:
      {
	int res = set_readopt(&hd->sd_rdopt,arg);
	if ( LIS_DEBUG_IOCTL )
	    printk("strioctl: I_SRDOPT: stream %s arg=%ld rslt=%d\n",
				hd->sd_name, arg, res) ;
	RTN(res);
      }
    case I_GRDOPT:
	{
	    int rmode;
	    if ((err=lis_check_umem(f,VERIFY_WRITE,(char*)arg,sizeof(int)))<0)
		RTN(err);
	    rmode=hd->sd_rdopt;
	    if ( LIS_DEBUG_IOCTL )
		printk("strioctl: I_GRDOPT: stream %s mode=0x%x\n",
				hd->sd_name, rmode);
	    err = lis_copyout(f,&rmode,(char*)arg,sizeof(int));
	}	
	break;
    case I_NREAD:		/* tested */
	{ 
	    int bytes=0;
	    if ((err=lis_check_umem(f,VERIFY_WRITE,(char*)arg,sizeof(int)))<0)
		RTN(err);
	    if (hd->sd_rq->q_first)
		bytes=lis_msgdsize(hd->sd_rq->q_first);
	    if ( LIS_DEBUG_IOCTL )
		printk("strioctl: I_NREAD: stream %s bytes=%d\n",
				hd->sd_name, bytes);
	    err = lis_copyout(f,&bytes,(char*)arg,sizeof(int));
	    if (err == 0)
		RTN(lis_qsize(hd->sd_rq)) ;	/* rtn # msgs in queue */
	}
	break;
    case I_FDINSERT:
      {
	/* This code compliments of Eric Levenez <levenez@club-internet.fr> */
	/* see magic garden 7.9.7 */
	strfdinsert_t	 strfdinsert;
	stdata_t	*ihd;
	queue_t		*q;
	mblk_t		*msg;
	int		 couldput;
	strbuf_t	*ctl;
	strbuf_t	*dat;

	/* primary sanity checks */

	if (F_ISSET (hd->sd_flag, STRDERR) || F_ISSET (hd->sd_flag, STWRERR))
	  RTN(-EINVAL);

	if (!hd->sd_wq->q_next)
	  RTN(-ENOSR);

	if (   (err = lis_check_umem (f, VERIFY_READ, (char *) arg,
				   sizeof (strfdinsert_t))) < 0
	    || (err = lis_copyin (f, &strfdinsert, (char *) arg,
				   sizeof (strfdinsert_t))) < 0
	   )
	    RTN(err);

	/* read structure command */


	/* sanity checks */
	if (   (strfdinsert.offset < 0)
	    || (strfdinsert.offset > 4096)
	    || (strfdinsert.ctlbuf.len < strfdinsert.offset + sizeof (char *))
	    || (strfdinsert.offset % sizeof (char *))
	    || (strfdinsert.fildes < 0)
	    || (strfdinsert.flags & ~RS_HIPRI))
	    RTN(-EINVAL);

	if ((ihd = lis_fd2str(strfdinsert.fildes)) == NULL)
	  RTN(-ENOSTR);

	if (   F_ISSET (ihd->sd_flag, STRDERR)
	    || F_ISSET (ihd->sd_flag, STWRERR))
	      RTN(-EINVAL);

	if (F_ISSET (ihd->sd_flag, STPLEX))
	  RTN(-ENXIO);

	if (   (   (strfdinsert.databuf.len > 0)
		&& (strfdinsert.databuf.len < hd->sd_minpsz)
	       )
	    || (   (hd->sd_maxpsz != INFPSZ)
		&& (strfdinsert.databuf.len > hd->sd_maxpsz)
	       )
	   )
	      RTN(-ERANGE);

	/* find driver queue */

	for (q = ihd->sd_wq; q->q_next; q = q->q_next)
	{
	    if (!LIS_CHECK_Q_MAGIC(q)) RTN(-EINVAL) ;
	}

	/* create the message a-la putpmsg */

	ctl = (strbuf_t *) (arg +
			    (unsigned long) &((strfdinsert_t *) NULL)->ctlbuf);
	dat = (strbuf_t *) (arg +
			    (unsigned long) &((strfdinsert_t *) NULL)->databuf);

	if ((err = copyin_msg (f, hd, &msg, ctl, dat)) < 0)
	  RTN(err);

	if (!msg)
	  RTN(-EINVAL);

	/* for priority message */

	if (F_ISSET (strfdinsert.flags, RS_HIPRI))
	{
	  if (lis_btype (msg) == M_PROTO)
	    lis_btype (msg) = M_PCPROTO;
	  else
	  {
	    lis_freemsg (msg);
	    RTN(-EINVAL);
	  }
	}

	/* insert the read queue */

	*((queue_t **) (msg->b_rptr + strfdinsert.offset)) = LIS_RD (q);

	/* wait for flow control (code from lis_strputpmsg) */

	do
	{
	  if (F_ISSET (hd->sd_flag, STWRERR))
	  {
	    freemsg (msg);
	    RTN(-EINVAL);
	  }
	  if (F_ISSET (hd->sd_flag, STRHUP|STRCLOSE))
	  {
	    freemsg (msg);
	    RTN(-ENODEV);
	  }
	  couldput = (lis_atomic_read(&hd->sd_wrcnt) == 0) &&
		     lis_bcanputnext (hd->sd_wq, (unsigned char) 0);
	  if (   !couldput
	      && (   should_notblock_wr (hd, f)
		  || F_ISSET (strfdinsert.flags, MSG_HIPRI)
		 )
	     )
	  {
	    freemsg (msg);
	    RTN(-EAGAIN);
	  }

	  if (!couldput)
	  {
	    err = lis_wait_on_wwrite (hd);	/* unlocks & relocks head */
	    if (err < 0)
	    {
	      lis_freemsg (msg);
	      RTN(err);
	    }
	  }
	}
	while (!couldput);

	/* send the message */

	if (LIS_DEBUG_IOCTL)
	  printk ("strioctl: I_FDINSERT: stream %s: q=%lx q2 = %lx \n",
		    hd->sd_name, (unsigned long) hd->sd_wq,
		    (unsigned long) LIS_RD (q));
	q = hd->sd_wq ;
	CLOCKADD() ;				/* exclude driver time */
	lis_spin_unlock(&hd->sd_lock) ;		/* unlock strm head */
	hd_locked = 0 ;
	lis_putnext (q, msg);			/* do not use PUTNEXT */
	CLOCKON() ;
	RTN(0);
      }
      break;

    case I_STR:	
	if (F_ISSET(hd->sd_flag,STRHUP|STRCLOSE))
	    RTN(-ENXIO);
	else {
	    strioctl_t ioc;
	    int res;
	    if (   (err=lis_check_umem(f,VERIFY_READ,
				    (char*)arg,sizeof(strioctl_t)))<0
		|| (err=lis_check_umem(f,VERIFY_WRITE,
				    (char*)arg,sizeof(strioctl_t)))<0
		|| (err=lis_copyin(f,&ioc,(char*)arg,sizeof(strioctl_t)))<0
	       )
		RTN(err);
	    CLOCKADD() ;
	    res=lis_strdoioctl(f,hd,&ioc,&creds,1);
	    CLOCKON() ;
	    if (res == 0)
		res=lis_copyout(f,&ioc,(char*)arg,sizeof(strioctl_t));
	    if ( LIS_DEBUG_IOCTL )
	     printk(
	      "strmioctl: I_STR: "
	      "stream %s ic_cmd=0x%x ic_timout=%d ic_len=%d rslt=%d\n",
			    hd->sd_name,
			    ioc.ic_cmd, ioc.ic_timout, ioc.ic_len, res) ;
		
	    RTN(res);
	}
	break;
    case I_SWROPT:
	if ( LIS_DEBUG_IOCTL )
	    printk("strioctl: I_SWROPT: stream %s arg=0x%lx\n",
			    hd->sd_name, arg) ;
	hd->sd_wropt = arg ;
	RTN(0) ;
    case I_GWROPT:
	if ((err=lis_check_umem(f,VERIFY_WRITE,(char*)arg,sizeof(int)))<0)
	    RTN(err);

	if ( LIS_DEBUG_IOCTL )
	    printk("strioctl: I_GWROPT: stream %s mode=0x%x\n",
			    hd->sd_name, hd->sd_wropt);
	err = lis_copyout(f,&hd->sd_wropt,(char*)arg,sizeof(int));
	RTN(err);
    case I_SENDFD:
	if (LIS_DEBUG_IOCTL)
	    printk( "strioctl: I_SENDFD %d from %s\n",
		    (int)arg, hd->sd_name) ;
	RTN(lis_sendfd( hd, (unsigned int)arg, NULL ));
	
#if defined(I_E_RECVFD)
    case I_E_RECVFD:
#endif
    case I_RECVFD:
	if ((err = lis_check_umem( f, VERIFY_WRITE, (char*)arg,
				   sizeof(strrecvfd_t) )) < 0) {
	    RTN(err);
	} else {
	    strrecvfd_t recv;
	    
	    if (LIS_DEBUG_IOCTL)
		printk("strioctl: I_RECVFD on %s\n", hd->sd_name) ;
	
	    /*
	     *  synchronize with the sender if needed; we will wait
	     *  as long as an IOCTL would wait by default (i.e., 15 secs).
	     *  note that this potential wait can be "tuned" and/or avoided
	     *  by preceding the ioctl(I_RECVFD) with a poll(POLLIN), since
	     *  M_PASSFP is treated like a data message by poll().  If no
	     *  message is available after poll() has waited the specified
	     *  time, the caller can respond accordingly (possibly by not
	     *  calling ioctl(I_RECVFD) at all).
	     */
	    lis_lockq(hd->sd_rq) ;
	    if (!hd->sd_rq->q_first) {  /* any messages? */
		struct timer_list tl;
		int rslt;
		
		if (LIS_DEBUG_IOCTL)
		    printk( "strioctl: I_RECVFD waiting for an M_PASSFP\n" );
		lis_tmout( &tl, lis_do_rd_tmout,
			   (long)hd, SECS_TO(LIS_DFLT_TIM) );
		rslt = lis_sleep_on_wread(hd) ;	/* unlocks & relocks q */
		if (rslt  < 0) {
		    /* we got a signal... */
		    lis_unlockq(hd->sd_rq) ;
		    lis_untmout(&tl);
		    RTN(rslt);
		}
		if (LIS_DEBUG_IOCTL)
		    printk( "strioctl: I_RECVFD wait terminated\n" );
		lis_untmout(&tl);
	    }
	    
	    /*
	     *  we should have a message - it should be an M_PASSFP
	     *  if not, lis_recvfd will return an error
	     */
	    err = lis_recvfd( hd, &recv, f ) ;
	    lis_unlockq(hd->sd_rq) ;
	    if (err < 0)
		RTN(err);
	    
	    RTN(lis_copyout( f, (char *)&recv, (char *)arg,
			     sizeof(strrecvfd_t) ));
	}

    case I_LIST:
	if (!arg)		/* want number of modules */
	{
	    if ( LIS_DEBUG_IOCTL )
		printk("strioctl: I_LIST: stream %s arg=NULL, return %d\n",
				hd->sd_name, hd->sd_pushcnt+1) ;
	    RTN(hd->sd_pushcnt+1); /* +1 for driver  */
	}
	else {
	    str_list_t list;
	    queue_t   *q;
	    int	       nmods ;
	    if (   (err=lis_check_umem(f,VERIFY_READ,(char*)arg,
						sizeof(str_list_t)))<0
	        || (err=lis_check_umem(f,VERIFY_WRITE,(char*)arg,
						sizeof(str_list_t)))<0
		|| (err=lis_copyin(f,&list,(char*)arg,sizeof(str_list_t)))<0
	       )
		RTN(err);
	    if (list.sl_nmods<=0)
		RTN(-EINVAL);
	    if ((err=lis_check_umem(f,VERIFY_WRITE,
				 list.sl_modlist,
				 sizeof(struct str_mlist)*list.sl_nmods)) <0)
		RTN(err);

	    if (hd->sd_wq == NULL || (q=hd->sd_wq->q_next) == NULL)
	    {
		printk("strioctl: I_LIST: stream %s: NULL queue ptrs\n",
				hd->sd_name) ;
		RTN(0) ;			/* really a bug */
	    }

	    for (nmods = 0; 
		 q && list.sl_nmods-- ; 
		 q=q->q_next, list.sl_modlist++, nmods++
		)
	    {
		if (!LIS_CHECK_Q_MAGIC(q)) break ;
		if ((err = lis_copyout(f,lis_queue_name(q),
			    list.sl_modlist->l_name,
			    FMNAMESZ)) < 0)
		    RTN(err) ;
	    }
	    if ((err = lis_copyout(f, &nmods,
			    &((str_list_t *)arg)->sl_nmods, sizeof(int))) < 0)
		RTN(err) ;

	    if ( LIS_DEBUG_IOCTL )
		printk("strioctl: I_LIST: stream %s arg=non-NULL, "
		       "%d modules, return 0\n",
			hd->sd_name, nmods) ;
	    RTN(0);
	}
	break;
    case I_ATMARK:
	{
	    mblk_t	*mp ;
	    int		 rtn = 0 ;

	    /* Switch on arg and return whether the first msg in the
	     * stream head read queue is marked if arg is ANYMARK.
	     * Return whether is it the last marked message if arg is LASTMARK.
	     */
	    if (   hd->sd_wq == NULL		/* no queue */
		|| hd->sd_rq == NULL	/* no rd queue */
		|| (mp = hd->sd_rq->q_first) == NULL	/* no msg */
	       )
	    {
		rtn = 0 ;			/* obviously not marked */
	    }
	    else
	    switch (arg)
	    {
	    case ANYMARK:			/* is 1st msg marked?  */
		rtn = (mp->b_flag & MSGMARK) ? 1 : 0 ;
		break ;

	    case LASTMARK:			/* is 1st msg the last marked */
		if ( !(mp->b_flag & MSGMARK) )	/* 1st msg not marked at all */
		{
		    rtn = 0 ;			/* is not the last one, then */
		    break ;
		}

		rtn = 1 ;			/* assume no other msg marked */
		for (mp = mp->b_next; mp != NULL; mp = mp->b_next)
		{
		    if (mp->b_flag & MSGMARK)	/* some other msg is marked */
		    {
			rtn = 0 ;		/* condition not met */
			break ;
		    }
		}

		break ;

	    default:
		rtn = -EINVAL;			/* invalid argument */
	    }

	    if ( LIS_DEBUG_IOCTL )
		printk("strioctl: I_ATMARK: stream %s: arg=%ld, return %d\n",
				hd->sd_name, arg, rtn) ;
	    RTN(rtn) ;
	}
	break;
    case I_CKBAND:
	if ( LIS_DEBUG_IOCTL )
	    printk("strioctl: I_CKBAND: stream %s arg=%ld\n",
				hd->sd_name, arg) ;
	if ( arg > LIS_MAX_BAND )
	    RTN(-EINVAL);
	else {
	    long val = 0;
	    
	    err = lis_strqget(hd->sd_rq,QCOUNT,(char)arg, &val) ; 
	    if (err) 
		val = -err ;			/* failed */
	    else
		val = (val != 0) ;		/* Solaris semantics */

	    RTN(val);
	}
	break;
    case I_GETBAND:
	{
	    int res;
	    if ((err=lis_check_umem(f,VERIFY_WRITE,(char*)arg,sizeof(int)))<0)
		RTN(err);
	    if (!hd->sd_rq->q_first)
		RTN(-ENODATA);
	    res=hd->sd_rq->q_first->b_band;
	    if ( LIS_DEBUG_IOCTL )
		printk("strioctl: I_GETBAND: stream %s rslt=%d\n",
				hd->sd_name, res) ;
	    err = lis_copyout(f,&res,(char*)arg,sizeof(int));
	    RTN(err);
	}
	break;
    case I_CANPUT:
        {
	    int	res ;

	    if ( arg > LIS_MAX_BAND )
		RTN(-EINVAL);
	    else 
	    {
		res = lis_bcanputnext(hd->sd_wq, (unsigned char) arg);
		if ( LIS_DEBUG_IOCTL )
		    printk("strioctl: I_CANPUT: stream %s rslt=%d\n",
				hd->sd_name, res) ;
		RTN(res) ;
	    }
        }
	break;
    case I_SETCLTIME:
	if (arg >LIS_MAX_CLTIME)
	    RTN(-EINVAL);
	hd->sd_closetime=arg;
	if ( LIS_DEBUG_IOCTL )
	    printk("strioctl: I_SETCLTIME: stream %s arg=%ld\n",
				hd->sd_name, arg) ;
	RTN(0);
	break;
    case I_GETCLTIME:
	{
	    int res;
	    if ((err=lis_check_umem(f,VERIFY_WRITE,(char*)arg,sizeof(int)))<0)
		RTN(err);
	    res=hd->sd_closetime;
	    if ( LIS_DEBUG_IOCTL )
		printk("strioctl: I_GETCLTIME: stream %s rslt=%d\n",
				hd->sd_name, res) ;
	    err = lis_copyout(f,&res,(char*)arg,sizeof(int));
	    RTN(err);
	}
	break;
    case I_PLINK:		
	/*fall through...  */
    case I_LINK:		/* link arg(=fd) stream below this one */
				/* see magic garden 7.5.1              */
				/* also AT&T STREAMS Pgrmr's Guide Chapt 10 */
	if (F_ISSET(hd->sd_flag,STRDERR)||
	    F_ISSET(hd->sd_flag,STWRERR))
	    RTN(-EINVAL);
	if (F_ISSET(hd->sd_flag,STRHUP|STRCLOSE))
	    RTN(-ENXIO);

	if (   hd->sd_strtab->st_muxrinit == NULL
	    || hd->sd_strtab->st_muxwinit == NULL
	   )					/* must have lower qinits */
	    RTN(-EINVAL);

	CLOCKADD() ;
	err = lis_i_link(i, f, hd, (int) arg, cmd, &creds) ;
	CLOCKON() ;
	RTN(err) ;
	
    case I_PUNLINK:
	/*fall through...  */
    case I_UNLINK:
	if (F_ISSET(hd->sd_flag,STRHUP|STRCLOSE))
	    RTN(-ENXIO);
	if ((long) arg < -1)		/* compare as signed quantity */
	    RTN(-EINVAL);

	CLOCKADD() ;
	if (cmd == I_PUNLINK && arg == MUXID_ALL)
	    err = lis_unlink_all(i, f, hd, &creds) ;
	else
	    err = lis_i_unlink(i, f, hd, (int)arg, cmd, &creds ) ;
	CLOCKON() ;

	RTN(err) ;

    /*
     *  the following three _ioc_ routines do their own user-to-
     *  kernel space stuff internally; we just need to get 'arg' to
     *  them in appropriate form.
     */
    case I_LIS_PIPE:
	if (LIS_DEBUG_IOCTL)
	    printk( "strioctl: I_LIS_PIPE\n" );
	RTN(lis_ioc_pipe( (unsigned int *) arg ));
    case I_LIS_FATTACH:
	if (LIS_DEBUG_IOCTL)
	    printk( "strioctl: I_LIS_FATTACH %s to \"%s\"\n",
		    hd->sd_name, (char *)arg );
	RTN(lis_ioc_fattach( f, (char *) arg ));
    case I_LIS_FDETACH:
	if (LIS_DEBUG_IOCTL)
	    printk( "strioctl: I_LIS_FDETACH \"%s\"\n",
		    (char *)arg );
	RTN(lis_ioc_fdetach( (char *)arg ));
    /**/

    case I_LIS_QRUN_STATS:
      {
        lis_qrun_stats_t	stats ;

	err = lis_check_umem(f,VERIFY_WRITE, (char*)arg, sizeof(stats)) ; 
	if (err < 0)
	    RTN(err);

	stats.num_cpus = lis_num_cpus;
	stats.num_qrunners = lis_atomic_read(&lis_runq_cnt) ;
	stats.queues_running = lis_atomic_read(&lis_queues_running) ;
	stats.runq_req_cnt = lis_atomic_read(&lis_runq_req_cnt) ;
	memcpy(stats.runq_cnts, (void *)lis_runq_cnts, sizeof(stats.runq_cnts));
	memcpy(stats.queuerun_cnts, (void *)lis_queuerun_cnts,
						sizeof(stats.queuerun_cnts)) ;
	memcpy(stats.active_flags, (void *)lis_runq_active_flags,
						sizeof(stats.active_flags)) ;
	memcpy(stats.runq_pids, (void *)lis_runq_pids, sizeof(stats.runq_pids));
	memcpy(stats.setqsched_cnts, (void *)lis_setqsched_cnts,
					    sizeof(stats.setqsched_cnts));
	memcpy(stats.setqsched_isr_cnts, (void *)lis_setqsched_isr_cnts,
					    sizeof(stats.setqsched_isr_cnts));
	err = lis_copyout(f,&stats,(char*)arg,sizeof(stats));
	RTN(err);
      }

    case I_LIS_GETSTATS:		/* LiS only - return stats */
	err = lis_check_umem(f,VERIFY_WRITE,(char*)arg, sizeof(lis_strstats)) ; 
	if (err < 0)
	    RTN(err);

	err = lis_copyout(f,lis_strstats,(char*)arg,sizeof(lis_strstats));
	RTN(err);

    case I_LIS_PRNTSTRM:		/* LiS only - print stream */
	lis_print_stream(hd) ;
	break;

    case I_LIS_PRNTQUEUES:		/* LiS only - print queues */
	lis_print_queues() ;
	break;

    case I_LIS_PRNTMEM:			/* LiS only - print memory */
	lis_print_mem() ;		/* maybe an suser() check */
	break;

    case I_LIS_PRNTSPL:
	lis_print_spl_track() ;
	break ;

    case I_LIS_SDBGMSK:			/* LiS only - set debug mask */
        CLOCKOFF(IOCTLTIME) ;		/* may reset timing debug bit */
	lis_debug_mask = arg ;		/* maybe an suser() check */
        CLOCKON() ;
	break;
    case I_LIS_SDBGMSK2:		/* LiS only - set 2nd debug mask */
	lis_debug_mask2 = arg ;		/* maybe an suser() check */
	break;

    case I_LIS_GET_MAXMSGMEM:		/* LiS only - get max msg mem */
	err = lis_check_umem(f,VERIFY_WRITE,(char*)arg, sizeof(long)) ; 
	if (err < 0)
	    RTN(err);

	err=lis_copyout(f,&lis_max_msg_mem,(char*)arg,sizeof(lis_max_msg_mem));
	break;

    case I_LIS_SET_MAXMSGMEM:		/* LiS only - set max msg mem */
	if (!lis_suser(f))
	    RTN(-EINVAL) ;

	lis_max_msg_mem = arg ;
	lis_atomic_set(&lis_strstats[MSGMEMLIM][MAXIMUM], lis_max_msg_mem) ;
	break;

    case I_LIS_GET_MAXMEM:		/* LiS only - get max mem */
	err = lis_check_umem(f,VERIFY_WRITE,(char*)arg, sizeof(long)) ; 
	if (err < 0)
	    RTN(err);

	err = lis_copyout(f,&lis_max_mem,(char*)arg,sizeof(lis_max_mem));
	break;

    case I_LIS_SET_MAXMEM:		/* LiS only - set max mem */
	if (!lis_suser(f))
	    RTN(-EINVAL) ;

	lis_max_mem = arg ;
	lis_atomic_set(&lis_strstats[MEMLIM][MAXIMUM], lis_max_mem) ;
	break;


    default:
	/* encapsulate ioctl(fd,cmd,arg)as if the user specified 
	 * instead     ioctl(fd,I_STR,{cmd,arg})
	 * and pass down to strdoioctl()
	 */
	{
	    strioctl_t ioc;
	    int res;
	    ioc.ic_cmd=cmd;
	    ioc.ic_timout=LIS_LNTIME;
	    ioc.ic_len=TRANSPARENT;
	    ioc.ic_dp=(char*)arg;
	    CLOCKADD() ;
	    res=lis_strdoioctl(f,hd,&ioc,&creds,0);	/* no copyin */
	    CLOCKON() ;
	    if ( LIS_DEBUG_IOCTL )
		printk("strioctl: stream %s "
		       "ic_cmd=0x%x ic_timout=%d ic_len=0 rslt=%d\n",
			    hd->sd_name,
			    ioc.ic_cmd, ioc.ic_timout, res) ;
	    RTN(res);
	}
    }/*sw cmd*/

    /*
     * Fall into return point and return the value of "err".
     */
return_point:
    if (hd_locked)
	lis_spin_unlock(&hd->sd_lock) ;		/* unlock strm head */
    CLOCKOFF(IOCTLTIME) ;
    lis_atomic_dec(&lis_in_syscall) ;	/* done processing a system call */
    lis_runqueues();
    return(err) ;

#undef RTN

}/*lis_strioctl*/

/*  -------------------------------------------------------------------  */
/* lis_poll_bits - evaluate poll bits for a stream			 */
/*
 * Used by lis_strpoll and by lis_poll_2_1 to evaluate a stream and
 * return poll events appropriate to the stream.
 */
unsigned lis_poll_bits(stdata_t *hd)
{
    int			 revents   = 0 ;
    int			 mtype ;
    mblk_t		*mp ;

#define	RTN(msk)	do { revents = msk; goto rtn_point; } while (0)

    if ( F_ISSET(hd->sd_flag,(STRDERR|STWRERR)) )
	RTN(POLLERR) ;

    if (F_ISSET(hd->sd_flag,(STRHUP|STRCLOSE)))
	RTN(POLLHUP) ;

    if (F_ISSET(hd->sd_flag,STPLEX))
    {
	if (LIS_DEBUG_POLL)
	  printk("strpoll: stream %s: stream has been I_LINKed\n", hd->sd_name);

	RTN(POLLNVAL) ;
    }

    if ((mp = hd->sd_rq->q_first) != NULL)	/* mp is 1st msg */
    {						/* a msg can be read */
	mtype = lis_btype(mp) ;
	if (lis_hipri(mtype))
		revents |= POLLPRI ;
	else
	{
	    revents |= POLLIN ;			/* other than hi pri msg */
	    switch (mtype)
	    {
	    case M_DATA:
	    case M_PROTO:
		if (mp->b_band > 0)
		    revents |= POLLRDBAND ;
		else
		    revents |= POLLRDNORM ;
		break ;

	    case M_SIG:			/* I don't think this can happen */
		if (*mp->b_rptr == SIGPOLL)
		    revents |= POLLMSG ;
		break ;
	    }
	}
    }

    if (lis_bcanputnext(hd->sd_wq, 0))
	revents |= (POLLWRNORM | POLLOUT) ;	/* can write band 0 data */

    if (lis_bcanputnext_anyband(hd->sd_wq->q_next))
	revents |= POLLWRBAND ;			/* can write band > 0 */

rtn_point:
    if ( LIS_DEBUG_POLL )
	printk("lis_poll_bits: stream %s: revents:%s\n",
		hd->sd_name,
	        lis_poll_events((short)revents)) ;

    return(revents) ;				/* return events */

#undef RTN

} /* lis_poll_bits */

/*  -------------------------------------------------------------------  */
/* strpoll - called from lis_syspoll() poll() syscall handler
 *
 * The STREAMS poll function is called with the file info and
 * a pointer to a polldat structure.  The STREAMS routine
 * interrogates the file for the requested conditions according
 * to the 'pd_events' field in the polldat structure.  It returns
 * a bit-mask of the events that satisfy the conditions, if any.
 * The STREAMS routine also sets the pd_headp pointer to point to
 * its list head in its stream structure for the polldat list.
 * The caller takes care of linking the structure into the list.
 */
#ifdef PORTABLE_POLL
int
lis_strpoll(struct inode *i, struct file *f, void *ptr)
{
    polldat_t		*pdat_ptr  = (polldat_t *) ptr ;
    int			 revents   = 0 ;
    int			 events    = pdat_ptr->pd_events | POLLERR | POLLHUP ;
    stdata_t		*hd;
    unsigned long  	 time_cell = 0 ;

    (void) f ;				/* reference 'f' for compiler */

    if (   !(hd = FILE_STR(f))
	|| hd->magic != STDATA_MAGIC
       )
	return(POLLNVAL);

    CLOCKON() ;

    revents = lis_poll_bits(hd) & events ;	/* evaluate the stream */

    pdat_ptr->pd_headp = &hd->sd_polllist ;	/* ptr to list hd struct */
						/* allows insertion */
    if ( LIS_DEBUG_POLL )
	printk("strpoll: stream %s: events:%s revents:%s\n",
		hd->sd_name,
	      lis_poll_events((short)events), lis_poll_events((short)revents)) ;

    CLOCKOFF(POLLTIME) ;

    return(revents) ;				/* return events */

#undef RTN

} /* lis_strpoll */
#endif

/*  -------------------------------------------------------------------  */
/*  lis_wakeup_close
 *
 * This routine is used as a timout routine and as a utility function.
 * The argument is really a pointer to a streamhead structure.
 */
void	lis_wakeup_close(caddr_t arg)
{
    stdata_t	*hd = (stdata_t *) arg ;

    if (hd->magic == STDATA_MAGIC)
    {
	CP(hd,hd->sd_flag) ;
	untimeout(hd->sd_close_timer) ;		/* cancel timer, just in case */
	hd->sd_close_timer = 0 ;
	lis_up(&hd->sd_closing) ;			/* semaphore wakeup */
    }
    else
	printk("lis_wakeup_close: hd=%lx bad magic: %lx\n",
		(long)hd, hd->magic);

} /* lis_wakeup_close */

/*  -------------------------------------------------------------------  */
/* lis_wakeup_flush
 *
 * Called if the M_FLUSH sent downstream at close time does not come
 * back to the stream head in a timely manner.
 */
void	lis_wakeup_flush(caddr_t arg)
{
    stdata_t	*hd = (stdata_t *) arg ;

    if (hd->magic == STDATA_MAGIC)
    {
	CP(hd,hd->sd_flag) ;
	untimeout(hd->sd_close_timer) ;		/* cancel timer, just in case */
	hd->sd_close_timer = 0 ;
	F_CLR(hd->sd_flag, STRFLUSHWT) ;
	lis_wakeup_close_wt(hd) ;
    }
    else
	printk("lis_wakeup_flush: hd=%lx bad magic: %lx\n",
		(long)hd, hd->magic);

} /* lis_wakeup_flush */

/*  -------------------------------------------------------------------  */
/*  -------------------------------------------------------------------  */
/* deschedule - remove the queue from the scan list and q-run list
 *
 * The queue is unlocked when this routine is called.
 */
static void deschedule(queue_t *q)
{
    int		 psw;
    int		 psw1, psw2 ;
    queue_t	*rq, *wq ;
    queue_t	*p_scan ;
    queue_t	*p_prev ;

    rq = RD(q) ;
    wq = WR(q) ;
    CP(rq,0) ;
    CP(wq,0) ;
    lis_spin_lock_irqsave(&lis_qhead_lock, &psw) ;
    if (F_ISSET(wq->q_flag,QSCAN))
    {					/* remove from the scan list */
	CP(q,0) ;
	p_prev = NULL;
	p_scan = (queue_t *) lis_scanqhead;
	while (p_scan != NULL)
	{					/* only wr q in scan list */
	    if (!LIS_CHECK_Q_MAGIC(p_scan)) break ;

	    if (p_scan == wq)
	    {					/* p_scan points to our queue */
	      lis_atomic_dec(&lis_runq_req_cnt) ; /* one less thing to do */
	      LIS_QISRLOCK(p_scan, &psw1) ;	/* prevent ISR operations */
	      CP(p_scan,0) ;
	      if (p_prev != NULL)		/* trailing ptr valid?  */
		  p_prev->q_scnxt = p_scan->q_scnxt ;	/* link around q */
	      else				/* prev still is q_head */
		  lis_scanqhead = p_scan->q_scnxt ;	/* link hd around q */
						/* NULL if last elt in list */
	      if (lis_scanqtail == p_scan)	/* last elt in list? */
		  lis_scanqtail = p_prev ;	/* back up tail ptr */
						/* NULL if only elt in list */
	      /* do not move prev ptr */
	      LIS_QISRUNLOCK(p_scan, &psw1) ;
	    }
	    else				/* not our element */
		p_prev = p_scan;		/* keep trailing ptr */

	    p_scan = p_scan->q_scnxt ;		/* go to next elt of list */
	}

	LIS_QISRLOCK(wq, &psw1) ;		/* prevent ISR operations */
	wq->q_scnxt = NULL ;
	wq->q_flag &= ~QSCAN ;
	LIS_QISRUNLOCK(wq, &psw1) ;
    }

    if ( (wq->q_flag | rq->q_flag) & QENAB )
    {
	p_prev = NULL;
	p_scan = (queue_t *) lis_qhead;

	/* Unschedule the service procedures.
	 */
	CP(q,0) ;

	if (   (lis_qhead != NULL && lis_qtail == NULL)
	    || (lis_qhead == NULL && lis_qtail != NULL)
	   )
	    printk("LiS: deschedule before: Qhead error: "
		   "lis_qhead=%lx lis_qtail=%lx\n",
		   lis_qhead, lis_qtail) ;

	while (p_scan != NULL)
	{
	    if (!LIS_CHECK_Q_MAGIC(p_scan)) break ;

	    if (p_scan == rq || p_scan == wq)
	    {					/* p_scan points to our queue */
	      lis_atomic_dec(&lis_runq_req_cnt) ; /* one less thing to do */
	      CP(p_scan,lis_qhead) ;
	      CP(p_prev,lis_qtail) ;
	      if (p_prev != NULL)		/* trailing ptr valid?  */
		  p_prev->q_link = p_scan->q_link ;	/* link around q */
	      else				/* prev still is q_head */
		  lis_qhead = p_scan->q_link ;	/* link hd around q */
						/* NULL if last elt in list */
	      if (lis_qtail == p_scan)		/* last elt in list? */
		  lis_qtail = p_prev ;		/* back up tail ptr */
						/* NULL if only elt in list */
	      /* do not move prev ptr */
	      LisDownCount(QSCHEDS) ;
	    }
	    else				/* not our element */
		p_prev = p_scan;		/* keep trailing ptr */

	    p_scan = p_scan->q_link ;		/* go to next elt of list */
	}

	if (   (lis_qhead != NULL && lis_qtail == NULL)
	    || (lis_qhead == NULL && lis_qtail != NULL)
	   )
	    printk("LiS: deschedule after: Qhead error: "
		   "lis_qhead=%lx lis_qtail=%lx\n",
		   lis_qhead, lis_qtail) ;

	LIS_QISRLOCK(wq, &psw1) ;		/* prevent ISR operations */
	wq->q_link = NULL ;
	wq->q_flag &= ~QENAB ;
	LIS_QISRUNLOCK(wq, &psw1) ;
	LIS_QISRLOCK(rq, &psw2) ;
	rq->q_link = NULL ;
	rq->q_flag &= ~QENAB ;
	LIS_QISRUNLOCK(rq, &psw2) ;
    }

    lis_spin_unlock_irqrestore(&lis_qhead_lock, &psw) ;
}

/* mark_closing - mark both read and write queues as closing, all
 * the way down to the driver's queues.
 */
static void mark_closing(queue_t *sd_wq)
{
    queue_t	*sd_rq ;
    queue_t	*q ;
    int		 psw;
    int		 psw_wq, psw_rq ;

    CP(sd_wq,0) ;
    do
    {
	sd_rq = RD(sd_wq) ;

	CP(sd_wq,sd_rq) ;
						/* prevent qenable */
	lis_spin_lock_irqsave(&lis_qhead_lock, &psw) ;
	LIS_QISRLOCK(sd_wq, &psw_wq) ;		/* prevent ISR operations */
	LIS_QISRLOCK(sd_rq, &psw_rq) ;
	CP(sd_wq,0) ;
	sd_wq->q_flag |= (QNOENB | QCLOSING);	/* prevent qenables */
	sd_rq->q_flag |= (QNOENB | QCLOSING);	/* prevent qenables */
	LIS_QISRUNLOCK(sd_rq, &psw_rq) ;
	LIS_QISRUNLOCK(sd_wq, &psw_wq) ;
	lis_spin_unlock_irqrestore(&lis_qhead_lock, &psw) ;

	deschedule(sd_wq) ;			/* remove from sched lists */

	q = sd_wq ;				/* save for loop iteration */
	sd_wq = sd_wq->q_next ;
    }
    while (sd_wq != NULL && SAMESTR(q)) ;
}

/*  -------------------------------------------------------------------  */
/* doclose - dec refcnt & dismantle the stream when sd_refcnt gets 0.
 *
 * This routine is called from several different contexts.  One is from
 * lis_strclose which is on the straightforward close path in closing
 * a file.  Another is from the unlink code when a stream is unlinked
 * from below a multiplexor and this is the last use of the stream.  In the
 * latter case, both file and inode pointers will be NULL since the file
 * and inode are both long gone.  Finally, it is called from fdetach.
 */
int
lis_doclose( struct inode *i, struct file *f, stdata_t *head, cred_t *creds )
{
    int		   refcnt ;
    unsigned long  time_cell = 0 ;
    long           this_doclose ;
    int		   rslt ;

    CLOCKON() ;

    if (head == NULL || head->magic != STDATA_MAGIC)
    {
	printk("lis_doclose: stream head=%lx: NULL or bad magic (0x%x)\n",
		(long)head, head != NULL ? head->magic : 0) ;
	return(-EINVAL) ;
    }

    CP(head,0) ;
    this_doclose = lis_incr(&lis_doclose_cnt);

#if defined(LINUX)
    /*
     *  FIFOs/pipes: synchronize at close time before anything else
     */
    if (i && f && F_ISSET(head->sd_flag,STFIFO))
	lis_fifo_close_sync( i, f );
#endif

    lis_del_from_elist(&head->sd_siglist,PID(f),S_ALL);

    if ( LIS_DEBUG_CLOSE )
    {
	printk("doclose#%ld/%ld: %s refcnt=%d linkcnt=%d"
	       " (%ld open stream(s))\n",
	       this_doclose, lis_close_cnt,
	       head->sd_name,
	       LIS_REFCNT(head), head->sd_linkcnt, lis_stdata_cnt ) ;
	printk("         head 0x%x wq:\"%s\" next wq:\"%s\"\n",
	       (int)head, lis_queue_name(head->sd_wq),
	       (SAMESTR(head->sd_wq)?
		lis_queue_name(head->sd_wq->q_next):"<none>"));
    }

    /*
     *  if we're hungup and fattached but no M_PASSFPs are queued,
     *  fdetach now
     */
    if ((head->sd_flag & (STRHUP|STRATTACH)) == (STRHUP|STRATTACH) &&
	!(head->sd_rfdcnt)) {
	F_CLR(head->sd_flag,STRATTACH);
	CP(head,0) ;
	if (LIS_DEBUG_FATTACH)
	    printk("calling lis_fdetach_stream(0x%x) from %s line %d\n",
		   (int)head, __FILE__, __LINE__ );
	lis_fdetach_stream(head);
	if (LIS_DEBUG_FATTACH)
	    printk("called lis_fdetach_stream(0x%x) from %s line %d\n",
		   (int)head, __FILE__, __LINE__ );
    }

    lis_atomic_dec(&head->sd_refcnt) ;		/* decr ref count */
    refcnt = LIS_REFCNT(head) ;			/* will return this */

    /*
     * If the reference count is now down to the plink count
     * then all that is left are the plinks.  This means that
     * this is really the last close of the stream and the inode
     * is about to disappear.  We clear the inode ptr in the
     * stdata structure.  There will be no more accesses of
     * this stream from above via the file system.
     *
     * The stream remains allocated but cannot be reached via an
     * open.  Open will allocate a new stdata structure upon a
     * re-open.
     */
    if (head->sd_linkcnt != 0 && LIS_REFCNT(head) <= head->sd_linkcnt)
    {
	lis_free_elist(&head->sd_siglist);	/* do in signal masks */
    }
    
    if (LIS_REFCNT(head) != 0)			/* still in use */
    {
	CP(head,0) ;
	lis_spin_unlock(&head->sd_lock) ;	/* unlock strm head */
	goto return_point ;
    }

    CP(head,0) ;
    head->sd_flag |= STRCLOSE ;

    /*
     *  if we're a pipe, set the peer's STRHUP flag (but don't hang up
     *  the peer - it may still have to receive passed FDs, etc...)
     */
    if (head->sd_peer &&
	head->sd_peer != head &&
	!(head->sd_peer->sd_flag & (STRHUP|STRCLOSE)))
    {
	CP(head,0) ;
	head->sd_peer->sd_flag |= STRHUP;
	head->sd_peer->sd_peer = NULL;
	lis_wake_up_all_wwrite(head->sd_peer);
	lis_wake_up_all_wread(head->sd_peer);
	if (   POLL_WAITING(head->sd_peer)
	    || F_ISSET(head->sd_peer->sd_flag,STRSELPND)
	)
	    lis_wake_up_poll(head->sd_peer,POLLHUP);
#if 1
	if (F_ISSET(head->sd_peer->sd_flag,STRATTACH) &&
	    !(head->sd_peer->sd_rfdcnt))
	{
	    if (LIS_DEBUG_FATTACH)
		printk(
		    "calling lis_fdetach_stream(0x%x) from %s line %d\n",
		    (int)head->sd_peer, __FILE__, __LINE__ );
	    lis_fdetach_stream(head->sd_peer);
	    if (LIS_DEBUG_FATTACH)
		printk(
		    "called lis_fdetach_stream(0x%x) from %s line %d\n",
		    (int)head->sd_peer, __FILE__, __LINE__ );
	}
#endif
    }
    if (   head->sd_closetime != 0		/* have close time set */
	&& lis_qcountstrm(head->sd_wq) != 0	/* messages queued */
       )
    {					/* set a timer */
	CP(head,0) ;
	if ( LIS_DEBUG_CLOSE )
	  printk(
	    "doclose: stream %s waiting %dms for queues to drain\n",
		    head->sd_name,
		    head->sd_closetime);

	if (SAMESTR(head->sd_wq))
	    head->sd_wq->q_next->q_flag |= QCLOSEWT;
	head->sd_close_timer = lis_timeout_fcn(lis_wakeup_close,
				       (caddr_t) head,
				   lis_milli_to_ticks(head->sd_closetime),
				       "stream-close", MEM_TIMER) ;
	lis_spin_unlock(&head->sd_lock) ;	/* unlock strm head */
	lis_up(&lis_close_sem);			/* allow other closes */
	CLOCKADD() ;
	lis_atomic_dec(&lis_in_syscall) ;	/* "done" with a system call */
	lis_runqueues() ;
	lis_down(&head->sd_closing) ;		/* sleep for awhile */
	CLOCKON() ;
	if ((rslt = lis_down_nintr_head(head, &lis_close_sem)) < 0)
	{					/* very busted */
	    printk("lis_doclose: Error %d on down(&lis_close_sem)\n", rslt);
	    return(0) ;
	}
	lis_spin_lock(&head->sd_lock) ;		/* lock strm head */
	lis_atomic_inc(&lis_in_syscall) ;	/* processing a system call */
	if (head->sd_close_timer)
	{
	    untimeout(head->sd_close_timer) ;
	    head->sd_close_timer = 0 ;
	}
	CP(head,0) ;
    }

    lis_free_elist(&head->sd_siglist);		/* do in signal masks */
    if (head->sd_mux.mx_hd != NULL)		/* is a ctl stream */
    {
	CP(head,0) ;
	CLOCKADD() ;
	lis_i_unlink(i, f, head, -1,
		     I_PUNLINK, creds) ;	/* unlink all lowers */
	CLOCKON() ;
	CP(head,0) ;
    }

    /*
     * The possible I_UNLINK above is that last legitimate message that can
     * be sent on the stream.  From now on message traffic is suppressed.
     *
     * By locking the queue we assure ourselves that there are no put
     * procedures running on other CPUs.  Then we prevent further qenables
     * on the stream head queue pair.  This keeps the service procedures
     * from running.  Since we haven't called the driver's close routine
     * yet, there could still be some interrupt level activity on the q.
     *
     * lis_qdetach will sync up with any residual service procedures
     * running on other CPUs.
     */
    CP(head,head->sd_wq) ;
    mark_closing(head->sd_wq) ;

    /*
     * Flush the stream and wait for the M_FLUSH to come back to
     * the stream head.  Then close all the queues.  We run a short
     * timer in case the M_FLUSH gets lost downstream somewhere.
     *
     * In the case of a STREAMS pipe, if the other end has already
     * closed then there will be no "downstream" to flush.
     */
    if (head->sd_wq->q_next != NULL)		/* is there a "downstream" */
    {
	F_SET(head->sd_flag,STRCLOSEWT);	/* allows wakeups */
	F_SET(head->sd_flag,STRFLUSHWT);
	head->sd_close_timer = lis_timeout_fcn(lis_wakeup_flush,
					   (caddr_t) head,
					   lis_milli_to_ticks(50),
					   "stream-flush", MEM_TIMER) ;
	lis_snd_mflush(head->sd_wq, FLUSHRW, 0);
	CP(head,head->sd_flag) ;
	rslt = 0 ;
	while (   F_ISSET(head->sd_flag,STRFLUSHWT)
	       && (rslt = lis_sleep_on_close_wt(head)) == 0
	      )
	    CP(head,head->sd_close_timer) ;

	if (head->sd_close_timer)
	{
	    untimeout(head->sd_close_timer) ;
	    head->sd_close_timer = 0 ;
	}

	if (rslt < 0 && rslt != -EINTR)
	    printk("lis_doclose: sleep_on_close_wt returned %d\n", rslt) ;
    }

    CP(head,head->sd_flag) ;
    while (!pop_mod( head, head->sd_open_flags, creds )) {};

    CLOCKADD() ;
    CP(head,head->sd_wq) ;
    if (SAMESTR(head->sd_wq))
	lis_qdetach( LIS_RD(head->sd_wq->q_next), 1, 
		     head->sd_open_flags, creds );
    CLOCKON() ;

    if (f) 
    {
	SET_FILE_STR(f, NULL);
	f->f_op = NULL;
	lis_clean_up_file(f, head, 0) ;
    }

    if (i)
    {
	SET_INODE_STR(i, NULL);			/* detach from inode */
						/* caller will "iput" */
    }

    /*
     * lis_dismantle unlocks the "open lock".  It doesn't do much good as
     * the stream head is about to be returned to the memory allocator
     * anyway.
     */
    CP(head,0) ;
    lis_dismantle(head, creds);			/* deallocate head */
					 	/* must be last thing */

return_point:
    if ( LIS_DEBUG_CLOSE )
	printk("doclose#%ld/%ld: %ld open stream(s)\n",
	       this_doclose, lis_close_cnt, lis_stdata_cnt) ;

    CLOCKOFF(CLOSETIME) ;
    return(refcnt);
								     
} /*lis_doclose*/

#if defined(KERNEL_2_1)
#define RTNX	return(0)
int
#else
#define RTNX	return
void
#endif

lis_strclose(struct inode *i, struct file *f)
{
    struct stdata *head;
    cred_t	   creds;
    unsigned long  time_cell = 0 ;
    unsigned long  this_close ;
    int		   rslt ;
    
    CLOCKON() ;

    this_close = lis_incr(&lis_close_cnt);

    if (!i) {
	lis_error(LIS_ERROR,"strclose","called w/ null inode");
	CLOCKOFF(CLOSETIME) ;
	RTNX;
    }

    if ( LIS_DEBUG_CLOSE )
    {
	printk("strclose#%ld: "
	       "i_rdev=(%d,%d) i_count=%ld f_count=%d"
#ifdef KERNEL_2_1				/* linux >= 2.1.0 */
	       " d_count=%d"
#endif
	       " (%ld open stream(s))\n",
	       this_close, STR_MAJOR(i->i_rdev), STR_MINOR(i->i_rdev),
	       I_COUNT(i),
	       F_COUNT(f)
#if defined(KERNEL_2_1)				/* linux >= 2.1.0 */
# if defined(KERNEL_2_3)
	       , atomic_read(&f->f_dentry->d_count)
# else
	       , f->f_dentry->d_count
# endif
#endif
	       , lis_stdata_cnt
	       ) ;

	if (LIS_DEBUG_ADDRS)
#ifdef KERNEL_2_1				/* linux >= 2.1.0 */
		  printk("strclose#%ld: f=%lx f->dentry=%lx i=%lx\n",
		     this_close, (long) f, (long) f->f_dentry, (long) i) ;
#else						/* linux < 2.1.0 or portable */
	      printk("strclose#%ld: f=%lx f->f_inode=%lx i=%lx\n",
		     this_close, (long) f, (long) f->f_inode, (long) i) ;
#endif
    }

    if (!I_COUNT(i))
    {
	lis_error(LIS_ERROR,"strclose",
		  "called with unused inode (inode 0x%x file 0x%x)",
		  (long)i, (long)f);
	CLOCKOFF(CLOSETIME) ;
	RTNX;
    }

    if (!S_ISCHR(i->i_mode))
    {
	CLOCKOFF(CLOSETIME) ;
	RTNX;
    }

    if (!LIS_DEVOK(STR_MAJOR(i->i_rdev)))
    {
	CLOCKOFF(CLOSETIME) ;
	RTNX;
    }

    if (!(head = FILE_STR(f))){
	lis_error(LIS_ERROR,"strclose",
		  "called with null stream (inode 0x%x file 0x%x)",
		  (long)i, (long)f);
	CLOCKOFF(CLOSETIME) ;
	RTNX;
    }    

    creds.cr_uid  = (uid_t) EUID(f);
    creds.cr_gid  = (gid_t) EGID(f);
    creds.cr_ruid = (uid_t) UID(f);
    creds.cr_rgid = (gid_t) GID(f);

    if ((rslt = lis_down_nintr_head(head, &lis_close_sem)) < 0)
    {						/* very busted */
	printk("strclose: Error %d on down(&lis_close_sem)\n", rslt) ;
	RTNX ;
    }

    CP(head,0) ;
    lis_spin_lock(&head->sd_lock) ;		/* lock strm head */
    						/* lis_doclose unlocks */
    						/* or deallocates struct */
    lis_atomic_inc(&lis_in_syscall) ;		/* processing a system call */
    CLOCKADD() ;
    if (lis_doclose(i, f, head, &creds) == 0)	/* last close, gone */
    {
#ifdef S_ZERO_WR				/* Linux 2.0.36 */
	i->i_flags &= ~S_ZERO_WR ;
#endif
	;
    }

    CP(head,0) ;
    lis_up(&lis_close_sem) ;
    lis_atomic_dec(&lis_in_syscall) ;	/* done processing a system call */
    lis_runqueues();
    RTNX ;

} /* lis_strclose */

#undef RTNX

/*  -------------------------------------------------------------------  */

/* Initialize some glob vars...
 */
void lis_init_head( void )
{
#if defined(USE_CODE_PATH)
    lis_spin_lock_init(&lis_code_path_lock, "LiS_code_path_lock") ;
#endif
    /* We'd better use defaults from strconfig instead of magic #s.
     */
    lis_qhead = lis_qtail =  lis_scanqhead= lis_scanqtail = NULL;
    lis_atomic_set(&lis_queues_running, 0);
    lis_atomic_set(&lis_runq_req_cnt, 0) ;
    lis_nstrpush = 0;
    lis_strhold = 0;
    lis_strthresh = 0;
    /*lis_iocseq=whichever get's in the air -- so we have some kind of random
     *number for this seq # <-should be fixed*/

    lis_sem_init( &lis_stdata_sem, 1 );
    lis_sem_init( &lis_close_sem, 1 );
    lis_spin_lock_init(&lis_stdata_lock, "LiS-Stdata") ;
    lis_spin_lock_init(&lis_qhead_lock, "Qhead-Lock") ;
    lis_spin_lock_init(&lis_incr_lock, "Incr-Lock") ;
    lis_spin_lock_init(&lis_msg_lock, "Msg-Lock") ;
    lis_spin_lock_init(&lis_bc_lock, "Bufcall-Lock") ;
    lis_spin_lock_init(&lis_mem_lock, "Mem-Lock") ;
    lis_spin_lock_init(&lis_tlist_lock, "Tlist-Lock") ;

    lis_init_bufcall() ;
    lis_init_mod() ;
}/*lis_init_head*/

/*  -------------------------------------------------------------------  */
/*
 * Terminate code
 */
void lis_terminate_head(void)
{
    lis_terminate_mod() ;
    lis_terminate_msg() ;
    lis_terminate_bufcall() ;
    lis_terminate_dki() ;
    lis_terminate_mem() ;
#if !defined(USER)
    lis_free_all_pages() ;		/* from lis page allocator */
#endif
    SEM_DESTROY(&lis_stdata_sem);
    SEM_DESTROY(&lis_close_sem);

} /* lis_terminate_head */


/*----------------------------------------------------------------------
# Local Variables:      ***
# change-log-default-name: "~/src/prj/streams/src/NOTES" ***
# End: ***
  ----------------------------------------------------------------------*/
